/********************************************************************
 bbox.c --
        Bounding box of mesh points and cuboid operations. 
 
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University, Edinburgh    
 ********************************************************************/

/* $Id: bbox.c,v 1.16 2002/07/28 13:27:36 peteri Exp $ */

#include <assert.h>
#include "eg_lib.h"
#include "bbox.h"
#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "tetrah.h"
#include "find.h"
#include "basic_op.h"
#include "mesh_op.h"
#include "debug.h"

static void TriangulateCuboid (TetrahMesh *tetMesh, Vertex **vert);

#define LOW_X  (low[0])
#define LOW_Y  (low[1])
#define LOW_Z  (low[2])
#define HIGH_X  (high[0])
#define HIGH_Y  (high[1])
#define HIGH_Z  (high[2])
#define MESHPOINT_X(ii) (vertex[ii]->coords[0])
#define MESHPOINT_Y(ii) (vertex[ii]->coords[1])
#define MESHPOINT_Z(ii) (vertex[ii]->coords[2])

#define SIMPLE_CHECK_POINT \
  if (LOW_X > MESHPOINT_X(i)) LOW_X = MESHPOINT_X(i); \
  if (LOW_Y > MESHPOINT_Y(i)) LOW_Y = MESHPOINT_Y(i); \
  if (LOW_Z > MESHPOINT_Z(i)) LOW_Z = MESHPOINT_Z(i); \
  if (HIGH_X < MESHPOINT_X(i)) HIGH_X = MESHPOINT_X(i); \
  if (HIGH_Y < MESHPOINT_Y(i)) HIGH_Y = MESHPOINT_Y(i); \
  if (HIGH_Z < MESHPOINT_Z(i)) HIGH_Z = MESHPOINT_Z(i); 

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   BoundingBox  
 * DESCRIPTION
 *   Calculate coordinates of extreme points of mesh points.
 * ARGUMENTS
 *   nvertex - number of boundary vertices
 *   vertex - list of boundary vertices
 *   low - pointer to double array of size 3 to store low corner (OUT)
 *   high - pointer to double array of size 3 to store high corner (OUT) 
 * RETURN VALUE
 *   NONE
 ******/
void BoundingBox (int nvertex, Vertex **vertex, double *low, double *high)
{
  int i;

  /* check validity of other pointers */
  assert (low  != NULL);
  assert (high != NULL); 

  /* initialize bounding box points */
  LOW_X = MESHPOINT_X(0); 
  LOW_Y = MESHPOINT_Y(0);
  LOW_Z = MESHPOINT_Z(0);
  HIGH_X = MESHPOINT_X(0);
  HIGH_Y = MESHPOINT_Y(0);
  HIGH_Z = MESHPOINT_Z(0);

  /* loop through all mesh points */
  for (i = 1; i < nvertex; i++)
  {
    SIMPLE_CHECK_POINT
  }
} /* end of BoundingBox */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   MakeCuboid
 * DESCRIPTION
 *   Create cuboid vertices.
 *   Coordinates of cuboid vertices are calculated by scaling
 *   coordinates of mesh bounding box. Then cuboid is triangulated creating 
 *   five tetrahedra.
 * ARGUMENTS
 *   tetMesh - tetrahedral mesh,
 *   low, high - coordinates of extreme points of mesh bounding box
 * RETURN VALUE
 *   NONE
 ******/
void MakeCuboid (TetrahMesh *tetMesh, double *low, double *high)
{
  register int i;
  double center[3]; /* coordinates of center of bounding box */
  double span; /* span of bounding box in direction of coordinate  axis */
  double max; /* maximal span of bounding box */
  double mparam;
  
  Vertex *vertex[8]; /* cuboid vertices */
  
  /* calculate maximal span of bounding box */
  max = 0;
  for (i=0; i<3; i++)
  {
    center[i] = 0.5 * (low[i] + high[i]);
    span = high[i] - low[i];
    if ( max < span) max = span;
  } 

  /* mesh parameter is always large, so cuboid is not refined */
  mparam = max * 2.0;
  
  /* extend the bounding box to form initial cuboid */
  max = 0.5 * max * CUBOID_SCALE;
  for (i=0; i<3; i++)
  {
    high[i] = center[i] +  max;
    low[i]  = center[i] -  max;
  }

  /* create cuboid vertices */
  for (i=0; i<8; i++)
  {
    vertex[i] = VertexCreate();
    vertex[i]->flag |= POINT_CUBOID + POINT_BOUNDARY;
    vertex[i]->meshParam = mparam;
    PutVertexOnListHead(tetMesh->pList, vertex[i]); 
    /* advance counter of point in the mesh */
    (tetMesh->NMeshPoints)++; 
  }

  /* assign coordinates to cuboid vertices */
  vertex[0]->coords[0] = low[0];
  vertex[0]->coords[1] = low[1];
  vertex[0]->coords[2] = low[2];

  vertex[1]->coords[0] = high[0];
  vertex[1]->coords[1] = low[1];
  vertex[1]->coords[2] = low[2];

  vertex[2]->coords[0] = low[0];
  vertex[2]->coords[1] = high[1];
  vertex[2]->coords[2] = low[2];

  vertex[3]->coords[0] = high[0];
  vertex[3]->coords[1] = high[1];
  vertex[3]->coords[2] = low[2];

  vertex[4]->coords[0] = low[0];
  vertex[4]->coords[1] = low[1];
  vertex[4]->coords[2] = high[2];

  vertex[5]->coords[0] = high[0];
  vertex[5]->coords[1] = low[1];
  vertex[5]->coords[2] = high[2];

  vertex[6]->coords[0] = low[0];
  vertex[6]->coords[1] = high[1];
  vertex[6]->coords[2] = high[2];
  
  vertex[7]->coords[0] = high[0];
  vertex[7]->coords[1] = high[1];
  vertex[7]->coords[2] = high[2];
  
  /* triangulate cuboid  */
  TriangulateCuboid (tetMesh, vertex);

} /* end of MakeCuboid */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   TriangulateCuboid 
 * DESCRIPTION
 *   Triangulate initial cuboid in a standard way creating 5 tetrahedra. 
 * ARGUMENTS
 *   tetMesh - pointer to tetrahedral mesh data structure
 *   vert - pointer to array of pointer to cuboid vertices
 * RETURN VALUE
 *   NONE
 ******/
static void TriangulateCuboid (TetrahMesh *tetMesh, Vertex **vert)
{
  register int i; 
  register int j;
  int n;
  Tetrah *tet[5];
  /* standard cuboid tetrahedralization into 5 tetrahedrons */
  static int stdCuboidTet[5][8] =
      {
          {0, 1, 2, 4,  4, -1, -1, -1},
          {4, 5, 7, 1,  -1, 4, -1, -1}, 
          {1, 3, 2, 7,  -1, 4, -1, -1},
          {7, 6, 4, 2,  -1, 4, -1, -1},
          {7, 4, 2, 1,   0, 2,  1,  3}
      };

  /* create cuboid tetrahedra and put them on tetrahedron list */
  for (i=0; i<5; i++)
  {
    tet[i] = TetrahCreate();
    PutTetOnListHead(tetMesh->tList, tet[i]);
    /* advance counter of tetrahedra */
    (tetMesh->NTetrah)++;
  }    
  /* fill adjacency information */
  for (i=0; i<5; i++)
  {
    TET_FLAG(tet[i]) = 0;
    for (j=0; j<4; j++)
    {
      TET_VERTEX(tet[i], j) = vert[stdCuboidTet[i][j]];
      n =stdCuboidTet[i][j+4];
      TET_NEIGHBOUR(tet[i], j) = (n == -1) ? (Tetrah *)NULL : tet[n];
    }
  } 
  return;
} /* end of TriangulateCuboid */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   ExtractMeshFromCuboid
 * DESCRIPTION
 *   There are two ways to identify all the tetrahedra which make
 *   up the final mesh. In the first case all internal tetrahedra
 *   is identified, while in the other case all external tetrahedra
 *   is identified. However in the first case an internal point
 *   must be supplied for the implementation below, therefore the
 *   second is used currently !!!
 * ARGUMENTS
 *   mgds - pointer to mesh generator data structure 
 * RETURN VALUE
 *   NONE
 ******/
void ExtractMeshFromCuboid (MeshGenData *mgds)
{

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- Domain recovery ...\n");
  }
  
  IdentifyOuterTetFlood (mgds);
  DeleteOuterTetrah (&(mgds->tetMesh));
  RemoveUnconnectedVertices (&(mgds->tetMesh)); 

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- Domain recovery finished\n");
  }
  
} /* end of ExtractMeshFromCuboid */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   IdentifyOuterTetFlood
 * DESCRIPTION
 *   Identify all tetrahedra between the cuboid and the mesh
 * ARGUMENTS
 *   mgds - pointer to mesh generatro data structure
 * RETURN VALUE
 *   NONE
 ******/
void IdentifyOuterTetFlood (MeshGenData *mgds)
{
  Tetrah    *tet, *neighbour;
  int        nface;
  Tos_MStack checkStack;
  Face       face;
  int        np, i;
  Vertex    *p;
  
  checkStack = Tos_MStackNew ();

  /* the first 8 points should be the cuboid points */
  np = 0;
  p = mgds->tetMesh.pList->tail;
  for(i = 0; i < mgds->tetMesh.NMeshPoints; i++)
  {
    if((VERTEX_FLAG(p) & POINT_CUBOID) == POINT_CUBOID)
    {
      /* store all tetrahedron adjacent to the cuboid points and
       * mark them as well */
      Tos_MStackTraversalInit(p->adjElem);
      while((tet = Tos_MStackTraverse(p->adjElem)) != NULL)
      {
        Tos_MStackPush(checkStack, (void *)tet);
        TET_FLAG(tet) |= TET_OUT_DOMAIN;
      }
      np++;
      /* only eight cuboid points */
      if(np >= 8)
          break;
    }
  }
  assert(i == 7);
  
  /* go through the stack */
  while (Tos_MStackCount (checkStack) > 0)
  {
    tet = (Tetrah *) Tos_MStackPop (checkStack);
    for (nface=4; nface--;)
    {
      neighbour = TET_NEIGHBOUR (tet, nface); 
      /*assert (neighbour != NULL);*/
      GetFaceFromTet (tet, nface, &face);
      /* if neighbour is boundary of cuboid or already marked or
       * boundary of the inserted mesh then simply try other
       * tetrahedra */
      if (   neighbour == NULL
          || ((TET_FLAG(neighbour) & TET_OUT_DOMAIN) == TET_OUT_DOMAIN)
          || IsFaceOnMeshBoundary (mgds, 0, &face, POINT_EXT_BOUNDARY) == 1)
      {
        continue;
      }
      else
      {
        Tos_MStackPush (checkStack, (void *) neighbour);
        TET_FLAG(neighbour) |= TET_OUT_DOMAIN;
      }
    }
  }
  /* stack is empty, free it up */
  Tos_MStackFree (&checkStack);
  return; 
} /* end of IdentifyOuterTetFlood */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   DeleteOuterTetrah 
 * DESCRIPTION
 *   Delete from mesh all tetrahedra which are outside
 * ARGUMENTS
 *   tetMesh - pointer to tetrahedral mesh data structure
 * RETURN VALUE
 *   NONE
 ******/
void DeleteOuterTetrah (TetrahMesh *tetMesh)
{
  Tetrah *tet;
  Tetrah *tetToCheck;
  
  tet = LIST_TAIL(tetMesh->tList);

  while (tet != NULL)
  {
    tetToCheck = tet;
    tet = TET_NEXT(tet);  
    if ((TET_FLAG (tetToCheck) & TET_OUT_DOMAIN) == TET_OUT_DOMAIN)
    {
      RemoveTetFromMesh (tetMesh, tetToCheck);
      TetrahDelete (tetToCheck);
    }  
  } 
} /* end of DeleteOuterTetrah */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   RemoveUnconnectedVertices 
 * DESCRIPTION
 *   Reomves from the mesh vertices which are not connected to and elemetn.
 * ARGUMENTS
 *   tetMesh - pointer to tetrahedral mesh data structure.
 * RETURN VALUE
 *   NONE
 ******/
void RemoveUnconnectedVertices(TetrahMesh *tetMesh)
{
  Vertex *vertex;
  Vertex *vertexToCheck;
  int counter = 0;
  
  vertex = LIST_TAIL(tetMesh->pList);

  while (vertex != NULL)
  {
    vertexToCheck = vertex;
    vertex = VERTEX_NEXT(vertex);  
    if (Tos_MStackCount(vertexToCheck->adjElem) == 0)
    {
      RemoveVertexFromList(tetMesh->pList, vertexToCheck);
      (tetMesh->NMeshPoints)--;
      counter++;
      VertexDelete (vertexToCheck);
    }  
  }

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- %d unconnected vertices removed succesfully\n",
           counter);
  }
} /* end of RemoveUnconnectedVertices */



/* Modifications:
 * $Log: bbox.c,v $
 * Revision 1.16  2002/07/28 13:27:36  peteri
 * - list creation interface is modified
 * - it is possible to add to the head or the tail of the list
 * - all related code is updated
 *
 * Revision 1.15  2002/07/26 14:59:43  peteri
 * - mesh data is separated from boundary data
 * - now it is possible to setup the boundary data separately
 *   and refine that boundary (without having a MESH)
 *
 * Revision 1.14  2002/07/26 11:35:10  peteri
 * nana is removed for simplicity
 *
 * Revision 1.13  2002/07/20 18:11:24  peteri
 * Code clean-up, alternative implementation is removed
 *
 * Revision 1.12  2002/07/20 13:20:46  peteri
 * Cuboid points are also marked as boundary points,
 * so swapping will not consider them.
 *
 * Revision 1.11  2002/03/18 15:53:45  peteri
 * Some renaming and cleaning
 *
 * Revision 1.10  2002/02/24 19:00:58  romek
 * *** empty log message ***
 *
 * Revision 1.9  2002/02/16 18:55:44  romek
 * Doc fix.
 *
 */
