/* $Id: boundary.c,v 1.11 2002/07/28 18:24:25 peteri Exp $ */
#include <assert.h>

#include "nodelem.h"
#include "triangadj.h"
#include "basic_op.h"
#include "boundary.h"
#include "debug.h"

static void BoundEdges (MeshGenData *mgds);
static void BoundFaces (MeshGenData *mgds);
static void BoundPoints (MeshGenData *mgds);

/****f*
 * NAME
 *   GetBoundaryRep 
 * DESCRIPTION
 *   Create boundary representation of object to be triangulated.
 *        
 *   At the moment we assume that the input to the mesh generator 
 *   consists of already tiangulated surface meshes.
 *   In this case the only role of this function is to fill the tables
 *   of boundary points, boundary edges and boundary faces from the
 *   input meshes.
 * ARGUMENTS
 *   mgds - pointer to mesh generator data structure
 * RETURN VALUE
 *   NONE
 ******/
void GetBoundaryRep (MeshGenData *mgds)
{
  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- Getting Boundary Representation\n");
  }
  
  /* get boundary points */
  BoundPoints (mgds);
  /* get boundary faces */
  BoundFaces (mgds);
  /* get boundary edges */
  BoundEdges (mgds); 

} /* end of GetBoundaryRep */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 BoundPoints --
        This function creates points from input meshes.
        It also sets flags for boundary points.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void BoundPoints (MeshGenData *mgds)
{
  register int j;
  register int i;
  int pointsCounter=0; 

  if(vm_verbose_level >= 3)
  {
    printf("+ debug 3 -- BoundPoints\n");
  }
  
  assert (mgds != NULL);
  
  /* for each input mesh */
  for (i = 0; i < mgds->NMeshes; i++)
  {
    mgds->pTableLength[i] = (mgds->mgCtrl->inputMeshes[i])->NMeshPoints;
    /* allocate memory for boundary point table */
    mgds->pTable[i] = (Vertex **)e_calloc (mgds->pTableLength[i], 
                                           sizeof (Vertex *));
    /* fill table of boundary points */
    for (j = 0; j < mgds->pTableLength[i]; j++)
    {
      /* create point */
      mgds->pTable[i][j] = VertexCreate();
      pointsCounter++;
      /* copy coordinates */
      memcpy ((void *)(mgds->pTable[i][j]->coords),
              (void *)((mgds->mgCtrl->inputMeshes[i])->MeshPointCoords[j]),
              3*sizeof(double));
      /* set boundary point flag */
      (mgds->pTable[i][j])->flag |= POINT_BOUNDARY;
      if (i == 0)
      { /* set external boundary point flag */
        (mgds->pTable[i][j])->flag |= POINT_EXT_BOUNDARY;
      }  
      else
      { /* set internal boundary flag */
        (mgds->pTable[i][j])->flag |= POINT_INT_BOUNDARY; 
      }
      /* set mesh parameter */
      (mgds->pTable[i][j])->meshParam = 
          (mgds->mgCtrl->inputMeshes[i])->MeshParam.NodalMeshParams[j];

    }
  }

  if(vm_verbose_level >= 4)
  {
    printf("+ debug 4 -- \t Total number of input points: %d\n", 
           pointsCounter);
  }
  
} /* end of BoundPoints */




/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 BoundFaces --
      This function creates the boundary faces. The faces are
      referencing the boundary points in pTable.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void BoundFaces (MeshGenData *mgds)
{
  register int j;
  register int i;
  register int n, v;
  int facesCounter=0; 

  if(vm_verbose_level >= 3)
  {
    printf("+ debug 3 -- BoundFaces\n");
  }
  
  /* for each input mesh */
  for (i = 0; i < mgds->NMeshes; i++)
  {
    mgds->fTableLength[i] = (mgds->mgCtrl->inputMeshes[i])->NElemsTriang1;
    /* allocate memory for boundary point table */
    mgds->fTable[i] = (Face **)e_calloc(mgds->fTableLength[i], sizeof(Face));

    /* fill table of boundary faces */
    for (j = 0; j < mgds->fTableLength[i]; j++)
    {
      /* create face */
      mgds->fTable[i][j] = FaceCreate();
      mgds->fTable[i][j]->mesh = i;
      
      facesCounter++;
      /* set faces vertices */
      for (v=0; v<3; v++)
      {
        n = (mgds->mgCtrl->inputMeshes[i])->NodesTriang1[j][v] - 1;
        FACE_VERTEX (mgds->fTable[i][j], v) = mgds->pTable[i][n];
      }
    }
  }

  if(vm_verbose_level >= 4)
  {
    printf("+ debug 4 -- \t Total number of input faces: %d\n",
           facesCounter);
  }
  
} /* end of BoundFaces */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 BoundEdges --
        Retrieve boundary edges form boundary surface mesh.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void BoundEdges (MeshGenData *mgds)
{
  int   edgesCounter=0; 
  int   elem, side, neighbour, i;
  int   neighbSide, sanityCheck;
  Edge *edge;
  Face *face;
  int **adjElem = NULL;
  int  *nodeElemAdjLen = NULL;
  int **nodeElemAdj = NULL;
  
  if(vm_verbose_level >= 3)
  {
    printf("+ debug 3 -- BoundEdges\n");
  }
  
  /* for each input mesh */
  for (i = 0; i < mgds->NMeshes; i++)
  {
    mgds->eTableLength[i] = 0;
    mgds->eTable[i] = NULL;
    
    /* allocate memory for adjacency matrices */
    nodeElemAdj    = e_init_frag_i_matr (mgds->mgCtrl->inputMeshes[i]->NNodes);
    nodeElemAdjLen = e_alloc_i_vect (mgds->mgCtrl->inputMeshes[i]->NNodes);
  
    /* find nodes to elements adjacency of i-th input mesh */
    e_InitMeshPointElemConnection (mgds->mgCtrl->inputMeshes[i], nodeElemAdjLen,
                                   nodeElemAdj);
    
    /* find general adjacency matrix of surface mesh */
    Triang1Adjacency(mgds->mgCtrl->inputMeshes[i],
                     nodeElemAdjLen, nodeElemAdj, &adjElem);
    
    /* visit all element in i-th input mesh - but only once */
    for (elem=0; elem < mgds->fTableLength[i]; elem++)
    {
      face = mgds->fTable[i][elem];
      
      /* set visited flag in the face */
      face->flag |= FACE_BOUNDARY;
      for (side=0; side<3; side++)
      {
        neighbour = adjElem[elem][side];
        if (   (neighbour == 0)
            || (((mgds->fTable[i][neighbour-1])->flag & FACE_BOUNDARY) !=
                    FACE_BOUNDARY)
           )
        {
          edge = EdgeCreate ();
          edge->mesh = i;
          
          EDGE_VERTEX(edge,0) = FACE_VERTEX(mgds->fTable[i][elem],
                                            sideDef[side][0]);
          EDGE_VERTEX(edge,1) = FACE_VERTEX(mgds->fTable[i][elem],
                                            sideDef[side][1]);

          /* put edge in table of edges (growing table if necessary) */
          mgds->eTable[i] = (Edge **)
              e_incr_vp_vect(mgds->eTableLength[i],
                             (void **)mgds->eTable[i],
                             EDGE_ALLOC_INCR);
          mgds->eTable[i][mgds->eTableLength[i]] = edge;
          mgds->eTableLength[i]++;
          
          edgesCounter++;
          
          
          if(neighbour == 0) /* edge in the triangle side belongs only
                              * to one element - it is free edge
                              */  
          {
            FACE_NEIGHBOUR (mgds->fTable[i][elem], side) = edge;
            EDGE_NEIGHBOUR (edge, 0) = mgds->fTable[i][elem];
            EDGE_NEIGHBOUR (edge, 1) = NULL;
          }
          else
          {
            /* set edge-face adjacency */ 
            EDGE_NEIGHBOUR (edge, 0) = mgds->fTable[i][elem];
            EDGE_NEIGHBOUR (edge, 1) = mgds->fTable[i][neighbour-1];
            /* set face-edge adjacency */
            FACE_NEIGHBOUR (mgds->fTable[i][elem], side) = edge;
            sanityCheck = FaceHasEdge (mgds->fTable[i][neighbour-1], edge, 
                                       &neighbSide);
            assert (sanityCheck == 1);
            FACE_NEIGHBOUR (mgds->fTable[i][neighbour-1], neighbSide) = edge;
          }
        }
      }
    }
    /* free allocated memory */
    e_free_cont_i_matr(mgds->mgCtrl->inputMeshes[i]->NElemsTriang1, adjElem);
    
    e_free ((void *) nodeElemAdjLen); 
    e_free_frag_i_matr ((mgds->mgCtrl->inputMeshes[i])->NNodes, nodeElemAdj); 
  }

  if(vm_verbose_level >= 4)
  {
    printf("+ debug 4 -- \t Total number of input edges: %d\n",
           edgesCounter);
  }
  
} /* end of BoundEdges */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */


void UpdateSplitEdgeToFace(MeshGenData *mgds, Edge *edge)
{
  int     iside, imesh, inode, ene, sidenum, subside, check;
  Face   *face;
  Edge   *edge1, *edge2;
  struct Tos_MStack istack = {0, NULL, NULL};
  Tos_MStack edgeStack = &istack;
  Edge   *splitedge, *midedge, *sideedge;
  Face   *splitface;
  Vertex *v1, *v2;
  
  imesh = edge->mesh;
  assert(imesh >= 0);

  v1 = EDGE_VERTEX(edge, 0);
  v2 = EDGE_VERTEX(edge, 1);
  
  for(iside = 0; iside < 2; iside++)
  {
    face = EDGE_NEIGHBOUR(edge, iside);
    
    if(face != NULL)
    {
      /* check backward reference, also provides side info */
      check = FaceHasEdge(face, edge, &sidenum);
      assert(check == 1);

      /* get other edges of the face */
      edge1 = FACE_NEIGHBOUR(face, (sidenum + 1) % 3);
      edge2 = FACE_NEIGHBOUR(face, (sidenum + 2) % 3);
      assert(edge1 != NULL);
      assert(edge2 != NULL);
      
      /* collect all edges */
      /* add non-split edges */
      Tos_MStackPush(edgeStack, (void *)edge1);
      Tos_MStackPush(edgeStack, (void *)edge2);

      Tos_MStackTraversalInit(edge->splitEdges);
      while((splitedge = (Edge *)Tos_MStackTraverse(edge->splitEdges)) != NULL)
      {
        /* add split edge */
        Tos_MStackPush(edgeStack, (void *)splitedge);

        for(inode = 0; inode < 2; inode++)
        {
          /* if the end point of this split edge and the node opposite the
             split edge is a middle, new edge and not yet encountered then
             add this middle edge to the stack */
          if(   EDGE_VERTEX(splitedge, inode) != v1
             && EDGE_VERTEX(splitedge, inode) != v2
             && IsEdgeOnStack(edgeStack,
                              FACE_VERTEX(face, sidenum),
                              EDGE_VERTEX(splitedge, inode)) == 0)
          {
            /* add middle edge to table of boundary edges */
            midedge = EdgeCreate();
            assert(imesh >= 0);
            midedge->mesh = imesh;
            EDGE_VERTEX(midedge, 0) = FACE_VERTEX(face, sidenum);
            EDGE_VERTEX(midedge, 1) = EDGE_VERTEX(splitedge, inode);

            /* middle edges are added to the list of edges properly */
            mgds->eTable[imesh] = (Edge **)
                e_incr_vp_vect(mgds->eTableLength[imesh],
                               (void **)mgds->eTable[imesh],
                               EDGE_ALLOC_INCR);
            mgds->eTable[imesh][mgds->eTableLength[imesh]] = midedge;
            mgds->eTableLength[imesh]++;
          
            Tos_MStackPush(edgeStack, (void *)midedge);
          }
        }
      }
      
      /* reset adjacency information */
      FACE_NEIGHBOUR(face, 0) = NULL;
      FACE_NEIGHBOUR(face, 1) = NULL;
      FACE_NEIGHBOUR(face, 2) = NULL;
      if(EDGE_NEIGHBOUR(edge1, 0) == face)
          EDGE_NEIGHBOUR(edge1, 0) = NULL;
      else if(EDGE_NEIGHBOUR(edge1, 1) == face)
          EDGE_NEIGHBOUR(edge1, 1) = NULL;
      else
          e_GeneralError("no reverse 1");
      if(EDGE_NEIGHBOUR(edge2, 0) == face)
          EDGE_NEIGHBOUR(edge2, 0) = NULL;
      else if(EDGE_NEIGHBOUR(edge2, 1) == face)
          EDGE_NEIGHBOUR(edge2, 1) = NULL;
      else
          e_GeneralError("no reverse 2");
      
      /* go through all split edges */
      Tos_MStackTraversalInit(edge->splitEdges);
      while((splitedge = (Edge *)Tos_MStackTraverse(edge->splitEdges)) != NULL)
      {
        /* create new subface */
        splitface = FaceCreate();
        assert(imesh >= 0);
        splitface->mesh = imesh;
        Tos_MStackPush(face->splitFaces, (void *)splitface);
        FACE_VERTEX(splitface, 0) = FACE_VERTEX(face, sidenum);
        FACE_VERTEX(splitface, 1) = EDGE_VERTEX(splitedge, 0);
        FACE_VERTEX(splitface, 2) = EDGE_VERTEX(splitedge, 1);
        
        /* set adjacency information from edge to face and face to edge */
        Tos_MStackTraversalInit(edgeStack);
        while((sideedge = (Edge *)Tos_MStackTraverse(edgeStack)) != NULL)
        {
          if(FaceHasEdge(splitface, sideedge, &subside))
          {
            assert(FACE_NEIGHBOUR(splitface, subside) == NULL);
            FACE_NEIGHBOUR(splitface, subside) = sideedge;
            
            ene = 0;
            if(EDGE_NEIGHBOUR(sideedge, 0) != NULL) ene++;
            assert(EDGE_NEIGHBOUR(sideedge, ene) == NULL);
            
            EDGE_NEIGHBOUR(sideedge, ene) = splitface;
          }
        }
      }
      
      /* empty the collected edges */
      Tos_MStackFlush(edgeStack);
      
    } /* if(face != NULL) */
  }

}



void UpdateSplitFaceToEdge(MeshGenData *mgds, Face *face)
{
  Edge   *edge1, *edge2, *edge3;
  Face   *splitface;
  int     iside, imesh, ene, subside;
  Vertex *v1, *v2;
  struct Tos_MStack istack = {0, NULL, NULL};
  Tos_MStack edgeStack = &istack;
  Edge   *midedge, *sideedge;

  imesh = face->mesh;
  assert(imesh >= 0);
  
  edge1 = FACE_NEIGHBOUR(face, 0);
  edge2 = FACE_NEIGHBOUR(face, 1);
  edge3 = FACE_NEIGHBOUR(face, 2);
  assert(edge1 != NULL);
  assert(edge2 != NULL);
  assert(edge3 != NULL);
  
  /* add all side edges */
  Tos_MStackPush(edgeStack, (void *)edge1);
  Tos_MStackPush(edgeStack, (void *)edge2);
  Tos_MStackPush(edgeStack, (void *)edge3);

  /* add all middle edges */
  Tos_MStackTraversalInit(face->splitFaces);
  while((splitface = (Face *)Tos_MStackTraverse(face->splitFaces)) != NULL)
  {
    for(iside = 0; iside < 3; iside++)
    {
      v1 = FACE_VERTEX(splitface, sideDef[iside][0]);
      v2 = FACE_VERTEX(splitface, sideDef[iside][1]);
      if(IsEdgeOnStack(edgeStack, v1, v2) == 0)
      {
        midedge = EdgeCreate();
        assert(imesh >= 0);
        midedge->mesh = imesh;
        EDGE_VERTEX(midedge, 0) = v1;
        EDGE_VERTEX(midedge, 1) = v2;
        mgds->eTable[imesh] = (Edge **)
            e_incr_vp_vect(mgds->eTableLength[imesh],
                           (void **)mgds->eTable[imesh],
                           EDGE_ALLOC_INCR);
        mgds->eTable[imesh][mgds->eTableLength[imesh]] = midedge;
        mgds->eTableLength[imesh]++;
          
        Tos_MStackPush(edgeStack, (void *)midedge);
      }
    }
  }

  /* reset adjacency information */
  FACE_NEIGHBOUR(face, 0) = NULL;
  FACE_NEIGHBOUR(face, 1) = NULL;
  FACE_NEIGHBOUR(face, 2) = NULL;
  if(EDGE_NEIGHBOUR(edge1, 0) == face)
      EDGE_NEIGHBOUR(edge1, 0) = NULL;
  else if(EDGE_NEIGHBOUR(edge1, 1) == face)
      EDGE_NEIGHBOUR(edge1, 1) = NULL;
  else
      e_GeneralError("no reverse 1");
  if(EDGE_NEIGHBOUR(edge2, 0) == face)
      EDGE_NEIGHBOUR(edge2, 0) = NULL;
  else if(EDGE_NEIGHBOUR(edge2, 1) == face)
      EDGE_NEIGHBOUR(edge2, 1) = NULL;
  else
      e_GeneralError("no reverse 2");
  if(EDGE_NEIGHBOUR(edge3, 0) == face)
      EDGE_NEIGHBOUR(edge3, 0) = NULL;
  else if(EDGE_NEIGHBOUR(edge3, 1) == face)
      EDGE_NEIGHBOUR(edge3, 1) = NULL;
  else
      e_GeneralError("no reverse 1");

  /* set face-edge relation on split faces */
  Tos_MStackTraversalInit(face->splitFaces);
  while((splitface = (Face *)Tos_MStackTraverse(face->splitFaces)) != NULL)
  {
    Tos_MStackTraversalInit(edgeStack);
    while((sideedge = (Edge *)Tos_MStackTraverse(edgeStack)) != NULL)
    {
      if(FaceHasEdge(splitface, sideedge, &subside))
      {
        assert(FACE_NEIGHBOUR(splitface, subside) == NULL);
        FACE_NEIGHBOUR(splitface, subside) = sideedge;
      
        ene = 0;
        if(EDGE_NEIGHBOUR(sideedge, 0) != NULL) ene++;
        assert(EDGE_NEIGHBOUR(sideedge, ene) == NULL);
        EDGE_NEIGHBOUR(sideedge, ene) = splitface;
      }
    }
  }
  
  Tos_MStackFlush(edgeStack);
}


void SplitBoundaryEdge(MeshGenData *mgds, Edge *edge, Vertex *p)
{
  Edge *splitedge;
  
  /* mark the point as boundary and external or internal boundary point */
  p->flag |= POINT_BOUNDARY;
  assert(   EDGE_VERTEX(edge, 0)->flag & POINT_EXT_BOUNDARY
         || EDGE_VERTEX(edge, 0)->flag & POINT_INT_BOUNDARY);
  if(EDGE_VERTEX(edge, 0)->flag & POINT_EXT_BOUNDARY)
  {
    p->flag |= POINT_EXT_BOUNDARY;
  }
  else if(EDGE_VERTEX(edge, 0)->flag & POINT_INT_BOUNDARY)
  {
    p->flag |= POINT_INT_BOUNDARY;
  }
  /* FIXME: the point should be added to the list of boundary points */

  assert(Tos_MStackCount(edge->splitEdges) == 0);

  /* create first edge */
  splitedge = EdgeCreate();
  assert(edge->mesh >= 0);
  splitedge->mesh = edge->mesh;
  EDGE_VERTEX(splitedge, 0) = EDGE_VERTEX(edge, 0);
  EDGE_VERTEX(splitedge, 1) = p;
  /* store it as split edge */
  Tos_MStackPush(edge->splitEdges, (void *)splitedge);

  /* create second edge */
  splitedge = EdgeCreate();
  assert(edge->mesh >= 0);
  splitedge->mesh = edge->mesh;
  EDGE_VERTEX(splitedge, 1) = EDGE_VERTEX(edge, 1);
  EDGE_VERTEX(splitedge, 0) = p;
  /* store it as split edge */
  Tos_MStackPush(edge->splitEdges, (void *)splitedge);

  /* update edge-face adjacency */
  UpdateSplitEdgeToFace(mgds, edge);
}



void SplitBoundaryFace(MeshGenData *mgds, Face *face, Vertex *p)
{
  int   iside;
  Face *splitface;
  
  /* mark the point as boundary and external or internal boundary point */
  p->flag |= POINT_BOUNDARY;
  assert(   FACE_VERTEX(face, 0)->flag & POINT_EXT_BOUNDARY
         || FACE_VERTEX(face, 0)->flag & POINT_INT_BOUNDARY);
  if(FACE_VERTEX(face, 0)->flag & POINT_EXT_BOUNDARY)
  {
    p->flag |= POINT_EXT_BOUNDARY;
  }
  else if(FACE_VERTEX(face, 0)->flag & POINT_INT_BOUNDARY)
  {
    p->flag |= POINT_INT_BOUNDARY;
  }
  /* FIXME: the point should be added to the list of boundary points */

  assert(Tos_MStackCount(face->splitFaces) == 0);

  for(iside = 0; iside < 3; iside++)
  {
    /* create face */
    splitface = FaceCreate();
    assert(face->mesh >= 0);
    splitface->mesh = face->mesh;
    FACE_VERTEX(splitface, 0) = FACE_VERTEX(face, sideDef[iside][0]);
    FACE_VERTEX(splitface, 1) = FACE_VERTEX(face, sideDef[iside][1]);
    FACE_VERTEX(splitface, 2) = p;
    /* store it as split face */
    Tos_MStackPush(face->splitFaces, (void *)splitface);
  }
  
  /* update edge-face adjacency */
  UpdateSplitFaceToEdge(mgds, face);
}


Edge *InitSplitEdge(Edge *edge, Tos_MStack edgeIter)
{
  Edge *splitedge;
  
  Tos_MStackFlush(edgeIter);
  
  /* check, whether edge is split */
  if(Tos_MStackCount(edge->splitEdges) > 0)
  {
    Tos_MStackTraversalInit(edge->splitEdges);
    Tos_MStackPush(edgeIter, (void *)(edge->splitEdges));
    splitedge = NextSplitEdge(edgeIter);
    assert(splitedge != NULL);
    return(splitedge);
  }
  else
  {
    return(edge);
  }
}

Edge *NextSplitEdge(Tos_MStack edgeIter)
{
  Edge *splitedge;
  Tos_MStack splitStack;
  
  while(Tos_MStackCount(edgeIter) > 0)
  {
    splitStack = (Tos_MStack)Tos_MStackPop(edgeIter);
    
    splitedge = (Edge *)Tos_MStackTraverse(splitStack);
    if(splitedge != NULL)
    {
      Tos_MStackPush(edgeIter, (void *)splitStack);
      if(Tos_MStackCount(splitedge->splitEdges) > 0)
      {
        Tos_MStackTraversalInit(splitedge->splitEdges);
        Tos_MStackPush(edgeIter, (void *)splitedge->splitEdges);
      }
      else
      {
        return(splitedge);
      }
    }
  }
  
  return NULL;
}




Face *InitSplitFace(Face *face, Tos_MStack faceIter)
{
  Face *splitface;

  Tos_MStackFlush(faceIter);
  
  /* check, whether face is split */
  if(Tos_MStackCount(face->splitFaces) > 0)
  {
    Tos_MStackTraversalInit(face->splitFaces);
    Tos_MStackPush(faceIter, (void *)(face->splitFaces));
    splitface = NextSplitFace(faceIter);
    assert(splitface != NULL);
    return(splitface);
  }
  else
  {
    return(face);
  }
}


Face *NextSplitFace(Tos_MStack faceIter)
{
  Face *splitface;
  Tos_MStack splitStack;
  
  while(Tos_MStackCount(faceIter) > 0)
  {
    splitStack = (Tos_MStack)Tos_MStackPop(faceIter);
    
    splitface = (Face *)Tos_MStackTraverse(splitStack);
    if(splitface != NULL)
    {
      Tos_MStackPush(faceIter, (void *)splitStack);
      if(Tos_MStackCount(splitface->splitFaces) > 0)
      {
        Tos_MStackTraversalInit(splitface->splitFaces);
        Tos_MStackPush(faceIter, (void *)splitface->splitFaces);
      }
      else
      {
        return(splitface);
      }
    }
  }
  
  return NULL;
}


void PrintEdgeMesh(MeshGenData *data)
{
  int imesh, iedge;
  Edge *edge, *splitedge;
  struct Tos_MStack edgeIter = {0, NULL, NULL};
  
  for(imesh = 0; imesh < data->NMeshes; imesh++)
  {
    for(iedge = 0; iedge < data->eTableLength[imesh]; iedge++)
    {
      edge = data->eTable[imesh][iedge];
      for(splitedge = InitSplitEdge(edge, &edgeIter);
          splitedge != NULL;
          splitedge = NextSplitEdge(&edgeIter))
      {
        if(splitedge->mesh < 0)
        {
          printf("iedge: %d  mesh: %d\n", iedge, splitedge->mesh);
        }
      }
    }
  }
}

void PrintFaceMesh(MeshGenData *data)
{
  int imesh, iface;
  Face *face, *splitface;
  struct Tos_MStack faceIter = {0, NULL, NULL};
  
  for(imesh = 0; imesh < data->NMeshes; imesh++)
  {
    for(iface = 0; iface < data->fTableLength[imesh]; iface++)
    {
      face = data->fTable[imesh][iface];
      for(splitface = InitSplitFace(face, &faceIter);
          splitface != NULL;
          splitface = NextSplitFace(&faceIter))
      {
        if(splitface->mesh < 0)
        {
          printf("iface: %d  mesh: %d\n", iface, splitface->mesh);
        }
      }
    }
  }
}



