#include <assert.h>
#include <math.h>
#include "e_lib.h"
#include "eg_lib.h"
#include "basic_op.h"
#include "measures.h"


const char *tetTypeNames[] = {"Round", "Needle", "Wedge", "Spindle", "Sliver",
                             "Cap", "Unknown"};

void GetTetEdgesLengthTable (double *coords[4], double *ltable[4])
{
  int i, j;
  for (i=0; i<4; i++)
  {
    for (j=i+1; j<4; j++)
    {
      ltable[i][j] = eg_PointDist3D(coords[i], coords[j]);
      ltable[j][i] = ltable[i][j];
    }
    ltable[i][i] = 0.0;
  }
} /* end of GetTetEdgesLengthTable */

double SinHalfSolidAngle (int vertexIndex, double volume, double *ltable[4])
{
  static int ij_tab[4][3] = {{1,2,3}, {0,2,3}, {0,1,3}, {0,1,2}};
  int i,j,k,l;
  double p,a,result; 
   
  p = 1.0;
  for (k=0; k<3; k++)
  {
    i = ij_tab[vertexIndex][k];
    for (l=k+1; l<3; l++)
    {
      j = ij_tab[vertexIndex][l];
      a = ltable[vertexIndex][i] + ltable[vertexIndex][j];
      p = p * (a + ltable[i][j]) * (a - ltable[i][j]); 
    }
  }
  result = 12.0 * volume / sqrt (p);  
  return result;
} /* end of SinHalfSolidAngle */

double RadiusRatio (double *vcoords[4], double *ltable[4])
{
  static int ij_tab[4][3] = {{1,2,3}, {0,2,3}, {0,1,3}, {0,1,2}};  
  double areaSum = 0;
  double volume;
  double a, b, c;
  double factor, rho;
  int i;

  volume = eg_TetrahVolume (vcoords[0], vcoords[1], vcoords[2], vcoords[3]);

  /* sum of area of tetrahedron faces */
  for (i=0; i<4; i++)
  {
     areaSum = areaSum + eg_TriangArea3D (vcoords[ij_tab[i][0]],
                                       vcoords[ij_tab[i][1]],
                                       vcoords[ij_tab[i][2]]);
  }

  /* products of lengths of oposite edges */
  a = ltable[0][3]*ltable[1][2];
  b = ltable[0][1]*ltable[3][2];
  c = ltable[1][3]*ltable[0][2];

  factor = (a+b+c)*(a+b-c)*(a+c-b)*(b+c-a);
  rho = 216.0 * volume * volume / (areaSum * sqrt (factor));

  return rho; 
} /* end of RadiusRatio */

double MeanRatio(double volume, double *ltable[4])
{
  double edgeLenSum;
  double a, result;
  int i,j;

  edgeLenSum = 0.0;
  
  for (i=0; i<4; i++)
  {
    for (j=i+1; j<4; j++)
    {
      edgeLenSum = edgeLenSum + ltable[i][j] * ltable[i][j];
    }
  }
  a = pow (3.0*volume, 2.0/3.0);
  result = 12.0 * a / edgeLenSum;
  return result;
} 

double Tetrah_MinSigmaMeasure (Tetrah *tet, int *index)
{
   double *ltable[4];
   double sigma_min, sigma;
   double volume;
   double *vcoords[4];
   int i;
   double data[4][4];

   ltable[0] = &(data[0][0]);
   ltable[1] = &(data[1][0]);
   ltable[2] = &(data[2][0]);
   ltable[3] = &(data[3][0]);
 
   vcoords[0] = TET_VCOORDS(tet,0);
   vcoords[1] = TET_VCOORDS(tet,1);
   vcoords[2] = TET_VCOORDS(tet,2);
   vcoords[3] = TET_VCOORDS(tet,3);
   volume = eg_TetrahVolume (vcoords[0], vcoords[1], vcoords[2], vcoords[3]);
   GetTetEdgesLengthTable (vcoords, ltable);

   sigma_min = 2.0;
   for (i=0; i<4; i++)
   {
     sigma = SinHalfSolidAngle (i, volume, ltable);
     if (sigma_min > sigma)
     {
        sigma_min = sigma; 
        *index = i;
     }
   }
   /* normalize the sigma value to range [0,1]  value 1 for regular
      tetrahedron */
   sigma_min = sigma_min / 0.2721655; /* sqrt(6)/9 */
   return sigma_min;
} /* end of Tetrah_MinSigmaMeasure */

double Tetrah_RadiusRatio (Tetrah *tet)
{
   double *ltable[4];
   double *vcoords[4];
   double data[4][4];

   ltable[0] = &(data[0][0]);
   ltable[1] = &(data[1][0]);
   ltable[2] = &(data[2][0]);
   ltable[3] = &(data[3][0]);
 
   vcoords[0] = TET_VCOORDS(tet,0);
   vcoords[1] = TET_VCOORDS(tet,1);
   vcoords[2] = TET_VCOORDS(tet,2);
   vcoords[3] = TET_VCOORDS(tet,3);
   GetTetEdgesLengthTable (vcoords, ltable);
   return RadiusRatio (vcoords, ltable);
} /* end of Tetrah_RadiusRatio */

double Tetrah_MeanRatio (Tetrah *tet)
{
   double *ltable[4];
   double volume;
   double *vcoords[4];
   double data[4][4];

   ltable[0] = &(data[0][0]);
   ltable[1] = &(data[1][0]);
   ltable[2] = &(data[2][0]);
   ltable[3] = &(data[3][0]);
  
   vcoords[0] = TET_VCOORDS(tet,0);
   vcoords[1] = TET_VCOORDS(tet,1);
   vcoords[2] = TET_VCOORDS(tet,2);
   vcoords[3] = TET_VCOORDS(tet,3);
   volume = eg_TetrahVolume (vcoords[0], vcoords[1], vcoords[2], vcoords[3]);
   GetTetEdgesLengthTable (vcoords, ltable);

   return MeanRatio (volume, ltable);
} /* end of Tetrah_MeanRatio */

double Tetrah_SolidAngle (Tetrah *tet, int i)
{
  double *ltable[4];
  double sigma, theta;
  double volume, denom;
  double *vcoords[4];
  int j,k; 
  double data[4][4];
  double *v[3];
  double vvec[3][3];
  double vl[3];


  assert (i>=0 && i<4);
  assert (tet != NULL);

  ltable[0] = &(data[0][0]);
  ltable[1] = &(data[1][0]);
  ltable[2] = &(data[2][0]);
  ltable[3] = &(data[3][0]);
  v[0] = &(vvec[0][0]);
  v[1] = &(vvec[1][0]);
  v[2] = &(vvec[2][0]);
 
  vcoords[0] = TET_VCOORDS(tet,0);
  vcoords[1] = TET_VCOORDS(tet,1);
  vcoords[2] = TET_VCOORDS(tet,2);
  vcoords[3] = TET_VCOORDS(tet,3);
  volume = eg_TetrahVolume (vcoords[0], vcoords[1], vcoords[2], vcoords[3]);
  GetTetEdgesLengthTable (vcoords, ltable);

  assert (i>=0 && i<4);

  k=0;
  for (j=0; j<4; j++)
  {
    if (j == i) continue;
   
    v[k][0] = vcoords[j][0] - vcoords[i][0];
    v[k][1] = vcoords[j][1] - vcoords[i][1];
    v[k][2] = vcoords[j][2] - vcoords[i][2];
    vl[k] = data[i][j];
    k++;
  } 

  denom = vl[0]*vl[1]*vl[2];
  denom = denom + vl[0] * e_VectDotProd(3, v[1], v[2]);
  denom = denom + vl[1] * e_VectDotProd(3, v[0], v[2]);
  denom = denom + vl[2] * e_VectDotProd(3, v[0], v[1]);
 
  sigma = SinHalfSolidAngle (i, volume, ltable);
 
  if (denom < 0)
  {
     theta = 2*(M_PI - asin(sigma));
  }
  else
  {
     theta = 2*asin(sigma);
  }
  return theta;
} /* end of Tetrah_SolidAngle */

void Tetrah_MinAndMaxSolidAngle (Tetrah *tet,
                                 double *minsangle, int *vminind, 
                                 double *maxsangle, int *vmaxind)
{
  double *ltable[4];
  double sigma;
  double denom, theta;
  double volume;
  double *vcoords[4];
  int i,j,k; 
  double data[4][4];
  double *v[3];
  double vvec[3][3];
  double vl[3];

  ltable[0] = &(data[0][0]);
  ltable[1] = &(data[1][0]);
  ltable[2] = &(data[2][0]);
  ltable[3] = &(data[3][0]);
  v[0] = &(vvec[0][0]);
  v[1] = &(vvec[1][0]);
  v[2] = &(vvec[2][0]);
 
  vcoords[0] = TET_VCOORDS(tet,0);
  vcoords[1] = TET_VCOORDS(tet,1);
  vcoords[2] = TET_VCOORDS(tet,2);
  vcoords[3] = TET_VCOORDS(tet,3);
  volume = eg_TetrahVolume (vcoords[0], vcoords[1], vcoords[2], vcoords[3]);
  GetTetEdgesLengthTable (vcoords, ltable);

  *minsangle = 2*M_PI+1.0;
  *maxsangle = -1.0;
  *vminind = -9999;  /* dummy indices */
  *vmaxind = -9999;

  for (i=0; i<4; i++)
  {
     k=0;
     for (j=0; j<4; j++)
     {
       if (j == i) continue;
    
       v[k][0] = vcoords[j][0] - vcoords[i][0];
       v[k][1] = vcoords[j][1] - vcoords[i][1];
       v[k][2] = vcoords[j][2] - vcoords[i][2];
       vl[k] = data[i][j];
       k++;
     } 

    denom = vl[0]*vl[1]*vl[2];
    denom = denom + vl[0] * e_VectDotProd(3, v[1], v[2]);
    denom = denom + vl[1] * e_VectDotProd(3, v[0], v[2]);
    denom = denom + vl[2] * e_VectDotProd(3, v[0], v[1]);
 
    sigma = SinHalfSolidAngle (i, volume, ltable);
 
    if (denom < 0)
    {
       theta = 2*(M_PI - asin(sigma));
    }
    else 
    {  /* this is explanation */
       theta = 2*asin(sigma);
    }
    if (*minsangle > theta) 
    {
      *minsangle = theta;
      *vminind = i;
    }
    if (*maxsangle < theta)
    {
      *maxsangle = theta;
      *vmaxind = i;
    }
  }
} /* end of Tetrah_MinAndMaxSolidAngle */


double Tetrah_DihedralAngle (Tetrah *tet, int i)
{
  double data[2][3];
  double *n[2];
  double dangle, dotProd;
  int f1, f2, ev1, ev2;
  int ierr;

  assert (i>=0 && i<6);
  assert (tet != NULL);
 
  /* set pointers to static storage */
  n[0] = &data[0][0];
  n[1] = &data[1][0];
  
  /* find vertices of the edge */
  ev1 = edgeDef[i][0];
  ev2 = edgeDef[i][1];

  /* find indiced of faces contianing edge ev1-ev2 */
  TetGenericMapping (ev1, ev2, &f1, &f2);


  /* calculate normals of the faces. According to face specification
     normals always point outside tetrahedron
   */
  ierr = eg_TriangNormal3D (TET_VCOORDS(tet, faceDef[f1][0]),
                            TET_VCOORDS(tet, faceDef[f1][1]), 
                            TET_VCOORDS(tet, faceDef[f1][2]), n[0]);
  if (ierr != EG_SUCCESS)
  {
    e_GeneralError ("Tetrah_DihedranAngle : %s",
                    "two or more nodes of tetrahedron collide"); 
  }
  ierr = eg_TriangNormal3D (TET_VCOORDS(tet, faceDef[f2][0]),
                            TET_VCOORDS(tet, faceDef[f2][1]), 
                            TET_VCOORDS(tet, faceDef[f2][2]), n[1]);
  if (ierr != EG_SUCCESS)
  {
    e_GeneralError ("Tetrah_DihedranAngle : %s",
                    "two or more nodes of tetrahedron collide"); 
  }

  /* calculate dot product of normals */
  dotProd = n[0][0] * n[1][0] + n[0][1]*n[1][1] + n[0][2]*n[1][2]; 

  /* calculate the dihedral angle - you have to take the dot product
     with oposite sign to calculate the actual dihedral angle.
   */
  dangle = acos(-dotProd);

  return dangle;
} /* end of Tetrah_DihedralAngle */


void Tetrah_MinAndMaxDihedralAngle (Tetrah *tet,
                                    double *mindangle, int *eminind,
                                    double* maxdangle, int *emaxind)
{
  double data[4][3];
  double *n[4];
  double dangle, dotProd;
  int f1, f2, ev1, ev2;
  int i, ierr;

  assert (tet != NULL);
 
  /* set pointers to static storage */
  n[0] = &data[0][0];
  n[1] = &data[1][0];
  n[2] = &data[2][0];
  n[3] = &data[3][0];
  
  /* calculat normals to faces */
  for (i=0; i<4; i++)
  {
    ierr = eg_TriangNormal3D (TET_VCOORDS(tet, faceDef[i][0]),
                              TET_VCOORDS(tet, faceDef[i][1]), 
                              TET_VCOORDS(tet, faceDef[i][2]), n[i]);
    if (ierr != EG_SUCCESS)
    {
      e_GeneralError ("Tetrah_MinAndMaxDihedranAngle : %s",
                      "two or more nodes of tetrahedron collide"); 
    }
  }

  /* remark - normals are normalized */
  *mindangle = M_PI+1.0;
  *maxdangle = -1.0;
 
  /* go through all edges of tetrahedron */
  for (i=0; i<6; i++)
  {
    /* get edge vertices */
    ev1 = edgeDef[i][0];
    ev2 = edgeDef[i][1];

    /* find indices of faces contianing edge ev1-ev2 */
    TetGenericMapping (ev1, ev2, &f1, &f2);

    /* calculate dot product of normals */
    dotProd = n[f1][0] * n[f2][0] + n[f1][1]*n[f2][1] + n[f1][2]*n[f2][2]; 

    /* calculate the dihedral angle - you have to take the dot product
       with oposite sign to calculate the actual dihedral angle.
    */
    dangle = acos(-dotProd);
    if (dangle < *mindangle)
    {
      *mindangle = dangle;
      *eminind = i;     
    }
    if (dangle > *maxdangle)
    {
      *maxdangle = dangle;
      *emaxind = i;
    }
  }
} /* end of Tetrah_MinAndMaxDihedralAngle */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   Mesh_MinAndMaxMeasure
 * DESCRIPTION
 *   Calculates minimal and maximal measure of elements in the mesh .
 * ARGUMENTS
 *   tetMesh - pointer to mesh structure (IN)
 *   measure - type of the mesure 
 *   minMeasure - mimimal measure 
 *   maxMeasure - maximal measure 
 * RETURN VALUE
 *   NONE
 ******/
void Mesh_MinAndMaxMeasure (TetrahMesh *tetMesh, enum MeasureType measure, 
                            double *minMeasure, double* maxMeasure)
{
  Tetrah *tet;
  double currentMeasure;

  *minMeasure = 2.0;
  *maxMeasure = -1.0;
  tet = LIST_TAIL(tetMesh->tList);
  while (tet != NULL)
  {
    currentMeasure = Tetrah_Measure (tet, measure);
    if (*minMeasure > currentMeasure)
        *minMeasure = currentMeasure;
    if (*maxMeasure < currentMeasure)
        *maxMeasure = currentMeasure;
    tet = TET_NEXT(tet);
  }   
} /* end of Mesh_MinAndMaxMeasure */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   Mesh_MinAndMaxSolidAngle
 * DESCRIPTION
 *   Search for maximum and minimum solid angle in the mesh.
 * ARGUMENTS
 *   tetMesh - pointer to mesh structure (IN)
 *   tet_min - tetrahedron with min solid angle (OUT)
 *   sa_min  - minimal solid angle (OUT)
 *   index_min - index of the min solid angle in tetrahedron (OUT)
 *   tet_max - tetrahedron with max solid angle (OUT)
 *   sa_max  - maximal solid angle (OUT)
 *   index_max - index of the max solid angle in tetrahedron (OUT)
 * RETURN VALUE
 *   NONE
 ******/
void Mesh_MinAndMaxSolidAngle (TetrahMesh *tetMesh, 
                              Tetrah **tet_min, double *sa_min, int *index_min,
                              Tetrah **tet_max, double *sa_max, int *index_max)
{
  Tetrah *t;
  double maxSolidAngle, minSolidAngle;
  int i_min, i_max;

  *sa_max=-1.0, 
  *sa_min=2*M_PI;
  *tet_min = NULL;
  *tet_max = NULL;
  *index_min = -9999;
  *index_max = -9999;

  t = LIST_TAIL(tetMesh->tList);
  while (t != NULL)
  {
    Tetrah_MinAndMaxSolidAngle(t, &minSolidAngle, &i_min, &maxSolidAngle, &i_max);
    if (*sa_min > minSolidAngle)
    {
      *tet_min = t;
      *sa_min = minSolidAngle;
      *index_min = i_min;
    }
    if (*sa_max < maxSolidAngle)
    {
      *tet_max = t;
      *sa_max = maxSolidAngle;
      *index_max = i_max;
    }
    t = TET_NEXT(t);
  }   
} /* end of Mesh_MinAndMaxSolidAngle */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   Mesh_MinAndMaxDihedralAngle
 * DESCRIPTION
 *   Search for maximum and minimum dihedral angle in mesh in a mesh.
 * ARGUMENTS
 *   tetMesh - pointer to mesh structure (IN)
 *   mindangle - minimum dihedral angle (OUT)
 *   maxdangle - maximum dihedral angle (OUT)
 * RETURN VALUE
 *   NONE
 ******/
void Mesh_MinAndMaxDihedralAngle (TetrahMesh *tetMesh,
                                 double *mindangle, double *maxdangle)
{
  Tetrah *t;
  double danglemin, danglemax;
  int dummy1, dummy2;
  
  *maxdangle = 0.0 -1.0;
  *mindangle = M_PI + 1.0;

  t = LIST_TAIL(tetMesh->tList);
  while (t != NULL)
  {
    Tetrah_MinAndMaxDihedralAngle(t, &danglemin, &dummy1, &danglemax, &dummy2);
    if (danglemin < *mindangle)
    {
      *mindangle = danglemin;
    }
    if (danglemax > *maxdangle)
    {
      *maxdangle = danglemax; 
    }
    t = TET_NEXT(t);
  }
} /* end of Mesh_MinAndMaxDihedralAngle */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   IsRound -- macro
 * DESCRIPTION
 *   Checks if the tetrahedron clasifies as a round.
 *   A round tetrahedron has no bad angles of any type.
 * ARGUMENTS
 *   ts_min - the tetrahedron smallset solid angle 
 *   ts_max - the tetrahedron lagrest solid angle
 *   td_min - the tetrahedron smalest dihedral angle
 *   td_max - the tetrahedron lagrest dihedral angle
 *   ms_min - the mesh small solid angle
 *   ms_max - the mesh lagre solid angle
 *   md_min - the mesh small dihedral angle
 *   md_max - the mesh large dihedral angle
 * RETURN VALUE
 *   1 if tetrahedron is needle otherwise 0.
 ******/
#define IsRound(ts_min, ts_max, td_min, td_max, \
                 ms_min, ms_max, md_min, md_max) \
    (    (ts_min) > (ms_min) /*  does not have small solid angle */ \
      && (ts_max) <  (ms_max) /* does not have large solid */\
      && (td_min) >  (md_min) /* does not have small dihedral */\
      && (td_max) <  (md_max) /* does not have large dihedral */ \
    )


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   IsNeedle -- macro
 * DESCRIPTION
 *   Checks if the tetrahedron clasifies as a needle.
 *   A needle has one small solid angle but no small or large
 *   dihedrals.
 * ARGUMENTS
 *   ts_min - the tetrahedron smallset solid angle 
 *   ts_max - the tetrahedron lagrest solid angle
 *   td_min - the tetrahedron smalest dihedral angle
 *   td_max - the tetrahedron lagrest dihedral angle
 *   ms_min - the mesh small solid angle
 *   ms_max - the mesh lagre solid angle
 *   md_min - the mesh small dihedral angle
 *   md_max - the mesh large dihedral angle
 * RETURN VALUE
 *   1 if tetrahedron is needle otherwise 0.
 ******/
#define IsNeedle(ts_min, ts_max, td_min, td_max, \
                 ms_min, ms_max, md_min, md_max) \
    (    (ts_min) <= (ms_min) /* has small solid angle */ \
      && (ts_max) <  (ms_max) /* does not have large solid */\
      && (td_min) >  (md_min) /* does not have small dihedral */\
      && (td_max) <  (md_max) /* does not have large dihedral */ \
    )

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   IsWedge -- macro
 * DESCRIPTION
 *   Checks if the tetrahedron clasifies as a wedge.
 *   A wedge has small but not large dihedrals and no large angles of any kind.
 * ARGUMENTS
 *   ts_min - the tetrahedron smallset solid angle 
 *   ts_max - the tetrahedron lagrest solid angle
 *   td_min - the tetrahedron smalest dihedral angle
 *   td_max - the tetrahedron lagrest dihedral angle
 *   ms_min - the mesh small solid angle
 *   ms_max - the mesh lagre solid angle
 *   md_min - the mesh small dihedral angle
 *   md_max - the mesh large dihedral angle
 * RETURN VALUE
 *   1 if tetrahedron is wedge otherwise 0.
 ******/
#define IsWedge(ts_min, ts_max, td_min, td_max, \
                 ms_min, ms_max, md_min, md_max) \
    (    (td_min) <= (md_min) /* has small dihedral */ \
      && (td_max) <  (md_max) /* does not have large dihedral */\
      && (ts_max) <  (ms_max) /* does not have large solid */ \
    )

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   IsSpindle -- macro
 * DESCRIPTION
 *   Checks if the tetrahedron clasifies as a spindle.
 *   A spindle has small but not large solid angles and large 
 *   but not small dihedrals.
 * ARGUMENTS
 *   ts_min - the tetrahedron smallset solid angle 
 *   ts_max - the tetrahedron lagrest solid angle
 *   td_min - the tetrahedron smalest dihedral angle
 *   td_max - the tetrahedron lagrest dihedral angle
 *   ms_min - the mesh small solid angle
 *   ms_max - the mesh lagre solid angle
 *   md_min - the mesh small dihedral angle
 *   md_max - the mesh large dihedral angle
 * RETURN VALUE
 *   1 if tetrahedron is spindle otherwise 0.
 ******/
#define IsSpindle(ts_min, ts_max, td_min, td_max, \
                ms_min, ms_max, md_min, md_max) \
    (    (ts_min) <= (ms_min) /* has small solid */ \
      && (ts_max) <  (ms_max) /* does not have large solid */\
      && (td_min) >  (md_min) /* does not have small dihedral */ \
      && (td_max) >= (md_max) /* has large dihedral */ \
    )

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   IsSliver -- macro
 * DESCRIPTION
 *   Checks if the tetrahedron clasifies as a sliver.
 *   A sliver  has small and large dihedrals but no large solid angle. 
 * ARGUMENTS
 *   ts_min - the tetrahedron smallset solid angle 
 *   ts_max - the tetrahedron lagrest solid angle
 *   td_min - the tetrahedron smalest dihedral angle
 *   td_max - the tetrahedron lagrest dihedral angle
 *   ms_min - the mesh small solid angle
 *   ms_max - the mesh lagre solid angle
 *   md_min - the mesh small dihedral angle
 *   md_max - the mesh large dihedral angle
 * RETURN VALUE
 *   1 if tetrahedron is sliver otherwise 0.
 ******/
#define IsSliver(ts_min, ts_max, td_min, td_max, \
                ms_min, ms_max, md_min, md_max) \
    (    (ts_max) <  (ms_max) /* does not have large solid */\
      && (td_min) <= (md_min) /* has  small dihedral */ \
      && (td_max) >= (md_max) /* has  large dihedral */ \
    )

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   IsCap -- macro
 * DESCRIPTION
 *   Checks if the tetrahedron clasifies as a cap.
 *   A cap has a large - nearly flat solid angle. 
 * ARGUMENTS
 *   ts_min - the tetrahedron smallset solid angle 
 *   ts_max - the tetrahedron lagrest solid angle
 *   td_min - the tetrahedron smalest dihedral angle
 *   td_max - the tetrahedron lagrest dihedral angle
 *   ms_min - the mesh small solid angle
 *   ms_max - the mesh lagre solid angle
 *   md_min - the mesh small dihedral angle
 *   md_max - the mesh large dihedral angle
 * RETURN VALUE
 *   1 if tetrahedron is cap otherwise 0.
 ******/
#define IsCap(ts_min, ts_max, td_min, td_max, \
                ms_min, ms_max, md_min, md_max) \
    (  (ts_max) >= (ms_max) /* does have large solid */\
    )


/****f*
 * NAME
 *   GoodAnglesLimits -- function 
 * DESCRIPTION
 *   Calculates lower and uper bound on the solid and dihedral angles in the
 *   mesh. Angles which are smaller or equal than the mesh small angle or
 *   grater or equal than mesh large angle are considered to be the bad angles.
 *
 *   The mesh small dihedral angle is defined as a minimum of 30 degrees and
 *   the minimum dihedral angle in mesh plus 10 degrees.
 * 
 *   The mesh large dihedral angle is defined as maximum of 150 degrees and
 *   maximum dihedral angle in the mesh minus 20 degrees.
 *
 *   The mesh small solid angle is defined as minimum of 5 degrees and the minimum
 *   solid angle in the mesh plus 5 degrees.
 * 
 *   The mesh large solid angle is defined as maximum of 240 degrees and
 *   maximum solid angle in the mesh minus 60 degrees.
 *
 * ARGUMENTS
 *   tetMesh - tetrahedral mesh (IN)
 *   ms_min - the mesh small solid angle (OUT)
 *   ms_max - the mesh lagre solid angle (OUT)
 *   md_min - the mesh small dihedral angle (OUT)
 *   md_max - thAnglee mesh large dihedral angle (OUT)
 * RETURN VALUE
 *   NONE
 ******/
void GoodAnglesLimits (TetrahMesh *tetMesh, 
                  double *ms_min, double *ms_max,
                  double *md_min, double *md_max)
{
  int dummy_index;
  double meshMinSolidAngle, meshMaxSolidAngle;
  double meshMinDihAngle, meshMaxDihAngle;
  double minAngle, maxAngle;
  Tetrah *t;
  double deg2rad = M_PI/180.0;
  double lim1, lim2;
  meshMaxDihAngle = -1.0;
  meshMinDihAngle = M_PI + 1.0;
  meshMaxSolidAngle =-1.0; 
  meshMinSolidAngle = 2*M_PI;

  /* calculate mesh maximal and minimal dihedral and solid angle */
  t = LIST_TAIL(tetMesh->tList);
  while (t != NULL)
  {
    if(   ((TET_VERTEX (t, 0)->flag & POINT_CUBOID)  == POINT_CUBOID)
       || ((TET_VERTEX (t, 1)->flag & POINT_CUBOID)  == POINT_CUBOID)
       || ((TET_VERTEX (t, 2)->flag & POINT_CUBOID)  == POINT_CUBOID)
       || ((TET_VERTEX (t, 3)->flag & POINT_CUBOID)  == POINT_CUBOID) )
    {
      t = TET_NEXT(t);
      continue;
    }
    
    Tetrah_MinAndMaxSolidAngle(t, &minAngle, &dummy_index, &maxAngle,
                               &dummy_index);
    if (minAngle < meshMinSolidAngle)
    {
      meshMinSolidAngle = minAngle;
    }
    if (maxAngle > meshMaxSolidAngle)
    {
      meshMaxSolidAngle = maxAngle;
    }

    Tetrah_MinAndMaxDihedralAngle(t, &minAngle, &dummy_index, &maxAngle,
                                     &dummy_index);
    if (minAngle < meshMinDihAngle)
    {
      meshMinDihAngle = minAngle;
    }
    if (maxAngle > meshMaxDihAngle)
    {
      meshMaxDihAngle = maxAngle;
    }

    t = TET_NEXT(t);
  }   

  lim1 = 4*deg2rad;
  lim2 = meshMinSolidAngle + 4*deg2rad;
  *ms_min = lim1 < lim2 ? lim1 : lim2;

  lim1 = 240*deg2rad;
  lim2 = meshMaxSolidAngle - 60*deg2rad; 
  *ms_max = lim1 < lim2 ? lim2 : lim1; 

  lim1 = 30*deg2rad;
  lim2 = meshMinDihAngle + 10*deg2rad;
  *md_min = lim1 < lim2 ? lim1 : lim2;

  lim1 = 150*deg2rad;
  lim2 = meshMaxDihAngle - 20*deg2rad;
  *md_max = lim1 < lim2 ? lim2 : lim1;
} /* end of GoodAnglesLimits */


/****f*
 * NAME
 *   ClassifyTetrah
 * DESCRIPTION
 *   Flags each tetrahedron according to its type. Each tetrahedron can
 *   belong to only one type!
 * ARGUMENTS
 *   tetMesh - tetrahedral mesh
 ******/
void ClassifyTetrah (TetrahMesh *tetMesh)
{
  double ms_min, ms_max;  /* small and large mesh solid angle */
  double md_min, md_max;  /* small and large mesh dihedral angle */
  Tetrah *t;
 
  GoodAnglesLimits (tetMesh, &ms_min, &ms_max, &md_min, &md_max);

  t = LIST_TAIL(tetMesh->tList);
  while (t != NULL)
  {
    TetrahSetTypeFlag(t, ms_min, ms_max, md_min, md_max);
    t = TET_NEXT(t);
  }   
} /* end of ClassifyTetrah */



/****f*
 * NAME
 *   CountTetrahType
 * DESCRIPTION
 *   Count how many tetrahedra of particular kind (round, needle, wedge,
 *   spindle, sliver, cap) is in the mesh; The same time it flags each
 *   tetrahedron according to its type.
 *
 * ARGUMENTS
 *   tetMesh - tetrahedral mesh (IN)
 *   tetTypeCount - must point to and array holing ot he size minimum
 *                  NTYPES_TET the array will be filled with counters of
 *                  the particular tetrahedron type.
 ******/
void CountTetrahType(TetrahMesh *tetMesh, int* tetTypeCount)
{
  double ms_min, ms_max;  /* small and large mesh solid angle */
  double md_min, md_max;  /* small and large mesh dihedral angle */
  int i;
  Tetrah *t;
 
  for(i = 0; i < NTYPES_TET; i++)
  {
    tetTypeCount[i] = 0;
  }

  GoodAnglesLimits(tetMesh, &ms_min, &ms_max, &md_min, &md_max);

  t = LIST_TAIL(tetMesh->tList);
  while(t != NULL)
  {
    tetTypeCount[TetrahSetTypeFlag(t, ms_min, ms_max, md_min, md_max)]++;
    t = TET_NEXT(t);
  }
} /* end of CountTetrahType */


/****f*
 * NAME
 *   TetrhaSetTypeFlag 
 * DESCRIPTION
 *   Flag each tetrahedron according to its type (round, needle, wedge,
 *   spindle, sliver, cap);
 *   It returns enum value according to each type - see TetrahType in
 *   mesaures.h.
 * ARGUMENTS
 *   tet - tetrahedron
 *    ms_min - lower limit on mesh solid angle 
 *    ms_max - uper limit on mesh solid angle 
 *    md_min - lower limit on mesh dihedral angle
 *    md_max - uper limit on mesh dihedral angle
 ******/
enum TetrahType TetrahSetTypeFlag (Tetrah *tet, double ms_min, double ms_max,
                                   double md_min, double md_max)
{
  double ts_min, ts_max;  /* min and max tetrah solid angle */
  double td_min, td_max;  /* min and max tetrah  dihedral angle */
  int dummy;
  enum TetrahType result;


  if(   ((TET_VERTEX (tet, 0)->flag & POINT_CUBOID)  == POINT_CUBOID)
     || ((TET_VERTEX (tet, 1)->flag & POINT_CUBOID)  == POINT_CUBOID)
     || ((TET_VERTEX (tet, 2)->flag & POINT_CUBOID)  == POINT_CUBOID)
     || ((TET_VERTEX (tet, 3)->flag & POINT_CUBOID)  == POINT_CUBOID) )
  {
    tet->flag |= TET_UNKNOWN_TYPE;
    return UNKNOWN_TET_TYPE;
  }
  
  Tetrah_MinAndMaxSolidAngle (tet, &ts_min, &dummy, &ts_max, &dummy);
  Tetrah_MinAndMaxDihedralAngle (tet, &td_min, &dummy, &td_max, &dummy);

  /* clear all tetrahedron type flags */
  tet->flag &= ~(  TET_ROUND | TET_NEEDLE | TET_WEDGE | TET_SPINDLE 
                 | TET_SLIVER | TET_CAP | TET_UNKNOWN_TYPE);

  /* slivers are difficult to remove with traditional methods,
     always must be counted */
  if (IsSliver(ts_min, ts_max, td_min, td_max, ms_min, ms_max, md_min, 
               md_max))
  {
     tet->flag |= TET_SLIVER;
     result = SLIVER_TET_TYPE;
  }
  else if (IsCap(ts_min, ts_max, td_min, td_max, ms_min, ms_max, md_min,
                 md_max))
  {
     tet->flag |= TET_CAP;
     result = CAP_TET_TYPE;
  }
  else if (IsSpindle(ts_min, ts_max, td_min, td_max, ms_min, ms_max, md_min, 
                     md_max))
  {
     tet->flag |= TET_SPINDLE;
     result = SPINDLE_TET_TYPE;
  }
  else if (IsWedge(ts_min, ts_max, td_min, td_max, ms_min, ms_max, md_min, 
                   md_max))
  {
     tet->flag |= TET_WEDGE;
     result = WEDGE_TET_TYPE;
  }
  else if (IsNeedle(ts_min, ts_max, td_min, td_max, ms_min, ms_max, md_min, 
                    md_max))
  {
     tet->flag |= TET_NEEDLE;
     result = NEEDLE_TET_TYPE;
  }
  else if (IsRound(ts_min, ts_max, td_min, td_max, ms_min, ms_max, md_min, 
                   md_max))
  {
     tet->flag |= TET_ROUND;
     result = ROUND_TET_TYPE;
  }
  else
  {
     tet->flag |= TET_UNKNOWN_TYPE;
     result = UNKNOWN_TET_TYPE;
  }
  return result;
} /* end of TetrahSetTypeFlag */


/****f*
 * NAME
 *   Tetrah_MinAndMaxAngle 
 * DESCRIPTION
 *   Calculates minimal and maximal angle of a tetrahedron.
 * ARGUMENTS
 *   tet      - pointer to tetrahedron (IN)
 *   angle    - angle type (see AngleType enum definition) (IN)
 *   minAngle - minimal angle in tet (OUT)
 *   minind   - index of the part ot tetrahdron related to that angle (OUT)
 *   maxAngle - maximal angle in tet (OUT)
 *   maxind -   index of the part ot tetrahdron related to that angle (OUT)
 * RETURN VALUE
 *   NONE
 ******/
void Tetrah_MinAndMaxAngle (Tetrah *tet, enum AngleType angle,
                            double *minAngle, int *minind, 
                            double *maxAngle, int *maxind)
{
  switch (angle)
  {
    case DIHEDRAL_ANGLE:
      Tetrah_MinAndMaxDihedralAngle (tet, minAngle, minind, maxAngle, maxind);
      break;
    case SOLID_ANGLE:
      Tetrah_MinAndMaxSolidAngle (tet, minAngle, minind, maxAngle, maxind);
      break;
    default:
      e_GeneralError ("%s:%d : Internal error - unknown angle type", 
                      __FILE__, __LINE__); 
  }
} /* end of Tetrah_MinAndMaxAngle */


/****f*
 * NAME
 *   Tetrah_Measure 
 * DESCRIPTION
 *   Calculates measure of a tetrahedron.
 * ARGUMENTS
 *   tet -  
 *   measure - measure type (see MeasureType enum definition) (IN)
 * RETURN VALUE
 *   NONE
 ******/
double Tetrah_Measure (Tetrah *tet, enum MeasureType measure)
{
  int dummy;
  double mval=-999.0;
  switch (measure)
  {
    case RADIUS_RATIO_MEASURE:
      mval = Tetrah_RadiusRatio (tet);
      break;
    case MEAN_RATIO_MEASURE:
      mval = Tetrah_MeanRatio (tet);
      break;
    case SIGMA_MEASURE:
      mval = Tetrah_MinSigmaMeasure (tet, &dummy);
      break;
    default:
      e_GeneralError ("%s:%d : Internal error - unknown measure type", 
                      __FILE__, __LINE__); 
  }
  return mval; 
} /* end of Tetrah_Measure */

/****f*
 * NAME
 *   FindMeasureLimits 
 * DESCRIPTION
 *   Calculates minimal and maximal measure of an assembly of tetrahedra.
 *   The assembly is passed as a stack of pointers to tetrahedra.
 * ARGUMENTS
 *   tetAssembly - stack of pointers to tetrahedra (IN)
 *   measure - measure type (see MeasureType enum definition) (IN)
 *   minMeasure - minimal measure in the assembly (OUT)
 *   maxMeasure - maximal mesure in the assembly (OUT)
 * RETURN VALUE
 *   NONE
 ******/
void FindMeasureLimits (Tos_MStack tetAssembly, enum MeasureType measure, 
                        double *minMeasure, double *maxMeasure)
{
   Tetrah *tet;
   double currentMeasure;
  
   /* all measures are non-negative and are bounded from above by 1.0 */
   *maxMeasure = -1.0;
   *minMeasure = 2.0;

  if (Tos_MStackCount (tetAssembly) <= 0)
  {
    e_GeneralError ("%s:%d : empty assembly of tetrahedra", __FILE__,
                     __LINE__);

  }
  Tos_MStackTraversalInit (tetAssembly);
  while ((tet = Tos_MStackTraverse (tetAssembly)) != NULL)
  {
    currentMeasure = Tetrah_Measure (tet, measure);
    if (*minMeasure > currentMeasure)
        *minMeasure = currentMeasure;
    if (*maxMeasure < currentMeasure)
        *maxMeasure = currentMeasure;
  }
} /* end of FindMeasureLimits */
