#include <assert.h>

#include "eg_lib.h"
#include "find.h"
#include "basic_op.h"
#include "mesh_op.h"
#include "boundary.h"
#include "debug.h"



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 InsertPointOnEdge --
        For given edge and point on it inserts that point into
        tetrahedral mesh maintaining updates of neighbouring 
        adjacency, node-elements adjacency and number of points
        in mesh.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void InsertPointOnEdge (TetrahMesh *tetMesh, Vertex *point, Edge *edge)
{
  register int i;
  register int vertex;
  Tos_MStack edgeShell; 
  Tos_MStack newTet; 
  Tetrah *newtet, *tet;
  Vertex *ne[2];
  int sanityCheck=0;

  /* initialize stacks */
  edgeShell = Tos_MStackNew ();
  newTet = Tos_MStackNew ();

  /* find edge shell */
  sanityCheck = FindEdgeShell (edge, edgeShell); 
  assert (sanityCheck > 0); /* There must be at least one elements
                             * in the edge shell */

  /* process each element in edge shell */
  Tos_MStackTraversalInit (edgeShell); 
  while ((tet = (Tetrah *) Tos_MStackTraverse (edgeShell)) != NULL)
  {
    /* find two vertices of tet which are not endpoint of edge */
    i=0;
    /* This  trick is to make the loop a bit faster, equivalent to
     * for (vertex=0; vertex<4; vertex++)
     */
    for (vertex=4; vertex--; )
    {
      if (   TET_VERTEX(tet, vertex) != EDGE_VERTEX(edge, 0)
          && TET_VERTEX(tet, vertex) != EDGE_VERTEX(edge, 1) )
      {
        ne[i] = TET_VERTEX (tet, vertex);
        i++;
      }
      if (i==2)
      {
        break;
      }
    }
    assert (i==2); /* there always must be two such vertices */

    /* construct new tetrahedra and push them on newTet stack */
    newtet = TetrahCreate ();
    TET_VERTEX (newtet, 0) = EDGE_VERTEX (edge, 0);
    TET_VERTEX (newtet, 1) = point;
    TET_VERTEX (newtet, 2) = ne[0];    
    TET_VERTEX (newtet, 3) = ne[1];    
    Tos_MStackPush (newTet, (void *) newtet);

    newtet = TetrahCreate ();
    TET_VERTEX (newtet, 0) = EDGE_VERTEX (edge, 1);
    TET_VERTEX (newtet, 1) = point;
    TET_VERTEX (newtet, 2) = ne[0];    
    TET_VERTEX (newtet, 3) = ne[1];    
    Tos_MStackPush (newTet, (void *) newtet);
  }

  /* Remove shell tetrahedra from the mesh and delete them */
  while (Tos_MStackCount (edgeShell) > 0)
  {
     tet = (Tetrah *) Tos_MStackPop (edgeShell); 
     RemoveTetFromMesh (tetMesh, tet);
     TetrahDelete (tet);
  }

  /* add new tetrahedra to mesh construction */
  while (Tos_MStackCount (newTet) > 0)
  {
    tet = (Tetrah *) Tos_MStackPop (newTet);
    AddTetToMesh (tetMesh, tet);
  }

  /* add point to the list of point in tetMesh */
  PutVertexOnListHead(tetMesh->pList, point);
  /* advance the mesh points counter in tetMesh */
  (tetMesh->NMeshPoints)++;

  /* delete empty stacks */
  Tos_MStackFree (&newTet);
  Tos_MStackFree (&edgeShell);

  return;
} /* end of InsertPointOnEdge */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 InsertPointOnFace --
        Inserts point on specified face spliting face shell,
        creating new tetrahedra and adds them to mesh structure
        and removing the shell ones
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void InsertPointOnFace (TetrahMesh *tetMesh, Vertex *point, Face *face)
{
  register int i;
  register int j;
  register int k;
  Tetrah *faceShell[2];
  int faceNum[2]; /* generic numbers of common face in shell elements */
  int shellCount; /* number of tetrahedra in shell */
  Tetrah *newTet[6]; /* array of new tetrahedra */
  int sanityCheck; 
  
  /* find shell associated with face */
  shellCount = FindFaceShell (face, faceShell);

  if (shellCount == 0)
  {
    /* face is not in tetrahedral construction */
    e_GeneralError ("InsertPointOnFace: face is not in tetrahedral mesh");
  }
  
  j=0;
  for (i=0; i<shellCount; i++)
  {
    /* find generic number of common face is shell element */
    sanityCheck = TetHasFace (faceShell[i], face, &(faceNum[i]));
    assert (sanityCheck == TRUE);  
    /* create new tetrahedra */
    for (k=3; k--; )
    {
      newTet[j] = TetrahCreate();
      /* set vertices in created tetrahedra */
      TET_VERTEX (newTet[j], 0) = TET_VERTEX (faceShell[i], faceNum[i]);
      TET_VERTEX (newTet[j], 1) = point;
      TET_VERTEX (newTet[j], 2) = FACE_VERTEX (face, sideDef[k][0]); 
      TET_VERTEX (newTet[j], 3) = FACE_VERTEX (face, sideDef[k][1]); 
      /* advance number of created tetrahedra */
      j++;
    }
  }
  for (i=0; i<shellCount; i++)
  {
    /* Remove shell elements from the mesh structure */
    RemoveTetFromMesh (tetMesh, faceShell[i]);

    /* Delete shell elements */
    TetrahDelete (faceShell[i]);
  }

  /* add new tetrahedra to the mesh */
  for (k=0; k<j; k++)
  {
    AddTetToMesh (tetMesh, newTet[k]);
  }

  /* add point to tetMesh points list */
  PutVertexOnListHead(tetMesh->pList, point);
  /* advance points counter in tetMesh */
  tetMesh->NMeshPoints++;
 
  return;
} /* end of InsertPointOnFace */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 AddTetToMesh --
        Adds tetrahedron to TetrahMesh data structure: 
        -- update adjacency information in neighbouring tetrahedra
        -- update node-element information in tetrahedron vertices
        -- add tetrahedron to the list of tetrahedra in mesh
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void AddTetToMesh (TetrahMesh *tetMesh, Tetrah *tet)
{
  register int faceNum; 
  int neighbFaceNum; 
  Vertex *point;
  Face *face;
  Face  dummy_face;
  Tetrah *faceShell[2]; /* face shell - at most two adjacent tetrahedra */
  int shellCount; /* counter of number of elements in the shell */
  int commonFacesCount; /* count how many faces are shared between
                         * new tetrah and mesh 
                         */   
  int sanityCheck;
  /* as faces are implemented with Tetrah data structure
   * dummy_tet is used to avoid dynamic allocation of memory for
   * face structure
   */
  face = &(dummy_face);

  commonFacesCount = 0;

  for (faceNum=0; faceNum<4; faceNum++)
  {
    GetFaceFromTet (tet, faceNum, face);
    shellCount = FindFaceShell (face, faceShell);
    /* there should be only one tetrahedron (or none) in the mesh
     * structure adjacent to given face (otherwise after inserting 
     * new tetrah mesh will be invalid
     */  
    assert (shellCount <= 1); 
    commonFacesCount += shellCount;

    /* check if there is a tetrahedron to be glued with new one */
    if (shellCount == 1)
    {
      sanityCheck = TetHasFace (faceShell[0], face, &neighbFaceNum); 
      assert (sanityCheck == TRUE); /* face must be in faceShell[0] */ 
      if (TET_NEIGHBOUR(faceShell[0], neighbFaceNum) != NULL)
      {
        /* tetrahedron (shellFace[0]) to which new tetrahedron 
         * should be glued hasn't got a free neighbour pointer
         * at given face which means that somenthing went wrong
         * and tetrahedral mesh is not consistent
         * When optimizing this check can be skipped if you trust the 
         * code enough
         */  
         e_GeneralError
          ("AddTetToMesh: Cannot add tetrahedron to mesh (line %d)", 
           __LINE__); 
      }
      /* update neighbouring information */
      TET_NEIGHBOUR (tet, faceNum) = faceShell[0];
      TET_NEIGHBOUR (faceShell[0], neighbFaceNum) = tet;
    }
  }
  for (faceNum=0; faceNum<4; faceNum++)
  {
    /* update node-element adjacency information */
    point = TET_VERTEX (tet, faceNum);
    Tos_MStackPush (point->adjElem, (void *) tet);
  }
  /* add tetrahedron to the list of tetrahedra in the mesh data 
   * structure 
   */
  PutTetOnListHead(tetMesh->tList, tet);  
  /* advance number of tetrahedra in mesh data structure */
  (tetMesh->NTetrah)++;
  
  return;  
} /* end of AddTetToMesh */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RemoveTetFromMesh --
        Removes tetrahedron from tetrahedral mesh structure:
        - update neighbouring information 
        - update node-elements adjacency information
        - removes tetrahedron from list of tetrahedra in the mesh
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void RemoveTetFromMesh (TetrahMesh *tetMesh, Tetrah *tet)
{
  register int faceNum;   /* face number in tet */
  register int neighbFaceNum; /* face number in neibgbour of tet
                               * sharing the same face
                               */
  Vertex *point;
  Tetrah *neighbour;

  for (faceNum=0; faceNum<4; faceNum++)
  {
    /* set to NULL neighbour pointer in tetrahedra
     * that have common face with tet */  
    neighbour = TET_NEIGHBOUR (tet, faceNum);
    if (neighbour != NULL)
    {
      for (neighbFaceNum=0; neighbFaceNum<4; neighbFaceNum++)
      {
        if (TET_NEIGHBOUR (neighbour, neighbFaceNum) == tet)
        {
          TET_NEIGHBOUR (neighbour, neighbFaceNum) = NULL;
          break;
        }
      }
    }
    /* delete tet from adjacency stack of its vertex */
    point = TET_VERTEX (tet, faceNum);
    Tos_MStackRemove (point->adjElem, (void *) tet); 
  }  

  /* remove tet from list of tetrahedra in the mesh */
  RemoveTetFromList (tetMesh->tList, tet);

  /* decrease tetrahedra counter */
  (tetMesh->NTetrah)--; 

  return;
} /* end of RemoveTetFromMesh */




/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 SegmentTetMeshIntersect --
        Find intersection of a segment and components of tetrahedral
        mesh (faces and edges). It returns an intersection point 
        closest to starting endpoint of the segment.
 Arguments:
          tetMesh -- the tetrahedral mesh (IN)
          nodeA   -- the endpoint of the segment (IN)
          nodeB   -- the endpoint of the segment (IN)
          tet     -- the tetrahedron adjacent to nodeA containing
                     the intersection point
 intersectionType -- flag defining the type of the intersection
                     (EG_ONEDGE - for intersection on tetrahedron edge
                      EG_INSIDE - for intersection inside the tetrahdedron
                                  face
                      EG_ONVERTEX - for intersection on tetrahedron vertex
                     )
 intersectionPoint -- the point of the intersection
         vertices -- the indices of vertices in tetrahedron which constitute
                     the intersected entity
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int SegmentTetMeshIntersect (TetrahMesh *tetMesh, Vertex *nodeA,
                             Vertex *nodeB, Tetrah **tet,
                             int *intersectionType, 
                             Vertex **intersectionPoint, int *vertices)
{
  register int face;
  int          gennum;
  double       isCoords[3];
  double       dummy[3];
  double       mu;
  Tetrah      *currtet;
  int          ipFound = FALSE;
  int          intersectFaceNum=-1, intersectGenNum=-1;
  int          where;
  int          side;
  double       checkVolume[3];
  int          badTetrah;
  int          badSide[3] = {0, 0, 0};
  int          count, count1;

  /* traverse adjacency stack at node A and check if any 
   * of the tetrahedron contains node B; if so edge A-B is
   * present in mesh
   */
  Tos_MStackTraversalInit (nodeA->adjElem);
  while ((currtet = Tos_MStackTraverse (nodeA->adjElem)) != NULL)
  {
    for (face=0; face<4; face++)
    {
      if (TET_VERTEX (currtet, face) == nodeB)
      {
        *tet = currtet;
        *intersectionPoint = nodeB;
        *intersectionType = EG_ONVERTEX;
        vertices[0] = face;
        return TRUE;
      }
    }
  }
  
  /* create intersection point */
  *intersectionPoint = VertexCreate();

  /* prepare to traverse adjacency stack at node A */
  Tos_MStackTraversalInit (nodeA->adjElem);
  while ((currtet = Tos_MStackTraverse (nodeA->adjElem)) != NULL)
  {
    for (face=0; face<4; face++)
    {
      /* skip faces which contain node A */
      if (   TET_VERTEX (currtet, faceDef[face][0]) == nodeA
          || TET_VERTEX (currtet, faceDef[face][1]) == nodeA
          || TET_VERTEX (currtet, faceDef[face][2]) == nodeA )
      {
        continue; 
      }
      
      where =
          eg_V1_TriangLineIntersect3D(TET_VCOORDS (currtet, faceDef[face][0]),
                                      TET_VCOORDS (currtet, faceDef[face][1]),
                                      TET_VCOORDS (currtet, faceDef[face][2]),
                                      nodeA->coords, nodeB->coords,
                                      isCoords, &gennum, dummy, &mu);
      if (where == EG_ONVERTEX || where == EG_ONEDGE || where == EG_INSIDE)
      {
        if (mu>= 0.0 && mu<=1.0)
        {
          intersectFaceNum = face;
          intersectGenNum = gennum;
          *intersectionType = where;
          *tet = currtet;
          (*intersectionPoint)->coords[0] = isCoords[0];
          (*intersectionPoint)->coords[1] = isCoords[1];
          (*intersectionPoint)->coords[2] = isCoords[2];
          ipFound = TRUE;
          break;
        }
      }
    }
    
    if (ipFound == TRUE) break;
  }
  
  /* FIXME: Handle the special case when intersection point
     has not been found */
  if (ipFound != TRUE)
  {
    printf("*** Handling special case in segment mesh intersection ***\n");

    count=0;
    count1=0; 
    Tos_MStackTraversalInit (nodeA->adjElem);
    while ((currtet = Tos_MStackTraverse (nodeA->adjElem)) != NULL)
    {
      for (face=0; face<4; face++)
      {
        /* skip faces which contain node A */
        if (   TET_VERTEX (currtet, faceDef[face][0]) == nodeA
            || TET_VERTEX (currtet, faceDef[face][1]) == nodeA
            || TET_VERTEX (currtet, faceDef[face][2]) == nodeA)
        {
          continue;
        }
        count1++;
        where = eg_DoSegmentTriangleIntersect(
              TET_VCOORDS (currtet, faceDef[face][0]),
              TET_VCOORDS (currtet, faceDef[face][1]), 
              TET_VCOORDS (currtet, faceDef[face][2]), 
              nodeA->coords, nodeB->coords); 
        if (where == 1)
        {
          count++;
        }
      }
    }
  }
  assert (ipFound == TRUE);

  if (*intersectionType == EG_INSIDE)
  {
    vertices[0] = faceDef[intersectFaceNum][0];
    vertices[1] = faceDef[intersectFaceNum][1];
    vertices[2] = faceDef[intersectFaceNum][2];
    /* check the volume of tetrahedrons created by any
       two vertices of face intersection point and node A
     */
    badTetrah=0;
    for (side=0; side<3; side++)
    {
      checkVolume[side] =
          eg_Orient3D (nodeA->coords,
                       (*intersectionPoint)->coords,
                       TET_VCOORDS(*tet, vertices[sideDef[side][0]]),
                       TET_VCOORDS(*tet, vertices[sideDef[side][1]])  );
      if (E_ABS(checkVolume[side]) < 10.e-10)
      {
        badSide[badTetrah] = side;
        badTetrah++;
      }
    }
    
    switch (badTetrah)
    {
      case (0):
      {
        break;
      }
      case (1):
      {
        *intersectionType = EG_ONEDGE;
        intersectGenNum = badSide[0];
        break; 
      }
      case (2):
          /* FALLTHRU */
      case (3):
      {
        *intersectionType = EG_ONVERTEX;
        intersectGenNum = 3 - badSide[0] - badSide[1] - badSide[2]; 
        break;
      }
    }
  }
  
  if (*intersectionType == EG_ONEDGE)
  {
    vertices[0] = faceDef[intersectFaceNum][sideDef[intersectGenNum][0]];
    vertices[1] = faceDef[intersectFaceNum][sideDef[intersectGenNum][1]];
    badTetrah=0;
    for (side=0; side<2; side++)
    {
      checkVolume[side] =
          eg_Orient3D (nodeA->coords,
                       (*intersectionPoint)->coords,
                       TET_VCOORDS(*tet,
                                   faceDef[intersectFaceNum][intersectGenNum]),
                       TET_VCOORDS(*tet, vertices[side]) );
      if (E_ABS(checkVolume[side]) < 10.e-10)
      {
        *intersectionType = EG_ONVERTEX;
        vertices[0] = vertices[side]; 
        break;
      }
    }
  }
  else if (*intersectionType == EG_ONVERTEX)
  {
    vertices[0] = faceDef[intersectFaceNum][intersectGenNum];
  }
  
  /* return FALSE which means that edge A-B is not present in mesh */  
  return FALSE;
} /* end of SegmentTetMeshIntersect */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 IsFaceOnMeshBoundary --
        Check if face is specified in surface
 mesh 'surfMeshNum' (or in list of subfaces created on 'surfMeshNum'
 surface mesh)
 
 Arguments:
        face -- face under consideration 
        surfMeshNum -- number of surface mesh to be checked
        
 Return value:
        1 - if face is present in the surface mesh
        0 - otherwise
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int IsFaceOnMeshBoundary (MeshGenData *mgds, int surfMeshNum,
                          Face *surface, int boundFlag)
{
  int i;
  Face *face, *splitface;
  struct Tos_MStack faceIter = {0, NULL, NULL};
  
  /* do the quick check of flags of nodes */
  if(IsFaceOnBoundary(surface, boundFlag) == 0)
  {
    return 0;
  }
  
  /* Scan through input mesh faces.
   * This is necessary, since in the case of a non-convex domain there can be
   * a face which has got three boundary points, but not part of the
   * boundary itself !
   */
  for(i = 0; i < mgds->fTableLength[surfMeshNum]; i++)
  {
    face = mgds->fTable[surfMeshNum][i];

    for(splitface = InitSplitFace(face, &faceIter);
        splitface != NULL;
        splitface = NextSplitFace(&faceIter))
    {
      if(CompareFaces(surface, splitface) == 1)
          return 1;
    }
  }
  return 0; 
} /* end of IsFaceOnMeshBoundary */



/* Modifications:
 * $Log: mesh_op.c,v $
 * Revision 1.13  2002/07/28 13:27:42  peteri
 * - list creation interface is modified
 * - it is possible to add to the head or the tail of the list
 * - all related code is updated
 *
 * Revision 1.12  2002/07/26 13:29:31  peteri
 * - recursive functions are replaced with iterative functions
 *
 * Revision 1.11  2002/07/26 13:07:03  peteri
 * minor modifiactions to remove compiler warnings
 *
 * Revision 1.10  2002/07/26 11:50:23  peteri
 * - minor editing
 * - nana is removed for simplicity
 * - boundary face detection is adjusted to new data structure,
 *   though it is recursive at the moment !
 *
 * Revision 1.9  2002/03/18 15:58:26  peteri
 * Comments are corrected.
 *
 * Revision 1.8  2002/02/27 22:56:20  romek
 * Fixed bug with Nana configuration.
 *
 * Revision 1.7  2002/02/24 19:03:53  romek
 * *** empty log message ***
 *
 */
