/********************************************************************
 mstack.c -- 
       Tos_MStack -- stack based on Tos_MemPool memory allocator.

 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University, Edinburgh    
 ********************************************************************/

/* $Id: mstack.c,v 1.5 2002/07/19 20:43:50 peteri Exp $ */

#include "mstack.h"
#include "assert.h"
#include <stdlib.h>

/* initialise the stack memory pool */
void MStackMemPoolInit(void)
{
  ;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackNew -- 
        Allocate and initialize new stack.
 Arguments: 
 Return:               
        Return empty stack;
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Tos_MStack Tos_MStackNew (void) 
{
  Tos_MStack stk;
  
  /* allocate memory for new stack */
  stk = (Tos_MStack)calloc(1, sizeof(struct Tos_MStack));
  stk->count = 0;
  stk->head = NULL;
  stk->pathItem = NULL;
  return stk; 
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackEmpty -- 
        Check if there are elements on stack; 

 Arguments:
        stk - stack
 Return:       
        If stack is empty return 1, otherwise return 0.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int Tos_MStackEmpty (Tos_MStack stk)
{
  assert (stk != NULL);
  return stk->count == 0;
} 

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackPush --  
        Push element on stack.
        
 Arguments:       
        stk - stack
        x - void pointer to element pushed on stack
 Return:
        NONE
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void Tos_MStackPush (Tos_MStack stk, void *x)
{
  struct elem *t;
  assert (stk != NULL);
  t = (struct elem *) calloc (1, sizeof (struct elem)); 
  t->x = x;
  t->next = stk->head;
  stk->head = t;
  stk->count++;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackPop -- 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void* Tos_MStackPop (Tos_MStack stk)
{
  void *x;
  struct elem *t;

  assert (stk != NULL);
  assert (stk->count > 0);
  t = stk->head;
  stk->head = t->next;
  (stk->count)--;
  x = t->x;
  free (t); 
  return x; 
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackFree -- 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void Tos_MStackFree (Tos_MStack *stk)
{
  struct elem *t, *u;
  
  assert (stk != NULL);
  assert ((*stk) != NULL);

  for ( t = (*stk)->head; t; t=u)
    {
      u = t->next;
      free (t); 
    }  
  free ((void *) *stk); 
  stk = NULL;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackFlush -- 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void Tos_MStackFlush (Tos_MStack stk)
{
  struct elem *t, *u;
  
  assert (stk != NULL);

  for ( t = stk->head; t; t=u)
    {
      u = t->next;
       free (t);
    }  
  stk->count = 0;
  stk->head = NULL;
  stk->pathItem = NULL;
}



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackTraversalInit -- 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void Tos_MStackTraversalInit (Tos_MStack stk)
{
  stk->pathItem = stk->head; 
  return;
}  

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackTraverse --
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void *Tos_MStackTraverse (const Tos_MStack stk)
{
  struct elem *t; 
  t = stk->pathItem;
  if (stk->pathItem != NULL)
  {
     t = stk->pathItem;
     stk->pathItem = (stk->pathItem)->next;
     return t->x;
  }
  return NULL;
}  /* end of Tos_MStackTraverse */
 
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackSearch
        Search for element on the stack 
 Retrun value:
        0 - if element is not on the stack
        1 - if element is on the stack 

 Complexity O(n) where n is stack size.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int Tos_MStackSearch (Tos_MStack stk, void *x)
{
  int found=0;
  struct elem *t = stk->head;

  while (t != NULL)
  {
    if (t->x == x) 
    {
      found = 1;
      break;
    }
    t = t->next;
  }
  return found;
} /* end of Tos_MStackSearch */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackSearchN
        Search how many times element occures on the stack. 
 Retrun value:
        Number of occurences of he element on the stack.

 Complexity O(n) where n is stack size.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int Tos_MStackSearchN (Tos_MStack stk, void *x)
{
  int found=0;
  struct elem *t = stk->head;

  while (t != NULL)
  {
    if (t->x == x) 
    {
      found++;
    }
    t = t->next;
  }
  return found;
} /* end of Tos_MStackSearchN */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackRemove
        Remove element from the stack.
 Complexity O(n) where n is stack size.

 Retrun:
        If element to be removed is on the stack returns 1, 
        otherwise returns 0.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int Tos_MStackRemove (Tos_MStack stk, void *x)
{
  int found=0;
  struct elem *prev = NULL;
  struct elem *t = stk->head;

  while (t != NULL)
  {
    if (t->x == x) 
    {
      found = 1;
      break;
    }
    prev = t;
    t = t->next;
  }
  if (found == 1)
  {
    if (prev != NULL)
    {
      prev->next = t->next;
    }
    else
    {
      stk->head = t->next;
    } 
    free (t);
   
    (stk->count)--; 
  }
  return found;
}



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_MStackCount --
        Return number of elements on stack.

 Note: This probably should be turned to be macro for efficiency 
 reasons (need timing...).
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* It has been changed to macro *
        int
Tos_MStackCount (Tos_MStack stk)
{
  return stk->count;
}
*/

/* Modifications:
 * $Log: mstack.c,v $
 * Revision 1.5  2002/07/19 20:43:50  peteri
 * Mempool removed for simplicity
 *
 * Revision 1.4  2002/02/24 19:01:14  romek
 * *** empty log message ***
 *
 * Revision 1.3  2002/02/14 10:25:22  romek
 * Added function Tos_MStackSearchN.
 *
 * Revision 1.2  2002/01/31 21:27:54  peteri
 * Major reorganization in the code.
 * - It is split into smaller modules.
 * - Some types, functions are renamed
 * - Proper flipping of edges is implemented
 *
 * Revision 1.1.1.1  2002/01/30 10:20:55  peteri
 * Imported using TkCVS
 *
 * Revision 1.3  2001/08/29 12:44:01  romek
 * Tos_MStackFluh added
 *
 * Revision 1.2  2001/07/14 17:42:44  romek
 * last working version .
 *
 * Revision 1.6  2000/07/25 16:23:33  romek
 * (Tos_MStackSearch) : removed unused variable prev
 *
 * Revision 1.5  2000/06/23 15:32:00  romek
 * (Tos_MStackFree) : fixed bug in assertion
 *
 * Revision 1.4  2000/06/14 11:26:23  romek
 * (Tos_MStackSearch) : New function.
 *
 * Revision 1.3  2000/02/10 14:49:54  romek
 * Tos_MStackCount changed to macro and moved to mstack.h
 *
 * Revision 1.2  2000/02/10 11:23:27  romek
 * Tos_MStackCount added
 *
 * Revision 1.1  2000/02/06 17:14:24  romek
 * Initial revision
 *
 */ 
            
