/********************************************************************
 gentrianadj.c --
        Generalized adjaceny for Triang1 elements mesh. 
 
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University, Edinburgh    
 ********************************************************************/

/* $Id: triangadj.c,v 1.3 2002/03/18 16:00:30 peteri Exp $ */
            
#include <assert.h>

#include "e_lib.h"
#include "nodelem.h"
#include "triangadj.h"


/* definition of generic number of nodes on side */
static int sideDef[3][2] = {{1, 2}, {2, 0}, {0, 1}};


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Triang1ElemsAdjToSide --
        For given Triang1 element 'elem' return "only one" element
        adjacent to 'side'.
        It returns the number of adjacent elements.
        
 Arguments:
        mesh - pointer to mesh of Triang1 elements   (IN)
        elem - element index                         (IN)
        side - side number (0, 1 or 2)               (IN) 
        nodeElemAdjLen - vector (of lenght NMeshPoints) of lengths
                of rows of nodeElemAdj matrix        (IN)  
        nodeElemAdj - nodes:->elements adjacency matrix (IN)
        sideAdj                                         (OUT)        

 Return:
       Number of elements adjacent to side >=0.  
       
 Note:
        sideAdj should be a NULL pointer otherwise the memory pointed
        by it will be lost.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int Triang1ElemsAdjToSide (MESH *mesh, 
                                  int elem, int side, 
                                  int *nodeElemAdjLen, 
                                  int **nodeElemAdj, 
                                  int *adjElem)
{
  int node1, node2; /* nodes of the face */
  int found = FALSE;
  int pointelem, i, j;
  int adjcount=0;

  *adjElem = 0;
  /* get the numbers of nodes on face */
  node1 = mesh->NodesTriang1[elem][sideDef[side][0]];
  node2 = mesh->NodesTriang1[elem][sideDef[side][1]];
  
  for(i = 0; i < nodeElemAdjLen[node1-1]; i++)
  { /* for each element adjacent to node1 */
    pointelem = nodeElemAdj[node1-1][i]-1;
    found = FALSE;
    for(j = 0; j < 3; j++)
    {
      if(mesh->NodesTriang1[pointelem][j] == node2)
      {
        found = TRUE;
        break;
      }
    }
    if(   found == TRUE
       && pointelem != elem)
    {
      if(adjcount == 0)
      {
        *adjElem = pointelem+1;
      }
      adjcount++;
    }
  }
  return adjcount;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Triang1Adjacency
        Build adjacency matrix for triangular (Triang1) element mesh.
 Two elements are considered adjacent if they share  at least two
 nodes. This function can be safely called for meshes in which 
 a side is shared by more than two elements. The length of
 adjacency list is equal to the number of interfaces (which is 3 for Triang1
 element). When more than one element share a side the program
 issues an error message and aborts !
 If a side in an element is a boundary side (it is shared with no
 other element) it is represented as 0 in the adjacency list.

 The sides in the elements are numbered in such a way that side 'n' is
 opposite to vertex 'n'.

 Arguments:
        mesh - mesh containing only Triang1 elements          (IN)
        nodeElemAdjLen, nodeElemAdj - node-element adjacency
                                      information             (IN)
        AdjElems - array of lists of adjacent elements        (OUT)
 Return:       
        If mesh is not of single element type or is not Triang1 
        mesh return 0, otherwise return 1.
        In case of any memory allocation problem this fucntions
        calls e_GenralError which causes the program to abort. 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int Triang1Adjacency (MESH *mesh,
                      int *nodeElemAdjLen, int **nodeElemAdj,
                      int ***AdjElems)
{
  int  sideAdjLen = 0;  /* length of side adjacency vector */
  int  i, side;

  if (mesh->NElemsTriang1 != mesh->TotalNElems)
  {
    e_GeneralError("Mesh contains unsupported elements");
    return 0;
  }

  /* allocate memory */
  *AdjElems = e_alloc_cont_i_matr (mesh->NElemsTriang1, 3);
  if (*AdjElems == NULL)
    e_GeneralError ("Triang1Adjacency: Cannot alocate memory",
                    __LINE__);
  
  for (i=0; i<mesh->NElemsTriang1; i++)
  {
    for (side=0; side <3; side++)
    {
      sideAdjLen = Triang1ElemsAdjToSide(mesh, i, side,
                                         nodeElemAdjLen, nodeElemAdj,
                                         &((*AdjElems)[i][side]));
      if(sideAdjLen > 1)
      {
        e_GeneralError("Only 2D manifold objects can be used !");
        return 0;
      }
    }
  }
  
  return 1;
}


/* Modifications:
 * $Log: triangadj.c,v $
 * Revision 1.3  2002/03/18 16:00:30  peteri
 * Simplified adjacency determination.
 *
 * Revision 1.2  2002/02/24 19:01:18  romek
 * *** empty log message ***
 *
 * Revision 1.1  2002/01/31 21:27:39  peteri
 * Major reorganization in the code.
 * - It is split into smaller modules.
 * - Some types, functions are renamed
 * - Proper flipping of edges is implemented
 *
 * Revision 1.1.1.1  2002/01/30 10:20:55  peteri
 * Imported using TkCVS
 *
 * Revision 1.1  2001/07/14 17:42:44  romek
 * Initial revision
 *
 * Revision 1.2  2000/02/09 13:28:21  romek
 * Tos_Triang1Adjacency function added
 * change in sideDef[][]
 *
 * Revision 1.1  2000/02/08 19:07:24  romek
 * Initial revision
 *
 */ 
