/********************************************************************
 write_dx.c --     
        Functions for writing mesh related data in DX format. 
 
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University Edinburgh    
 ********************************************************************/

/* $Id: write_dx.c,v 1.4 2002/02/02 17:18:38 peteri Exp $ */

#include <stdio.h>
#include <time.h>
#include <assert.h>
#include "e_lib.h"
#include "eg_lib.h"

#include "convert.h"
#include "measures.h"
#include "mesh.h"
#include "write_dx.h"

/* Some error messages uses gcc __FUNCTION__ macro which 
 * is not defined in Sun cc  
 */
#ifndef __GNUC__
#   define __FUNCTION__ __FILE__
#endif

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *    Tos_OutputOuterSurfaceDXS
 * DESCRIPTION
 * ARGUMENTS
 * RETURN VALUE
 *   NONE
 ******/
void Tos_OutputOuterSurfaceDX (MeshGenData *mgds)
{
  FILE *filePtr;
  char *nameSuffix = "_outer_surface.dx";
  char *fileName;
  fileName = (char *) e_alloc_c_vect (strlen(mgds->mgCtrl->outputFileName)
                     + strlen(nameSuffix)+ 1);
  if (fileName == NULL)
  {
    e_GeneralError ("%s(%d) Cannot allocate memory for file name", __FILE__,
                    __LINE__ );
  }
  strcpy (fileName, mgds->mgCtrl->outputFileName);
  strcat (fileName, nameSuffix);
  filePtr = fopen (fileName, "w");
  if (filePtr == NULL)
  {
     e_GeneralError ("Cannot open file %s for writing", fileName);
  }
  Tos_WriteSurfaceMeshDX (mgds, filePtr, 0);
} /* Tos_OutputOuterSurfaceDX */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *    Tos_OutputDXShape
 * DESCRIPTION
 * ARGUMENTS
 * RETURN VALUE
 *   NONE
 ******/
void Tos_OutputDXShape (MeshGenData *mgds)
{
  FILE *filePtr;
  char *nameSuffix = "_shape.dx";
  char *fileName;
  fileName = (char *) e_alloc_c_vect (strlen(mgds->mgCtrl->outputFileName)
                     + strlen(nameSuffix)+ 1);
  if (fileName == NULL)
  {
    e_GeneralError ("%s(%d) Cannot allocate memory for file name", __FILE__,
                    __LINE__ );
  }
  strcpy (fileName, mgds->mgCtrl->outputFileName);
  strcat (fileName, nameSuffix);
  filePtr = fopen (fileName, "w");
  if (filePtr == NULL)
  {
     e_GeneralError ("Cannot open file %s for writing", fileName);
  }
  Tos_WriteMeshAndShapeDataDX (&(mgds->tetMesh), filePtr);
} /* Tos_OutputDXShape */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *    Tos_MeshAndShapeDataDX
 * DESCRIPTION
 * ARGUMENTS
 * RETURN VALUE
 *   NONE
 ******/
void Tos_WriteMeshAndShapeDataDX (TetrahMesh *tetMesh, FILE *filePtr)
{
  Vertex *vertex;
  Tetrah *tet;
  double *vc[4];
  double sigma, rho, eta;
  fprintf (filePtr,
       "object 1 class array type float rank 1 shape 3 items %d data follows\n",   
        tetMesh->NMeshPoints);

  vertex = LIST_TAIL(tetMesh->pList);
  while (vertex != NULL)
  {
    fprintf (filePtr, "\t%f\t%f\t%f\n", vertex->coords[0], 
                                          vertex->coords[1],
                                          vertex->coords[2]);
    vertex = VERTEX_NEXT(vertex);  
  }  

  fprintf (filePtr,
       "object 2 class array type int rank 1 shape 4 items %d data follows\n",   
        tetMesh->NTetrah);
  tet = LIST_TAIL(tetMesh->tList);
  while (tet != NULL)
  {
    vc[0] = TET_VCOORDS(tet,0);
    vc[1] = TET_VCOORDS(tet,1);
    vc[2] = TET_VCOORDS(tet,2);
    vc[3] = TET_VCOORDS(tet,3);
    if (eg_Orient3D(vc[0], vc[1], vc[2], vc[3]) > 0)
    {
       fprintf (filePtr, "\t%d\t%d\t%d\t%d\n", TET_VNUMBER(tet,0)-1,
           TET_VNUMBER(tet,1)-1, TET_VNUMBER(tet,2)-1, TET_VNUMBER(tet,3)-1);
    }
    else
    {
       fprintf (filePtr, "\t%d\t%d\t%d\t%d\n", TET_VNUMBER(tet,0)-1,
           TET_VNUMBER(tet,2)-1, TET_VNUMBER(tet,1)-1, TET_VNUMBER(tet,3)-1);
    }
    tet =TET_NEXT(tet);  
  }  
  fprintf (filePtr,"attribute \"element type\" string \"tetrahedra\"\n");
  fprintf (filePtr,"attribute \"ref\" string \"positions\"\n");

  fprintf (filePtr,
       "object 3 class array type float rank 1 shape 3 items %d data follows\n",   
        tetMesh->NTetrah);
  tet = LIST_TAIL(tetMesh->tList);
  while (tet != NULL)
  {
    int dummy;
    sigma = Tetrah_MinSigmaMeasure (tet, &dummy);
    rho = Tetrah_RadiusRatio (tet);
    eta = Tetrah_MeanRatio (tet);
    fprintf (filePtr, "\t%f\t%f\t%f\n", sigma, rho, eta);
    tet = TET_NEXT(tet);  
  }  
  fprintf (filePtr, "attribute \"dep\" string \"connections\"\n");
  fprintf (filePtr, "object \"tetrah mesh with shape measures\" class field\n");
  fprintf (filePtr, "component \"positions\" value 1\n");
  fprintf (filePtr, "component \"connections\" value 2\n");
  fprintf (filePtr, "component \"data\" value 3\n");
  fprintf (filePtr, "end\n");
} /* end of Tos_WriteMeshAndShapeDataDX */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *    Tos_WriteSurfaceMeshDX
 * DESCRIPTION
 * ARGUMENTS
 * RETURN VALUE
 *   NONE
 ******/
void Tos_WriteSurfaceMeshDX (MeshGenData *mgds, FILE *filePtr, int surfNum)
{
  Vertex  *vertex;
  Face    *face, *partFace;
  int      numOfElems=0;
  int      i;
  
  AssignNumbersToPoints (&(mgds->tetMesh));
  if (surfNum < 0 || surfNum >= mgds->NMeshes)
  {
    e_GeneralError ("%s(%d) Invalid number of surface %d", 
            __FILE__, __LINE__, surfNum);
  }
  
  /* count number of triangles */
  for (i = 0; i < mgds->fTableLength[surfNum]; i++)
  {
    face = mgds->fTable[surfNum][i];
    if(Tos_MStackCount(face->splitFaces) > 0)
        numOfElems += Tos_MStackCount(face->splitFaces);
    else
        numOfElems++;
  }
  
  fprintf (filePtr,
           "object 1 class array type float rank 1 shape 3 items %d data follows\n",
           mgds->tetMesh.NMeshPoints);

  vertex = LIST_TAIL(mgds->tetMesh.pList);
  while (vertex != NULL)
  {
    fprintf (filePtr, "\t%f\t%f\t%f\n",
             vertex->coords[0], 
             vertex->coords[1],
             vertex->coords[2]);
    vertex = VERTEX_NEXT(vertex);  
  }

  fprintf (filePtr,
           "object 2 class array type int rank 1 shape 3 items %d data follows\n",   
           numOfElems);

  /* triangles */
  for (i = 0; i < mgds->fTableLength[surfNum]; i++)
  {
    face = mgds->fTable[surfNum][i];
    if(Tos_MStackCount(face->splitFaces) > 0)
    {
      Tos_MStackTraversalInit(face->splitFaces);
      while((partFace = Tos_MStackTraverse(face->splitFaces)) != NULL)
      {
        fprintf (filePtr, "\t%d\t%d\t%d\n",
                 FACE_VNUMBER(partFace,0)-1,
                 FACE_VNUMBER(partFace,1)-1,
                 FACE_VNUMBER(partFace,2)-1);
      }
    }
    else
    {
      fprintf (filePtr, "\t%d\t%d\t%d\n",
               FACE_VNUMBER(face,0)-1,
               FACE_VNUMBER(face,1)-1,
               FACE_VNUMBER(face,2)-1);
    }
  }  
  fprintf (filePtr,"attribute \"element type\" string \"triangles\"\n");
  fprintf (filePtr,"attribute \"ref\" string \"positions\"\n");

  fprintf (filePtr,
           "object 3 class array type int rank 0 items %d data follows\n",   
           numOfElems);


  for (i = 0; i < mgds->fTableLength[surfNum]; i++)
  {
    face = mgds->fTable[surfNum][i];
    if(Tos_MStackCount(face->splitFaces) > 0)
    {
      Tos_MStackTraversalInit(face->splitFaces);
      while((partFace = Tos_MStackTraverse(face->splitFaces)) != NULL)
      {
        fprintf (filePtr, "\t%d\n", 1);
      }
    }
    else
    {
      fprintf (filePtr, "\t%d\n",0);
    }
  }  
  
  fprintf (filePtr, "attribute \"dep\" string \"connections\"\n");
  fprintf (filePtr, "object \"tetrah mesh with shape measures\" class field\n");
  fprintf (filePtr, "component \"positions\" value 1\n");
  fprintf (filePtr, "component \"connections\" value 2\n");
  fprintf (filePtr, "component \"data\" value 3\n");
  fprintf (filePtr, "end\n");
} /* end of Tos_WriteMeshAndShapeDataDX */


/* Modifications:
 * $Log: write_dx.c,v $
 * Revision 1.4  2002/02/02 17:18:38  peteri
 * - edge list is transformed into edge table
 * - related code is updated
 * - kernel is simplify a little bit
 * - minor editing and formatting
 *
 * Revision 1.3  2002/02/02 15:19:30  peteri
 * - addedfaceslist member from MeshGenData is removed
 * - most of the functions are updated
 * - sup_domain member is deleted from vertex, edge, face, tetrah structures
 * - some functions are merged, we have less code
 * - some other minor changes, editing
 *
 * Revision 1.2  2002/01/31 21:27:58  peteri
 * Major reorganization in the code.
 * - It is split into smaller modules.
 * - Some types, functions are renamed
 * - Proper flipping of edges is implemented
 *
 * Revision 1.1.1.1  2002/01/30 10:20:55  peteri
 * Imported using TkCVS
 *
 * Revision 1.1  2001/10/27 18:31:22  romek
 * Initial revision
 *
 */
