/********************************************************************
 tosmg.c --
       Main driver for Tos - tetrahedral mesh generator.
 
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University Edinburgh    
********************************************************************/

/* $Id: tosmg.c,v 1.5 2002/07/26 15:49:30 peteri Exp $ */
#include <stdio.h>
#include <stdlib.h>

#if defined (HAVE_CONFIG_H)
#include "config.h"
#endif 

#ifndef WIN32

/* if compiled not on GNU system use getopt instead of getopt_long */
#if defined (HAVE_GETOPT_LONG)   
#    include <getopt.h>
#else
#    include <unistd.h>
#endif

#else
#include "getopt.h"
#endif

#include "meshgen.h"
#include "ctr_read.h"
#include "e_lib.h"
#include "debug.h"

#define VERSION_MAJOR 0
#define VERSION_MINOR 1

/* required number of command line arguments set to -1 if program can be called
 * with variable number of arguments
 */
#define REQNUMPARAM 1

void UsageInfo(const char *progname);
void AffilInfo(void); 
void VersionInfo(const char *progname);

int main(int argc, char *argv[])
{
  int   c, i;
  MESH *outMesh; 
  MeshGenCtrl *mgCtrl; 
  MeshGenData *mgds;
  
  /* set default verbose level to be 1 - basic messages */
  vm_verbose_level = 1; 

  /* process command line options */
  while (1) /* run util exited by break */
  {
#if defined (HAVE_GETOPT_LONG)
    int option_index = 0;
    static struct option long_options[] =
    {
      {"version", 0, 0, 'v'},
      {"verbose", 1, 0, 'V'}, 
      {"help", 0, 0, 'h'},
      {0, 0, 0, 0}
    };
    c = getopt_long (argc, argv, "vhV:", long_options, &option_index);
#else
    c = getopt (argc, argv, "vhV:");
#endif /* HAVE_GETOPT_LONG */    
    
    if (c == -1)
       break;

    switch (c)
    {

      case 'V':
	vm_verbose_level = atoi (optarg);
        break;

      case '?':
      case 'h': /* print usage and exit */
        UsageInfo (argv[0]);
        exit (EXIT_SUCCESS);
        break;

      case 'v':
        VersionInfo(argv[0]);      
        AffilInfo();
        exit(EXIT_SUCCESS);       
        break;

      default:
        fprintf (stderr, 
              "Internal error : getopt returned character code 0%o ??\n", c);
        exit (EXIT_FAILURE);
    }
  }

  if (REQNUMPARAM >= 0 && (optind+REQNUMPARAM) != argc)
  {
    UsageInfo(argv[0]);
    exit(EXIT_FAILURE);    
  }

  /* create new control structure */
  mgCtrl = vm_CreateMeshGenCtrl();
  if(mgCtrl == NULL)
  {
    e_GeneralError("Cannot allocate memory for mesh control data structure");
  } 

  /* setup control data structure from a file */
  MeshGenCtrlRead(mgCtrl, argv[optind]);

  assert(mgCtrl->nmeshes <= VM_MAX_NMESHES);
  
  /* read in the mesh */
  for(i = 0; i < mgCtrl->nmeshes; i++)
  {
    e_GetMeshData(&(mgCtrl->inputMeshes[i]), mgCtrl->inputFileNames[i]); 
    e_GetMeshParams(mgCtrl->inputMeshes[i],  mgCtrl->inputFileNames[i]); 
  }

  /* create the mesh generation data */
  mgds = vm_CreateMeshGenData(mgCtrl);

  /* throw away the meshes, theoretically free them up */
  for(i = 0; i < mgCtrl->nmeshes; i++)
  {
    mgCtrl->inputMeshes[i] = NULL;
  }

  /* generate the volume mesh */
  outMesh = vm_GenerateVolumeMESH(mgds);  

  /* write out the resulting mesh */
  if(outMesh != NULL)
  {
    e_PutMeshData(outMesh, mgCtrl->outputFileName);
    e_PutDecompData(outMesh, mgCtrl->outputFileName);
  }

  return (EXIT_SUCCESS);
} /* end of main */

void VersionInfo(const char *progname)
{
  fprintf (stdout, "%s version %d.%d\n", progname, 
		  VERSION_MAJOR, VERSION_MINOR);
}

void UsageInfo(const char *progname)
{
  fprintf(stdout, "\nusage: %s [options] file\n\n", progname);
  fprintf(stdout, "Options and arguments:\n");
  fprintf(stdout, "----------------------\n");
  fprintf(stdout, "-h       | --help          : prints this message\n");
  fprintf(stdout, "-v       | --version       : prints version info\n");
  fprintf(stdout, "-V level | --verbose=level : sets verbosity level" 
                  " (0 = no messages)\n\n");
  fprintf(stdout, "file                       : mesh generator control file "
                  "(.ctr) \n\n");
}           

void AffilInfo(void)
{
  fprintf(stdout, "Roman Putanowicz\n"); 
  fprintf(stdout, "SECT Research Group\n");
  fprintf(stdout, "Heriot-Watt University, Edinburgh\n");
}	

/* Modifications:
 * $Log: tosmg.c,v $
 * Revision 1.5  2002/07/26 15:49:30  peteri
 * - adjusted to mesh and boundary data separation
 *
 * Revision 1.4  2002/02/26 21:40:30  romek
 * Usage string improved.
 *
 * Revision 1.3  2002/02/24 19:00:10  romek
 * Added proper command line handlig.
 *
 * Revision 1.2  2002/02/14 10:43:35  romek
 * Test program update accordin to the changes in the source routines.
 *
 * Revision 1.1.1.1  2002/01/30 10:20:56  peteri
 * Imported using TkCVS
 *
 * Revision 1.1  2001/10/06 11:04:03  romek
 * Initial revision
 *
 */ 
            
