#include <math.h>
#include <e_lib.h>


#define SET_CURVE_POINT_APPROX_ZERO  1.0e-012


/* ////////////////////////////////////////////////////////////////////////// */
/* SetCurvePoint: To set the coordinates "x" of a point on a NURBS curve
   given the point's curve parameter "t" (0.0 <= t <= 1.0).                   */


void SetCurvePoint(NURBSCURVE *pCurve, double t, double *x)
{
   int j,l;

   double Denominator;
   double Factor0,Factor;
   double tFactor;


   if (E_ABS(t) < SET_CURVE_POINT_APPROX_ZERO)
   {
      for (j = 0; j < E_NDIMS; j++)
         x[j] = pCurve->CntlPoints[0][j];
   }
   else if (E_ABS(1.0 - t) < SET_CURVE_POINT_APPROX_ZERO)
   {
      for (j = 0; j < E_NDIMS; j++)
         x[j] = pCurve->CntlPoints[pCurve->Degree][j];
   }
   else
   {
      tFactor = t/(1.0 - t);
      Factor0 = pow(1.0 - t,(double)(pCurve->Degree));
      Factor = Factor0;
      Denominator = Factor*pCurve->CntlPoints[0][E_NDIMS];
      x[0] = Factor*pCurve->CntlPoints[0][0]*pCurve->CntlPoints[0][E_NDIMS];

      for (l = 1; l <= pCurve->Degree; l++)
      {
         Factor *= tFactor*(pCurve->Degree - l + 1)/l;
         Denominator += Factor*pCurve->CntlPoints[l][E_NDIMS];
         x[0] += Factor*pCurve->CntlPoints[l][0]*pCurve->CntlPoints[l][E_NDIMS];
      }

      x[0] /= Denominator;

      for (j = 1; j < E_NDIMS; j++)
      {
         Factor = Factor0;
         x[j] = Factor*pCurve->CntlPoints[0][j]*pCurve->CntlPoints[0][E_NDIMS];

         for (l = 1; l <= pCurve->Degree; l++)
         {
            Factor *= tFactor*(pCurve->Degree - l + 1)/l;
            x[j] += Factor*pCurve->CntlPoints[l][j]
                          *pCurve->CntlPoints[l][E_NDIMS];
         }

         x[j] /= Denominator;
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Example:


#include <math.h>
#include <e_lib.h>


   MESH *mesh;
   double x[E_NDIMS];

   e_GetMeshData(&mesh,"geom");
   e_GetGeomModel(mesh,"geom");
   SetCurvePoint(&mesh->Geom.NurbsCurves[0],0.5,x);
   printf("%20.8f  %20.8f  %20.8f\n",x[0],x[1],x[2]);


*/
/* / //////////////////////////////////////////////////////////////////////// */



