/********************************************************************
 etliCmdMesh.c --
      This file implements etli_mesh commands which give acces to
      MESH type objects from E-lib library. 
      
 ******************************************************************** 
** Copyright (C) 1999 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA 02111-1307, USA.
 ***********************************************************************
*/


 /* $Id: etliCmdMesh.c,v 1.1.1.1 2002/08/04 13:19:11 peteri Exp $ */

#include "etliInt.h"
#include "etliCmdMesh.h"

/* These procedures has been moved to separate file */
extern CmdReturn*  EtliMeshObjInfo (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);
              
extern CmdReturn* EtliMeshObjGet (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);

extern CmdReturn* EtliMeshObjPut (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);

extern CmdReturn* EtliMeshObjDraw (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);
/* Prototypes for procedures refernced in this file only */
static int EtliMeshCmd (ClientData clientData, Tcl_Interp *interp,
             int objc, Tcl_Obj *CONST objv[]);

static int EtliMeshObjCmd (ClientData clientData, Tcl_Interp *interp, 
                int objc, Tcl_Obj *CONST objv[]);

static void EtliMeshObjDelete (ClientData clientData);



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Etli_InitMeshCmd -- 

 Results:
         TCL_OK

 Side effects:
         Create etli_mesh command. 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        int        
Etli_InitMeshCmd (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "etli_mesh", EtliMeshCmd, 
                      (ClientData) NULL,  (Tcl_CmdDeleteProc *) NULL);

  return TCL_OK; 
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliMeshCmd -- 
             Create mesh object and accociated Tcl command
 Results:           
         The results in normally TCL_OK.
         TCL_ERROR will be returned if
           1) wrong numbers of argument is supplied
           2) allocation of new mesh object failed
         In any case apropriate error message will be left
         in result object.
         
 Side effects:
         Allocate memory of new mesh object.
         Create new Tcl command for mesh object.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        int
EtliMeshCmd (ClientData clientData, Tcl_Interp *interp,
             int objc, Tcl_Obj *CONST objv[])
{
  char name[20];  /* for name of created Tcl command */
  char *usage = NULL;
  static int id = 0; /* for unique command names */

  Tcl_Obj *resultObjPtr;  
  MeshObj *meshObjPtr;  

  /* get object for returning result */
  resultObjPtr = Tcl_GetObjResult(interp); 

  /* mesh command must be  called without any argument */
  if (objc != 1)
    {
      Tcl_WrongNumArgs (interp, 1, objv, usage);
      return TCL_ERROR; 
    }    

  /* create new mesh object */
  /* Use Tcl memory allocators to make possible monitor memory usage */ 
  meshObjPtr = (MeshObj *) Tcl_Alloc (sizeof (MeshObj)); 
  if ( meshObjPtr == NULL)
    {
      char error[40];
      sprintf (error, "Cannot alocate memory for mesh object");
      Tcl_SetStringObj (resultObjPtr, error, -1);
      return TCL_ERROR;
    }

  /* mesh member of meshObjPtr is not initialized here 
   * initialization of this pointer is done by get subcommand
   * All other commands chech if MESH pointer is initialized
   * and if not return error
   */  
  meshObjPtr->mesh = NULL;

  /* create name of mesh command associated with mesh object */
  sprintf (name, "etliMesh%d", id); 

  id++;
  
  Tcl_CreateObjCommand (interp, name, EtliMeshObjCmd, 
                  (ClientData) meshObjPtr, EtliMeshObjDelete);
  
  Tcl_SetStringObj (resultObjPtr, name, -1); 
  return TCL_OK; 
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliMeshObjCmd -- 
         Parse and process requests to mesh object
         
 Results:        
         The return value is normally TCL_OK.
         TCL_ERROR can be returned if command parse error ocures
         or request return error condition.

 Side effecs:
         Depending on the request.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static int
EtliMeshObjCmd (ClientData clientData, Tcl_Interp *interp, 
                int objc, Tcl_Obj *CONST objv[])
{
  int index;   
  char *subCmds[] = {"get", "put", "info", "draw", "delete", NULL};
  enum subCmdsIx {GetIx, PutIx, InfoIx, DrawIx, DeleteIx};  
  
  /* subcommands usage min. and maix. no. of arguments */
  static cmdDefinition definitions[5] = {
                         {"get -option filename", 4, 4},
                         {"put -option filename", 4, 4},
                         {"info -option ?args...?", 3, 6}, 
                    {"draw canvasname bb1 bb2 bb3 bb4 -option ?arg?", 8, 24},
                         {"delete", 2, 2} };
  int result = TCL_OK; /* return status */ 
  Tcl_Obj *resultObjPtr = NULL;  
  CmdReturn *returnStruct = NULL; /* structure for recieving return from
                                     subcommands */

  /* Check if at least a subcommand is provided 
   * If not this code can immediately  return as a error message is 
   * already in Tcl interpreter
   */
  if (objc < 2)
    {
      Tcl_WrongNumArgs (interp, 1, objv, "option ?arg ...?"); 
      return TCL_ERROR; 
    }
  
  /* Finf subcommand in the list of valid subcommands */
  if (Tcl_GetIndexFromObj (interp, objv[1], subCmds, "subcommand", 
                          TCL_EXACT, &index) != TCL_OK)
    {
      return TCL_ERROR;
    }

  /* Check that the argument count matches what is expected
   * for this cubcommand
   */  
  if ((objc < definitions[index].minArgCnt) ||
      (objc > definitions[index].maxArgCnt))
    {
      Tcl_WrongNumArgs (interp, 1, objv, definitions[index].usage); 
      return TCL_ERROR; 
    }      
  /* The subcommand is recognized and has a valid number of arguments.
   * Process the subcommand.
   */  
  switch (index)
    {
      case GetIx:
        {
          returnStruct = EtliMeshObjGet (clientData, interp, objc, objv);
          break;
        }
      case PutIx:
        {
          returnStruct = EtliMeshObjPut (clientData, interp, objc, objv);
          break;
        }
      case InfoIx:
        {
          returnStruct = EtliMeshObjInfo (clientData, interp, objc, objv);
          break;
        }
      case DrawIx:
        {
          returnStruct = EtliMeshObjDraw (clientData, interp, objc, objv);
          break;
        }  
      case DeleteIx:
        {
          Tcl_DeleteCommand (interp, Tcl_GetStringFromObj (objv[0], NULL));
          break;
        }          
      default:          
        {
          char error [100]; 
          snprintf (error, 99, 
                    "Bad sub-command %s. Has no switch entry",
                     Tcl_GetStringFromObj (objv[1], NULL)); 
          resultObjPtr = Tcl_NewStringObj (error, -1); 
          result = TCL_ERROR; 
        }
    } /* end of switch(index) */

  /* Extract an object to return from returnStruct */
  if (returnStruct != NULL)
    {
      resultObjPtr = returnStruct->object;
      result = returnStruct->status;
      Tcl_Free ((char *) returnStruct); 
    }
  
  /* Set the return value */
  if (resultObjPtr != NULL)
    {
      Tcl_SetObjResult (interp, resultObjPtr);
    }
  
  return result;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliMeshObjDelete --
         Delete mesh object and associated Tcl command

 Results:        
         NONE
 
 Side effects:
         Free memory allocated for mesh object.
         Delete Tcl command associated with mesh object. 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        void
EtliMeshObjDelete (ClientData clientData)
{
  MeshObj *meshObjPtr = (MeshObj *) clientData; 
   
  /* Free memory allocated for MESH type member of meshObjPtr */
  if (meshObjPtr->mesh != NULL)
    {
      e_FreeMeshData (meshObjPtr->mesh); 
    }

  /* Free meshObjPtr itself */
  /* meshObjPtr was allocated with Tcl allocator */ 
  Tcl_Free ((char *) meshObjPtr);
}

/* Modifications:
 * $Id: etliCmdMesh.c,v 1.1.1.1 2002/08/04 13:19:11 peteri Exp $
 */ 
