/********************************************************************
 etliCmdMeshDraw.c --
      This file implements draw subcommand for mesh objects. 
      
 ******************************************************************** 
** Copyright (C) 1999 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA 02111-1307, USA.
 ***********************************************************************
*/


 /* $Id: etliCmdMeshDraw.c,v 1.1.1.1 2002/08/04 13:19:11 peteri Exp $ */

#include <assert.h>
#include "etliInt.h"
#include "etliCmdMesh.h" /* for declaration of mesh internal procedures */
#include "mesh_supp.h"

/* Prototypes for procedures refernced in this file only */
static int mapBBox (double *wbbox, double *vbbox, double *mapping);

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 mapBBox --
 Description:
      Calculate coefficients of mapping between mesh coordinate
      system and vieport coordinate system using mesh bounding box
      and vieport bounding box.
 Arguments:      
      Arguments to this function are arrays of the following length
      wbbox - 6 (x1, y1, z1, x2, y2, z2)
      wbbox - 4 (x1, y1, x2, y2)
      mapping - 4 (A1, A2, B1, B2) where mapping is defined:
                 x' = A1 * x + A2
                 y' = B1 * y + B2
      Mapping is constructed in such way that aspect ratio of wbbox             
      is not changed after mapping.
 Result:     
     If no error returns 1 and mapping coefficients are in mapping 
     array. On error return 0. Error will occur if width or height
     of vieport bounding box is smaller than 1.
 Side effects:
     None 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static int
mapBBox (double *wbbox, double *vbbox, double *mapping)          
{
  static double worldRes = 1.0e-6; /* resolution for world coords */  
  static double vieportRes = 1.0; /* resolution for canvas */ 
  double heightW, widthW; /* world (mesh) height and width */
  double heightV, widthV; /* vieport (canvas) height and width */
    
  heightW = wbbox[4] - wbbox[1]; 
  widthW = wbbox[3] - wbbox[0]; 
  heightV = vbbox[3] - vbbox[1]; 
  widthV = vbbox[2] - vbbox[0]; 
  
  /* check the vieport bounding box */
  if ( (fabs (heightV) < vieportRes) || fabs (widthV) < vieportRes)
    {
      return 0;
    }
  /* check the world bounding box */
  if ( fabs (heightW) < worldRes )
    {
      mapping[2] = HUGE;
    }
  else
    {
      mapping[2] = heightV / heightW;
    }
  if ( fabs(widthW) < worldRes )
    {
      mapping[0] = HUGE;
    }
  else
    {
      mapping[0] = widthV / widthW;
    }
  /* coordinates in world are assumed to be:   | 
   *                                           o__
   * and coordinates in vieport (Tk canvas are o__
   *                                           |    
   * thus wyou have the equations beelow
   */   
  mapping[0] = -1.0 * E_MIN(mapping[0], mapping[2]);
  mapping[2] = mapping[0];
  mapping[1] = 0.5 * (vbbox[0]+vbbox[2] - mapping[0]*(wbbox[0]+wbbox[3]));
  mapping[3] = 0.5 * (vbbox[1]+vbbox[3] - mapping[2]*(wbbox[1]+wbbox[4]));
                 
  return 1;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 DrawMesh --
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define _CASE_ELEM(name1, name2) \
        {\
          llength = E_NVPE_##name1; \
          nodes =  meshObjPtr->mesh->Nodes##name2; \
          break; \
        } 

        static void
DrawMesh ( ClientData clientData, Tcl_Interp *interp, int objc, 
           Tcl_Obj *CONST objv[], int *result, Tcl_Obj **resultObjPtr)
{
  register int node;  
  register int i;
  register int j;
  double mapping[4];  /* mapping coeficients */
  char error[80];  
  Tcl_Obj **drawCmdList;
  Tcl_Obj *tag[4];  /* for tag items  for every element :
                         * meshname -> etliMesh%
                         * element number -> elem%
                         */   
  Tcl_Obj *tagOption;  /* for holding -tag option */
  int llength, element; 

  MeshObj *meshObjPtr = (MeshObj *) clientData; 
  MESH *mesh = meshObjPtr->mesh; /* convinience variable */

  double bbox[4]; 
  int *nodesList;

  /* get vieport bounding box coordinates from argument list */
  for (i=0; i<4; i++)
    {
      if (Tcl_GetDoubleFromObj (interp, objv[3+i], bbox+i) != TCL_OK)
        {
          *result = TCL_ERROR;
          return;
        }
    } 
  /* allocate memory for draw commnad list */        
  drawCmdList = (Tcl_Obj **) Tcl_Alloc ((2 * 8 + 20)* sizeof (Tcl_Obj *));

  /* calculate mapping coefficients */
  if ( mapBBox (meshObjPtr->bBox, bbox, mapping) != 1 )
    {
      sprintf (error, 
            "Invalid bounding box: must have nonzero width and height"); 
      *resultObjPtr = Tcl_NewStringObj (error, -1);
      *result = TCL_ERROR;
      return;
    }
  
  tagOption = Tcl_NewStringObj ("-tag", -1);
  tag[0] = objv[0];

  drawCmdList[0] = objv[2];
  drawCmdList[1] = Tcl_NewStringObj ("create", -1);
  drawCmdList[2] = Tcl_NewStringObj ("polygon", -1);
  for (element=1; element<=mesh->TotalNElems; element++)
    {
      RP_MeshElementVertices (mesh, element, &nodesList, &llength);  
      for (i=0; i<llength; i++)
        {
          node = nodesList[i]-1; 
          *(drawCmdList+3+2*i) = Tcl_NewDoubleObj 
               (mapping[0] * (*mesh->MeshPointCoords[node])+mapping[1]);       
          *(drawCmdList+4+2*i) = Tcl_NewDoubleObj 
               (mapping[2] * (*(mesh->MeshPointCoords[node]+1))+mapping[3]);       
        }
      i=3+2*llength; 
      /* CAUTION: a table error[80] will be used to create tag strings for tag
       * list
       */
      snprintf (error, 79, "elem%d", element); 
      tag[1] = Tcl_NewStringObj (error, -1);

      /* add tag option and tags list to polygon draw command */
      drawCmdList[i++] = tagOption;
      drawCmdList[i++] = Tcl_NewListObj (2, tag); 
              
      for (j=8; j<objc; j++)
        {
          drawCmdList[i++] = objv[j];
        }
      Tcl_EvalObjv (interp, i, drawCmdList, TCL_EVAL_DIRECT); 
    }
  Tcl_Free ((char *) drawCmdList);

  return;
}  
/*
 
  nodescoord = mesh->MeshPointCoords;
  nj = myobjv+3;
  nj1 = myobjv+4;
  for (element=1; element<=mesh->TotalNElems; element++)
    {
      RP_MeshElementVertices (mesh, element, &nodesList, &llength);  
      for (i=0; i<llength; i++)
        {
          node = nodesList[i]-1; 
          i2 = 2*i;
             *(nj+i2) = Tcl_NewDoubleObj 
               (206.29 * (*mesh->MeshPointCoords[node])-32.14);       
             *(nj1+i2) = Tcl_NewDoubleObj 
               (544.13* (*(mesh->MeshPointCoords[node]+1))-49.85);       
           }
         llength2=2*llength;
         myobjv[3+llength2] = outline;
         myobjv[4+llength2] = color;
         myobjv[5+llength2] = fill;
         myobjv[6+llength2] = fillcolor;
         myobjv[7+llength2] = outlinewidth;
         myobjv[8+llength2] = width;
         Tcl_EvalObjv (interp, 2*llength+9, myobjv, TCL_EVAL_DIRECT); 
       }
 
  if (meshObjPtr->elemType == ETLI_MIXED)
    {
     for (element=1; element<=mesh->TotalNElems; element++)
       {
         RP_MeshElementVertices (mesh, element, &nodesList, &llength);  
         for (i=0; i<llength; i++)
           {
             node = nodesList[i]-1; 
             myobjv[3+2*i] = Tcl_NewDoubleObj 
               (206.29 * (*mesh->MeshPointCoords[node])-32.14);       
             myobjv[3+2*i+1] = Tcl_NewDoubleObj 
               (544.13* (*(mesh->MeshPointCoords[node]+1))-49.85);       
           }
         myobjv[2+2*llength+1] = outline;
         myobjv[2+2*llength+2] = color;
         myobjv[2+2*llength+3] = fill;
         myobjv[2+2*llength+4] = fillcolor;
         myobjv[2+2*llength+5] = outlinewidth;
         myobjv[2+2*llength+6] = width;
         Tcl_EvalObjv (interp, 2*llength+9, myobjv, TCL_EVAL_DIRECT); 
       }
    }
  else
    {
      switch (meshObjPtr->elemType)
        {
          case E_ELEM_TYPE_LINK1: 
            _CASE_ELEM(LINK1, Link1)
          case E_ELEM_TYPE_LINK2: 
              _CASE_ELEM(LINK2, Link2)
          case E_ELEM_TYPE_LINK3: 
              _CASE_ELEM(LINK3, Link3)
          case E_ELEM_TYPE_LINK4: 
              _CASE_ELEM(LINK4, Link4)
          case E_ELEM_TYPE_LINK5: 
              _CASE_ELEM(LINK5, Link5)
          case E_ELEM_TYPE_TRIANG1: 
              _CASE_ELEM(TRIANG1, Triang1)
          case E_ELEM_TYPE_TRIANG2: 
              _CASE_ELEM(TRIANG2, Triang2)
          case E_ELEM_TYPE_TRIANG3: 
              _CASE_ELEM(TRIANG3, Triang3)
          case E_ELEM_TYPE_TRIANG4: 
              _CASE_ELEM(TRIANG4, Triang4)
          case E_ELEM_TYPE_TRIANG5: 
              _CASE_ELEM(TRIANG5, Triang5)
          case E_ELEM_TYPE_QUAD1: 
              _CASE_ELEM(QUAD1, Quad1)
          case E_ELEM_TYPE_QUAD3: 
              _CASE_ELEM(QUAD3, Quad3)
          case E_ELEM_TYPE_TETRAH1: 
              _CASE_ELEM(TETRAH1, Tetrah1)
          case E_ELEM_TYPE_BLOCK1: 
              _CASE_ELEM(BLOCK1, Block1)
        } 
         myobjv[2+2*llength+1] = outline;
         myobjv[2+2*llength+2] = color;
         myobjv[2+2*llength+3] = fill;
         myobjv[2+2*llength+4] = fillcolor;
         myobjv[2+2*llength+5] = outlinewidth;
         myobjv[2+2*llength+6] = width;
     for (element=0; element<mesh->TotalNElems; element++)
       {
         for (i=0; i<llength; i++)
           {
             node = *(nodes[element]+i)-1; 
             myobjv[3+2*i] = Tcl_NewDoubleObj 
               (206.29 * (*mesh->MeshPointCoords[node])-32.14);       
             myobjv[3+2*i+1] = Tcl_NewDoubleObj 
               (544.13* (*(mesh->MeshPointCoords[node]+1))-49.85);       
           }
         Tcl_EvalObjv (interp, 2*llength+9, myobjv, TCL_EVAL_DIRECT); 
       }
    } */ 

/* meshDrawProc is an array of pointer to Info* functions */
static void (*meshDrawProc[]) (ClientData clientData, 
               Tcl_Interp *interp, 
               int objc, 
               Tcl_Obj *CONST objv[],
               int *result,
               Tcl_Obj **resultObjPtr)={DrawMesh,
                                       }; 

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliMeshObjDraw -- 
        syntax: 
        etliMesh# draw .canvas bbox1 bbox2 bbox3 bbox4 -option ?args...?
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        CmdReturn*
EtliMeshObjDraw (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[])
{
  int index;
  static char *cmdOpts[2]={"-mesh", NULL};
  int result = TCL_OK; 
  CmdReturn *returnStructPtr = NULL;
  Tcl_Obj *resultObjPtr = NULL;

  MeshObj *meshObjPtr = (MeshObj *) clientData; 
  MESH *mesh = meshObjPtr->mesh; /* convinience variable */

  /* allocate the space for the return structure */
  /* Use Tcl memory allocators to make possible monitor memory usage */ 
  returnStructPtr = (CmdReturn *) Tcl_Alloc (sizeof (CmdReturn));


  /* check if mesh has been initialized */
  if (mesh == (MESH*) NULL)   
    {
      resultObjPtr = Tcl_NewStringObj 
          ("mesh not initialized; use \"get -mesh\" to initialize", -1); 
      result = TCL_ERROR; 
      goto done;
    }
  /* checking if canvas name is valid is done in each drowing procedure
   * separatelly
   */  
  /* find option in the list of valid options for this subcommad */
  if (Tcl_GetIndexFromObj (interp, objv[7], cmdOpts, 
                          "options", TCL_EXACT, &index) != TCL_OK)
    {
      result = TCL_ERROR;
      goto done;
    }

  /* calling appropriate info function */
  meshDrawProc[index](clientData, interp, objc, objv, 
                      &result, &resultObjPtr); 
  done:
  /* initialize return structure */
     returnStructPtr->object = resultObjPtr;
     returnStructPtr->status = result;

  return returnStructPtr;
}

/* Modifications:
 * $Id: etliCmdMeshDraw.c,v 1.1.1.1 2002/08/04 13:19:11 peteri Exp $
 */ 
