/***********************************************************************
 etliCmdNurbs.c --
      This file implements etli_nurbs command which give acces to
      NURBSCURVE type objects from E-Lib library.

 ***********************************************************************
** Copyright (C) 1999 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA 02111-1307, USA.
 ***********************************************************************
*/

/* $Id: etliCmdNurbs.c,v 1.1.1.1 2002/08/04 13:19:11 peteri Exp $ */

#include "etliInt.h"
#include "nurbs_supp.h" /* for NURBSCURVE interface */

/* Two global MESH structures.
 * The structure meshSingle is used to to output single NURBS curve
 * to file while meshCollect is used to output a set of NURBS curves
 * 
 * Asside comments -- the whole NURBS support in E-Lib library should
 * be rewriten
 */
static MESH meshSingle_Str;
static MESH meshCollect_Str;
static MESH *meshSingle = &meshSingle_Str;
static MESH *meshCollect = &meshCollect_Str;
static NURBSCURVE **nurbsCollect;

/* string array for error messages */
static char errorMessage[200];

/* nurbs utility from Jelle Muylle */
extern void crv_getcoords   (NURBSCURVE *curve,double t,double *crds);
/* anothe nurbs utility from Jelle Muylle */
/* This was done only for testing purposes */
/*extern void SetCurvePoint (NURBSCURVE *curve,double t,double *crds); */

/* Prototypes for procedures referenced in this file only */
static int EtliNurbsCmd (ClientData clientData, Tcl_Interp *interp,
             int objc, Tcl_Obj *CONST objv[]);

static int EtliSaveNurbsCmd (ClientData clientData, Tcl_Interp *interp,
             int objc, Tcl_Obj *CONST objv[]);

static int EtliSetEndpointsCmd (ClientData clientData, Tcl_Interp *interp,
             int objc, Tcl_Obj *CONST objv[]);

static int EtliNurbsObjCmd (ClientData clientData, Tcl_Interp *interp, 
                int objc, Tcl_Obj *CONST objv[]);

static void EtliNurbsObjDelete (ClientData clientData);

static CmdReturn* EtliNurbsObjGet (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);

static CmdReturn* EtliNurbsObjSet (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);

static CmdReturn* EtliNurbsObjPut (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);

static CmdReturn* EtliNurbsObjDraw (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);

static CmdReturn*  EtliNurbsObjInfo (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);

static CmdReturn*  EtliNurbsObjMesh (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[]);

static void InfoAll (ClientData clientData, Tcl_Interp *interp, int objc, 
              Tcl_Obj *CONST objv[], int *result, Tcl_Obj **resultObjPtr);

static void InfoControlPoint (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[], int *result,
              Tcl_Obj **resultObjPtr);

static void InfoCoords (ClientData clientData, Tcl_Interp *interp, int objc, 
              Tcl_Obj *CONST objv[], int *result, Tcl_Obj **resultObjPtr);

static void SetAll (ClientData clientData, Tcl_Interp *interp, int objc, 
              Tcl_Obj *CONST objv[], int *result, Tcl_Obj **resultObjPtr);


static void AddNurbsToMesh (ClientData clientData, Tcl_Interp *interp, int objc, 
              Tcl_Obj *CONST objv[], int *result, Tcl_Obj **resultObjPtr); 
static void RemoveNurbsFromMesh (ClientData clientData, Tcl_Interp *interp, 
            int objc, Tcl_Obj *CONST objv[], int *result, Tcl_Obj **resultObjPtr);  
static void InitSingleMesh (void);
void InitCollectMesh (MESH *mesh); 

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Etli_InitNurbsCmd -- 

 Results:
         TCL_OK

 Side effects:
         Create etli_nurbs command. 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        int        
Etli_InitNurbsCmd (Tcl_Interp *interp)
{
  Tcl_CreateObjCommand (interp, "etli_nurbs", EtliNurbsCmd, 
                     (ClientData) NULL,  (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateObjCommand (interp, "etli_save_nurbs", EtliSaveNurbsCmd, 
                     (ClientData) NULL,  (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateObjCommand (interp, "etli_nurbs_set_endpoints", EtliSetEndpointsCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  return (TCL_OK); 
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliNurbsCmd -- 
             Create nurbs object and associated Tcl command
 Results:           
         The results in normally TCL_OK.
         TCL_ERROR will be returned if
           1) wrong numbers of argument is supplied
           2) allocation of new nurbs object failed
         In any case apropriate error message will be left
         in result object.
         
 Side effects:
         Allocate memory of new nurbs object.
         Create new Tcl command for nurbs object.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static int
EtliNurbsCmd (ClientData clientData, Tcl_Interp *interp,
             int objc, Tcl_Obj *CONST objv[])
{
  char name[30];  /* for name of created Tcl command */
  char nurbsname[30];  /* for name of the nurbs curve */
  char *usage = "degree ep1 ep2 x1 y1 z1 w1 ... xn, yn, zn, wn";
  static int id = 0;  /* for unique command names */

  Tcl_Obj *resultObjPtr;  
  NurbsObj *nurbsObjPtr;  
  NURBSCURVE *nurbs; /* convinience variable */
  int degree;
  int i,j;

  /* get result object to report errors */
  resultObjPtr = Tcl_GetObjResult(interp);
  
  /* check if at least degree supplied */
  if (objc < 2 )
    {
      Tcl_WrongNumArgs (interp, 1, objv, usage);
      return (TCL_ERROR); 
    }    
  /* extract drgree from command arguments */ 
  if (Tcl_GetIntFromObj (interp, objv[1], &degree) != TCL_OK)
    {
      return (TCL_ERROR);
    } 

  /* check if degree is in range 1...ETLI_MAX_NURBS_DEGREE */
  if (degree < 1 || degree > ETLI_MAX_NURBS_DEGREE)
    {
      snprintf (errorMessage, 79, 
               "Invalid degree value; only degrees 1-%d supported",
                ETLI_MAX_NURBS_DEGREE);
      Tcl_SetStringObj (resultObjPtr, errorMessage, -1); 
      return (TCL_ERROR);
    }

  /* check if proper number of parameters supplied: 
   * 1 - command name
   * 1 - degree
   * 2 - end points indices
   * 4 * (degree+1) coordeinates (x,y,z) and weight of control points
   */  
  if ( objc != (1 + 1 + 2 + (degree+1)*4))
    {
      Tcl_WrongNumArgs (interp, 1, objv, usage);  
      return (TCL_ERROR); 
    }    

  nurbsObjPtr = (NurbsObj *) Tcl_Alloc(sizeof (NurbsObj)); 

  if (nurbsObjPtr == NULL) 
    {
      snprintf (errorMessage, 39, "Cannot alocate memory for nurbs object");
      Tcl_SetStringObj (resultObjPtr, errorMessage, -1); 
      return (TCL_ERROR);
    }

  /*  initialize nurbsObjPtr->nurbs to empty object */
  nurbsObjPtr->nurbs = RP_InitNurbs();
  RP_EmptyNurbs(nurbsObjPtr->nurbs, degree); 

  nurbs = nurbsObjPtr->nurbs;  /* only for easier calling */
  /* fill nurbs structure */
   
  /* fill created nurbs object using command arguments */
  if (Tcl_GetIntFromObj (interp, objv[2], 
             &(nurbs->EndPoint1)) != TCL_OK) return (TCL_ERROR);
  if (Tcl_GetIntFromObj (interp, objv[3], 
             &(nurbs->EndPoint2)) != TCL_OK) return (TCL_ERROR);
  for (i=0; i<degree+1; i++)
    for (j=0; j<4; j++)         
      {
       if (Tcl_GetDoubleFromObj (interp, objv[1+1+2+i*4+j], 
             *(nurbs->CntlPoints)+4*i+j) != TCL_OK) return (TCL_ERROR);
      }   
  
  /* create name of Tcl command associated with nurbs object */
  sprintf (name, "etliNurbs%d", id); 
  sprintf (nurbsname, "etliNurbs%04d", id); 

  
  /* create name of the nurbs object */
  nurbs->Name = (char *) Tcl_Alloc(sizeof (char) * (strlen (nurbsname)+1));
  if (nurbs->Name == NULL) 
    {
      snprintf (errorMessage, 39, "Cannot alocate memory for nurbs name");
      Tcl_SetStringObj (resultObjPtr, errorMessage, -1); 
      return (TCL_ERROR);
    }
  strcpy (nurbs->Name, nurbsname);
  
  id++;
   
  Tcl_CreateObjCommand (interp, name, EtliNurbsObjCmd, 
                  (ClientData) nurbsObjPtr, EtliNurbsObjDelete);
  
  Tcl_SetStringObj (resultObjPtr, name, -1); 
  return (TCL_OK); 
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliSaveNurbsCmd -- 
       Save collection of NURBS added to mesh.

 Results:           
         The results is normally TCL_OK.
         TCL_ERROR will be returned if
           1) wrong numbers of argument is supplied
         In any case apropriate error message will be left
         in result object.
         
 Side effects:
         Save NURBS to file .
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static int
EtliSaveNurbsCmd (ClientData clientData, Tcl_Interp *interp,
             int objc, Tcl_Obj *CONST objv[])
{
  char *usage = "file";
  char *fileName;
  int i;

  Tcl_Obj *resultObjPtr;  
 

  /* get result object to report errors */
  resultObjPtr = Tcl_GetObjResult(interp);
  
  /* check if file name was supplied */
  if (objc != 2 )
    {
      Tcl_WrongNumArgs (interp, 1, objv, usage);
      return (TCL_ERROR); 
    }    

  /* extract file name for command argumens */ 
  fileName = Tcl_GetStringFromObj (objv[1], NULL); 
  
  /* check if there are any nurbs to save */
  if (meshCollect->Geom.NNurbsCurves == 0)
    {
      snprintf (errorMessage, 79, "Warning: No NURBS to save");
      Tcl_SetStringObj (resultObjPtr, errorMessage, -1); 
      return (TCL_OK);
    }

  /* free old structures from meshCollect */
  e_free (meshCollect->Geom.NurbsCurves);

  /* allocate memory for new structures */
  meshCollect->Geom.NurbsCurves = 
                    (NURBSCURVE*) 
           e_calloc (meshCollect->Geom.NNurbsCurves, sizeof (NURBSCURVE));
  
  for (i=0; i<meshCollect->Geom.NNurbsCurves; i++)
  {
     memcpy (&(meshCollect->Geom.NurbsCurves[i]), nurbsCollect[i], 
                     sizeof (NURBSCURVE));
  } 
  /* e-Lib functions will "explode" on error so there is no point 
    to do this at the moment -- SOMEON HAS TO FIX IT
   */
  if (e_PutGeomModel (meshCollect, fileName) != 1)
  {
    sprintf(errorMessage, "Writing NURBS to file %s failed", 
                             fileName); 
    Tcl_SetStringObj (resultObjPtr, errorMessage, -1); 
    return (TCL_ERROR);
  } 

  e_free  (nurbsCollect); 
  return (TCL_OK); 
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliSetEndpointsCmd -- 
         Set the EndPointsCoords array in Geom strucure in meschCollect.
         This function is ad hoc solution to the problem of savin
         NURBS objects. 
         
 Results:           
         The results is normally TCL_OK.
         TCL_ERROR will be returned if
           1) wrong numbers of argument is supplied
         In any case apropriate error message will be left
         in result object.
         
 Side effects:
          Allocat if necesar and set element of 
          meshCollect->Geom.EndPointCoords, 
          set  meshCollect->Geom.NNurbsCurves.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static int
EtliSetEndpointsCmd (ClientData clientData, Tcl_Interp *interp,
             int objc, Tcl_Obj *CONST objv[])
{
  char *usage = "NEndPoints pointNumber x-coord y-coord";
  int nEndPoints, pointNumber;
  double xcoord, ycoord;

  Tcl_Obj *resultObjPtr;  
 

  /* get result object to report errors */
  resultObjPtr = Tcl_GetObjResult(interp);
  
  /* check if file name was supplied */
  if (objc != 5 )
    {
      Tcl_WrongNumArgs (interp, 1, objv, usage);
      return (TCL_ERROR); 
    }    

  /* extract total number of end points */ 
  if (Tcl_GetIntFromObj (interp, objv[1], 
             &nEndPoints) != TCL_OK) return (TCL_ERROR);

  /* check the range */
  if ( nEndPoints < 1)
    {
      snprintf (errorMessage, 79, 
         "Error: Invalid total number of end points %d (should be > 0", 
          nEndPoints);
      Tcl_SetStringObj (resultObjPtr, errorMessage, -1); 
      return (TCL_OK);
    }
  /* allocate memory for end points */
  if (nEndPoints > meshCollect->Geom.NEndPoints) 
  {
     meshCollect->Geom.EndPointCoords = e_ext_cont_d_matr (
                meshCollect->Geom.NEndPoints, 3,  
              meshCollect->Geom.EndPointCoords, 
              nEndPoints -  meshCollect->Geom.NEndPoints);
    meshCollect->Geom.NEndPoints =  nEndPoints; 
  }
  /* extract the point number */
  if (Tcl_GetIntFromObj (interp, objv[2], 
             &pointNumber) != TCL_OK) return (TCL_ERROR);

  /* check pointNumber range */
  if ( pointNumber < 1 || pointNumber > meshCollect->Geom.NEndPoints)
    {
      snprintf (errorMessage, 79, 
         "Error: Point number %d out of range [1, %d]", 
          pointNumber, meshCollect->Geom.NEndPoints );
      Tcl_SetStringObj (resultObjPtr, errorMessage, -1); 
      return (TCL_OK);
    }
  /* get point x and y coordinates */
  if (Tcl_GetDoubleFromObj (interp, objv[3], 
             &xcoord) != TCL_OK) return (TCL_ERROR);
  if (Tcl_GetDoubleFromObj (interp, objv[4], 
             &ycoord) != TCL_OK) return (TCL_ERROR);

  /* set the endpoint coordnates */
  meshCollect->Geom.EndPointCoords[pointNumber -1][0] = xcoord;           
  meshCollect->Geom.EndPointCoords[pointNumber -1][1] = ycoord;           
  meshCollect->Geom.EndPointCoords[pointNumber -1][2] = 0.0;           

  return (TCL_OK); 
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliNurbsObjCmd -- 
         Parse and process requests to nurbs object
         
 Results:        
         The return value is normally TCL_OK.
         TCL_ERROR can be returned if command parse error ocures
         or request return error condition.

 Side effecs:
         Depending on the request.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static int
EtliNurbsObjCmd (ClientData clientData, Tcl_Interp *interp, 
                int objc, Tcl_Obj *CONST objv[])
{
  int index;   
  static char *subCmds[] = {"put", "info", "set", "delete", "mesh", NULL};
  enum subCmdsIx {PutIx, InfoIx, SetIx, DeleteIx, MeshIx};  

  /* subcommand usage, min and max no. of arguments */
  static cmdDefinition definitions[5] = {
                            {"put filename", 3, 3},
                            {"info key ?arg?", 3, 4}, 
                            {"set key value ?option value?", 4, 
                              6+4*(1+ETLI_MAX_NURBS_DEGREE)},  
                            {"delete", 2, 2},
                            {"mesh key", 3, 3}}; 
  int result = TCL_OK;   /* return status */
  Tcl_Obj *resultObjPtr = NULL;
  CmdReturn *returnStruct = NULL; /* structure used for passing return form
                                   subcommnads */

  /* Check if at least a subcommand is provided 
   * If not this code can immediately  return as a error message is 
   * already in Tcl interpreter
   */
  if (objc < 2)
    {
       Tcl_WrongNumArgs (interp, 1, objv, "option ?arg ...?"); 
       return (TCL_ERROR); 
    }

  /* Find subcommand in the list of valid subcommands */
  if (Tcl_GetIndexFromObj (interp, objv[1], subCmds, "subcommand", TCL_EXACT, 
                          &index) != TCL_OK)
    {
       return (TCL_ERROR);
    }

  /* Check that the argument count matches 
   *  what is expected for this subcommand
   */
  if ((objc < definitions[index].minArgCnt) ||
      (objc > definitions[index].maxArgCnt))
     {
       Tcl_WrongNumArgs (interp, 1, objv, definitions[index].usage); 
       return (TCL_ERROR); 
     }      
         
  /* The subcommand is recognized, and has a valid number of arguments.
   *  Process the command.   
   */  
  switch (index)
    {
      case PutIx:
        {
          returnStruct = EtliNurbsObjPut (clientData, interp, objc, objv); 
          break;
        }          
      case InfoIx:
        {
          returnStruct = EtliNurbsObjInfo (clientData, interp, objc, objv); 
          break;
        }
      case SetIx:
        {
          returnStruct = EtliNurbsObjSet (clientData, interp, objc, objv); 
          break;
        }
      case DeleteIx:
        {
          Tcl_DeleteCommand (interp, Tcl_GetStringFromObj (objv[0], NULL));
          break;
        }          
      case MeshIx:
        {
          returnStruct = EtliNurbsObjMesh (clientData, interp, objc, objv);
          break;
        }
      default:          
        {
          snprintf (errorMessage, 99, 
                    "Bad sub-command %s. Has no switch entry",
                     Tcl_GetStringFromObj (objv[1], NULL)); 
          resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
          result = TCL_ERROR; 
        }
    }

  /* Extract an object to return from returnStruct */
  if (returnStruct != NULL)
    {
      resultObjPtr = returnStruct->object;
      result = returnStruct->status;
      Tcl_Free ((char *) returnStruct); 
    } 
  
  /* Set the return value */
  if (resultObjPtr != NULL)
    {
      Tcl_SetObjResult (interp, resultObjPtr);
    }
  
  return (result);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliNurbsObjDelete --
         Delete nurbs object and associated Tcl command

 Results:        
         NONE
 
 Side effects:
         Free memory allocated for nurbs object.
         Delete Tcl command associated with nurbs object. 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static void
EtliNurbsObjDelete (ClientData clientData)
{
   NurbsObj *nurbsObjPtr = (NurbsObj *) clientData; 
    
   /* free memory allocated for NURBSCURVE member of nurbsObjPtr */
   RP_FreeNurbs (nurbsObjPtr->nurbs);
   /* free nurbsObjPtr itself */
   Tcl_Free ((char *) nurbsObjPtr); 
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliNurbsObjSet -- 
             Set nurbs object parameters
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static CmdReturn*
EtliNurbsObjSet (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[])
{
  int intValue;
  double doubleValue; 
  int index, subindex;
  static char *subOpts[] = {"-all","-ep1", "-ep2", "-ctrlpoint", "-name", NULL};
  enum subOptsIx {AllIx, Ep1Ix, Ep2Ix, CtrlIx, NameIx};
  static char *subSubOpts[] = {"-coords", "-weight", NULL};
  enum subSubOptsIx {CoordsIx, WeightIx};

  int result = TCL_OK; 
  CmdReturn *returnStructPtr = NULL;
  Tcl_Obj *resultObjPtr = NULL;

  NurbsObj *nurbsObjPtr = (NurbsObj *) clientData; 
  NURBSCURVE *nurbs = nurbsObjPtr->nurbs; /* convinience variable */

  /* allocate the space and initialize the return structure */
  returnStructPtr = (CmdReturn *)  Tcl_Alloc (sizeof (CmdReturn));

  /* find otption in the list of valid options for this subcommand */
  if (Tcl_GetIndexFromObj (interp, objv[2], subOpts, "option", TCL_EXACT, 
                          &index) != TCL_OK)
    {
       result = TCL_ERROR; 
       goto done;
    }

  switch (index) 
    {
      case AllIx:
        {       
          SetAll (clientData, interp, objc, objv, 
                          &result, &resultObjPtr);             
          break;
        }
      case Ep1Ix:
        {
          if (Tcl_GetIntFromObj (interp, objv[3], &intValue) != TCL_OK)
            {
              result = TCL_ERROR; 
              break;
            }    
          nurbs->EndPoint1 = intValue; 
          break;
        }
      case Ep2Ix:
        {
          if (Tcl_GetIntFromObj (interp, objv[3], &intValue) != TCL_OK)
            {
              result = TCL_ERROR; 
              break;
            }    
          nurbs->EndPoint2 = intValue; 
          break;
        }
      /* The case bellow should be put in separate function.
       * This is must be done in the next code upgrade 
       */  
      case CtrlIx: /* setting control points attributes */
        {
          int pointNumber; 
           if (Tcl_GetIntFromObj (interp, objv[3], &pointNumber) != TCL_OK)
             {
               result = TCL_ERROR; 
               break;
             }    
           if ( pointNumber<1 || pointNumber>nurbs->Degree+1 )
             {
               sprintf(errorMessage, "wrong point number: expected integer 1-%d", 
                             nurbs->Degree+1); 
               resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
               result = TCL_ERROR; 
               break; 
             }               
           pointNumber--; /* point are numbered from 1 but strored in C array */

           if (Tcl_GetIndexFromObj (interp, objv[4], subSubOpts, 
                               "suboption", TCL_EXACT, &subindex) != TCL_OK)
             {
               result = TCL_ERROR; break; 
             }
           if (objc == 6) /* setting weight parameters */
             {
               if (Tcl_GetDoubleFromObj (interp, objv[5], &doubleValue) != TCL_OK)
                 {
                   result = TCL_ERROR; break;
                 }    
               *(nurbs->CntlPoints[pointNumber]+3)=doubleValue; break;
             }
           else if (objc == 8) /* setting coordinates */
             {
               int i;  
               for (i=0; i<3; i++)  /* for x, y, z coord */ 
                  {
                    if (Tcl_GetDoubleFromObj (interp, objv[5+i], 
                                              &doubleValue) != TCL_OK)
                      {
                        result = TCL_ERROR; break;
                      }    
                    *(nurbs->CntlPoints[pointNumber]+i)=doubleValue;
                  }
             } 
           else /* invalid number of parameters */
             {
               Tcl_WrongNumArgs (interp, 1, objv, "wrong # args");  
               result = TCL_ERROR;
             }   
          break;
        } /* end of setting control point attributes */          
      case NameIx: /* setting NURBS title */
        {
           int length;
           char *name;
           name = Tcl_GetStringFromObj (objv[3], &length);
           nurbs->Name = Tcl_Alloc ((length+1)*sizeof (char));
           strcpy (nurbs->Name, name);
           break; 
        }
    } /* end of switch */                
  done:

  /* initialize return structure */
        returnStructPtr->object = resultObjPtr; 
        returnStructPtr->status = result; 

  return (returnStructPtr);        
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 SetAll --
          Change all parameters of existing nurbs object  
          This function does not set the nurbs title.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static void
SetAll (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[], int *result, 
              Tcl_Obj **resultObjPtr)
{
  char *usage = "-all degree ep1 ep2 x1 y1 z1 w1 ... xn, yn, zn, wn";
  
  NurbsObj *nurbsObjPtr = (NurbsObj *) clientData;  
  NURBSCURVE *nurbs = nurbsObjPtr->nurbs; /* convinience variable */
  int degree, olddegree;
  int i,j;

  /* check if proper number of argument supplied*/
  if (objc < 4 )
    {
      Tcl_WrongNumArgs (interp, 1, objv, usage);
      *result = TCL_ERROR;
      return; 
    }    
  /* extract drgree from command arguments */ 
  if (Tcl_GetIntFromObj (interp, objv[3], &degree) != TCL_OK)
    {
      *result = TCL_ERROR;
      return;
    } 
 /* check if degree is in range 1...ETLI_MAX_NURBS_DEGREE */
  if (degree < 1 || degree > ETLI_MAX_NURBS_DEGREE)
    {
      snprintf (errorMessage, 79, 
               "Invalid degree value; only degrees 1-%d supported",
                ETLI_MAX_NURBS_DEGREE);
      *resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
      *result = TCL_ERROR;
      return; 
    }

  /* check if proper number of parameters supplied: 
   * 1 - command name (etliNurbs%)
   * 1 - subcommand (info)
   * 1 - option (-all)
   * 1 - degree
   * 2 - end points indices
   * 4 * (degree+1) coordinates (x,y,z) and weight of control points
   */  
  if ( objc != (1+ 1 + 1 + 1 + 2 + (degree+1)*4))
    {
      Tcl_WrongNumArgs (interp, 1, objv, usage);  
      *result = TCL_ERROR;
      return; 
    }    

  /* check if new degree is the same as an old one
   * if not free nurbs and allocate new once again
   */
  olddegree = nurbs->Degree;
  if (olddegree != degree)
    {
      /* free memory allocated for NURBSCURVE member of nurbsObjPtr */
      RP_FreeNurbs (nurbs);
      /* allocate new nurbs object and nullify it*/
      nurbs = RP_InitNurbs();
      RP_EmptyNurbs(nurbs, degree); 
    }
  
  /* set nurbs structure */
  if (Tcl_GetIntFromObj (interp, objv[4], &(nurbs->EndPoint1)) != TCL_OK) 
    {
      *result = TCL_ERROR;
      return;
    }
  if (Tcl_GetIntFromObj (interp, objv[5], &(nurbs->EndPoint2)) != TCL_OK)
    {
      *result = TCL_ERROR; 
      return;
    }

  /* set coordinates and weights to new value */
  for (i=0; i<degree+1; i++)
    for (j=0; j<4; j++)         
      {
       if (Tcl_GetDoubleFromObj (interp, objv[6+i*4+j], 
             *(nurbs->CntlPoints)+4*i+j) != TCL_OK) 
         {
           *result = TCL_ERROR;
           return;
         }
      }   
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliNurbsObjPut --
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static CmdReturn*
EtliNurbsObjPut (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[])
{
  char *fileName;
  CmdReturn *returnStructPtr;
  Tcl_Obj *resultObjPtr=NULL;
  int result;
  NurbsObj *nurbsObjPtr = (NurbsObj *) clientData; 

  fileName = Tcl_GetStringFromObj (objv[2], NULL); 

  result = TCL_OK;
  resultObjPtr = NULL; 

  InitSingleMesh ();
  
  meshSingle->Geom.NurbsCurves = nurbsObjPtr->nurbs; 
  memcpy (meshSingle->Geom.EndPointCoords[0],
                  nurbsObjPtr->nurbs->CntlPoints[0], 3*sizeof(double));
  memcpy (meshSingle->Geom.EndPointCoords[1],
                  nurbsObjPtr->nurbs->CntlPoints[1], 3*sizeof(double));

  /* e-Lib functions will "explode" on error so there is no point 
    to do this at the moment -- SOMEON HAS TO FIX IT
   */
  if (e_PutGeomModel (meshSingle, fileName) != 1)
  {
    sprintf(errorMessage, "Writing NURBS to file %s failed", 
                             fileName); 
    resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
    result = TCL_ERROR;
  } 

  
  
  /* allocate the space and initialize the return structure */
  returnStructPtr = (CmdReturn *)  Tcl_Alloc (sizeof (CmdReturn));
  returnStructPtr->status = result;
  returnStructPtr->object = resultObjPtr; 

  return (returnStructPtr);        
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliNurbsObjMesh -- 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static CmdReturn*
EtliNurbsObjMesh (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[])
{
  int index;
  static char *subOpts[] = {"-add", "-remove", NULL};
  enum subOptsIx {AddIx, RemoveIx};

  int result = TCL_OK; 
  Tcl_Obj *resultObjPtr = NULL;
  CmdReturn *returnStructPtr = NULL;

  /* allocate the space for the return structure */
  returnStructPtr = (CmdReturn *) Tcl_Alloc (sizeof (CmdReturn));
  
  /* find option in the list of valid options for this subcommand */
  if (Tcl_GetIndexFromObj (interp, objv[2], subOpts, "option", TCL_EXACT, 
                          &index) != TCL_OK)
    {
       result = TCL_ERROR; 
       goto done;
    }

  switch (index)
    {
      case AddIx:
        {
           AddNurbsToMesh (clientData, interp, objc, objv, 
                             &result, &resultObjPtr); 
           break;
        }
      case RemoveIx:
        {
           RemoveNurbsFromMesh (clientData, interp, objc, objv, 
                             &result, &resultObjPtr); 
          break;
        }
    } /* end of switch */                
  done:
        returnStructPtr->object = resultObjPtr; 
        returnStructPtr->status = result; 

  return returnStructPtr;        
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EtliNurbsObjInfo -- 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static CmdReturn*
EtliNurbsObjInfo (ClientData clientData, Tcl_Interp *interp, 
              int objc, Tcl_Obj *CONST objv[])
{
  int index;
  static char *subOpts[] = {"-all", "-degree", "-ep1", "-ep2", 
                             "-ctrlpoint", "-coords", "-name", NULL};
  enum subOptsIx {AllIx, DegreeIx, Ep1Ix, Ep2Ix, CtrlIx, CoordsIx, NameIx};

  int result = TCL_OK; 
  Tcl_Obj *resultObjPtr = NULL;
  CmdReturn *returnStructPtr = NULL;

  NurbsObj *nurbsObjPtr = (NurbsObj *) clientData; 
  NURBSCURVE *nurbs = nurbsObjPtr->nurbs; /* convinience variable */


  /* allocate the space for the return structure */
  returnStructPtr = (CmdReturn *) Tcl_Alloc (sizeof (CmdReturn));
  
  /* find option in the list of valid options for this subcommand */
  if (Tcl_GetIndexFromObj (interp, objv[2], subOpts, "option", TCL_EXACT, 
                          &index) != TCL_OK)
    {
       result = TCL_ERROR; 
       goto done;
    }

  switch (index)
    {
      case AllIx:
        {
           InfoAll (clientData, interp, objc, objv, 
                             &result, &resultObjPtr); 
           break;
        }
      case DegreeIx:
        {
          resultObjPtr = Tcl_NewIntObj (nurbs->Degree); 
          break;
        }
      case Ep1Ix:
        {
          resultObjPtr = Tcl_NewIntObj (nurbs->EndPoint1); 
          break;
        }
      case Ep2Ix:
        {
          resultObjPtr = Tcl_NewIntObj (nurbs->EndPoint2);
          break; 
        }
      case CtrlIx: 
        {
           InfoControlPoint (clientData, interp, objc, objv, 
                             &result, &resultObjPtr); 
           break;
        }         
      case CoordsIx:
        {
          InfoCoords (clientData, interp, objc, objv, 
                          &result, &resultObjPtr); 
          break;
        }  
      case NameIx:
        {
          resultObjPtr = Tcl_NewStringObj (nurbs->Name, strlen (nurbs->Name));
          break;
        }
    } /* end of switch */                
  done:
        returnStructPtr->object = resultObjPtr; 
        returnStructPtr->status = result; 

  return returnStructPtr;        
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 InfoAll --
       Return list of all nurbs parametrs (in the form which is
       the same as when creating new nurbs object)
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static void
InfoAll (ClientData clientData, 
         Tcl_Interp *interp, 
         int objc, 
         Tcl_Obj *CONST objv[],
         int *result,
         Tcl_Obj **resultObjPtr)
{
  /* for holding return list */
  Tcl_Obj *listPtrPtr[3+4*(1+ETLI_MAX_NURBS_DEGREE)];
  int llength;  /* return list list length */
  int pointNumber; /*control point number */
  int j, i;
  NurbsObj *nurbsObjPtr = (NurbsObj *) clientData; 
  NURBSCURVE *nurbs = nurbsObjPtr->nurbs; /* convinience variable */

  /* check number of arguments */
  if (objc != 3)
    {
      sprintf(errorMessage, 
      "wrong # args: should be \"etli_nurbs info -all\"");
       *resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
       *result = TCL_ERROR;
       return;         
     }

  /* calculate length of the return list:
   *       3 items for   degree, ep1, ep2, 
   *  and  4* (degree+1) items for coords and weigth of each control
   *  point
   */
  llength = 3 + 4*(nurbs->Degree+1);
  
  /* add degree and end points number to return list */
  listPtrPtr[0] = Tcl_NewIntObj (nurbs->Degree);
  listPtrPtr[1] = Tcl_NewIntObj (nurbs->EndPoint1);
  listPtrPtr[2]= Tcl_NewIntObj (nurbs->EndPoint2);
  
  /* construc list of control point coords and weight */
  j=3; /* we already filled thre items */
  for (pointNumber=0; pointNumber < nurbs->Degree+1; pointNumber++)
    {
      for (i=0; i<4; i++)
        {
          listPtrPtr[j] = 
                 Tcl_NewDoubleObj (*(nurbs->CntlPoints[pointNumber]+i));
          j++;
        } 
    }

   *resultObjPtr = Tcl_NewListObj(llength, listPtrPtr); 

   return;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 InfoControlPoint --
           Returns control point coords and weight.
 Syntax:          
      nurbsObj info -ctrlpoint number
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static void
InfoControlPoint (ClientData clientData, 
                  Tcl_Interp *interp, 
                  int objc, 
                  Tcl_Obj *CONST objv[],
                  int *result,
                  Tcl_Obj **resultObjPtr)
{
  int pointNumber,j;
  Tcl_Obj *listPtrPtr[4]; /* for holding coords and weight */

  NurbsObj *nurbsObjPtr = (NurbsObj *) clientData; 
  NURBSCURVE *nurbs = nurbsObjPtr->nurbs; /* convinience variable */

  /* check if control point number is supplied for -ctrlpoint option */
  if (objc != 4)
    {
      sprintf(errorMessage, 
      "wrong # args: should be \"etli_nurbs info -ctrlpoint number\"");
       *resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
       *result = TCL_ERROR;
       return;         
     }
  if (Tcl_GetIntFromObj (interp, objv[3], &pointNumber) != TCL_OK)
    {
      *result = TCL_ERROR; 
      return;
    }  

  /* check if control point number is valid */
  if ( pointNumber<1 || pointNumber>nurbs->Degree+1 )
    {
      sprintf(errorMessage, "wrong point number: expected integer 1-%d", 
                     nurbs->Degree+1); 
      *resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
      *result = TCL_ERROR; 
      return; 
    }

  pointNumber--; /* control point are numbered from 1 but this is C */

  /* construc list of control point coords and weight */
  for (j=0; j<4; j++)
    {
      listPtrPtr[j] = Tcl_NewDoubleObj (*(nurbs->CntlPoints[pointNumber]+j));
    } 
   *resultObjPtr = Tcl_NewListObj(4, listPtrPtr); 

   return;
}    

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 InfoCoords --
        Return list of x,y,z coords of point on nurbs curve for
        given curve parameter t.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static void
InfoCoords (ClientData clientData, 
            Tcl_Interp *interp, 
            int objc, 
            Tcl_Obj *CONST objv[],
            int *result,
            Tcl_Obj **resultObjPtr)
{
  int j;
  double t; /* curve parameter */
  double coords[3]; /* curve point coords */
  Tcl_Obj *listPtrPtr[3]; /* for holding x,y,z  coords */

  NurbsObj *nurbsObjPtr = (NurbsObj *) clientData; 
  NURBSCURVE *nurbs = nurbsObjPtr->nurbs; /* convinience variable */

  /* check if curve parameter is supplied for -coords option */
  if (objc != 4)
    {
      sprintf(errorMessage, 
      "wrong # args: should be \"etli_nurbs info -coords t_param\"");
       *resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
       *result = TCL_ERROR;
       return;         
     }

  /* retrieve curve parameter from arguments list */ 
  if (Tcl_GetDoubleFromObj (interp, objv[3], &t) != TCL_OK)
    {
      *result = TCL_ERROR; 
      return;
    }  

  /* check if curve parameter is valid  <0,1>*/
  if ( t<0.0 || t>1.0 )
    {
      sprintf(errorMessage, "wrong curve parameter: expected double <0,1>"); 
      *resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
      *result = TCL_ERROR; 
      return; 
    }

  /* get curve point coords for given parameter t (Jelle function) */
      crv_getcoords (nurbs, t, coords); 
  /* new function supplied by Jelle, this was only for testing purposes */
  /*    SetCurvePoint (nurbs, t, coords); */
  
  
  /* construc list of x,y,z coords for giver t */
  for (j=0; j<3; j++)
    {
      listPtrPtr[j] = Tcl_NewDoubleObj (coords[j]);
    } 
   *resultObjPtr = Tcl_NewListObj(3, listPtrPtr); 

   return;
} 

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 InitSingleMesh --
        Initializes the meshSingl structure to hold one NURBS 
        curve.
        The action of this function are performed only once regardles
        how many times it is called.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void InitSingleMesh (void)
{
 static int called=0;

 if (called == 0)
 {
    called = 1;
    e_NullifyMesh (meshSingle);
    meshSingle->Geom.NEndPoints = 2;
    meshSingle->Geom.NNurbsCurves = 1;
    meshSingle->Geom.EndPointCoords = e_alloc_cont_d_matr (2,3) ; 
    meshSingle->SubStructSet[E_SS_NUM_GEOM] = TRUE; 
 }
 return;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 InitCollectMesh --
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void InitCollectMesh (MESH *mesh)
{
 static int called=0;

 if (called == 0 && mesh == NULL)
 {
    called = 1;
    meshCollect = &meshCollect_Str;
    e_NullifyMesh (meshCollect);
    meshCollect->SubStructSet[E_SS_NUM_GEOM] = TRUE; 
 }
 else if (called == 0 && mesh != NULL)
 {
   called = 1;
   meshCollect = mesh;
   meshCollect->SubStructSet[E_SS_NUM_GEOM] = TRUE; 
 }
 else if (called == 1 && mesh != NULL)
 {
   if (meshCollect != &meshCollect_Str)
   {
     mesh->Geom.NNurbsCurves = meshCollect->Geom.NNurbsCurves;
     mesh->Geom.NurbsCurves = meshCollect->Geom.NurbsCurves;
     meshCollect = mesh;
   }
 }
 return;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 AddNurbsToMesh --
           Adds nurbs objec to mesh.
           Actualy this function adds nurbs object to the
           array of pointers to NURBSCURVE pointers.
 Syntax:          
      nurbsObj  mesh -add
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static void
AddNurbsToMesh (ClientData clientData, 
                  Tcl_Interp *interp, 
                  int objc, 
                  Tcl_Obj *CONST objv[],
                  int *result,
                  Tcl_Obj **resultObjPtr)
{
  NurbsObj *nurbsObjPtr = (NurbsObj *) clientData; 
  NURBSCURVE *nurbs = nurbsObjPtr->nurbs; /* convinience variable */

  InitCollectMesh (NULL);

  nurbsCollect = (NURBSCURVE **) e_incr_vp_vect (
      meshCollect->Geom.NNurbsCurves, (void *) nurbsCollect, 1);

  nurbsCollect[meshCollect->Geom.NNurbsCurves] = nurbs; 
  
  meshCollect->Geom.NNurbsCurves++;

  *result = TCL_OK;
  *resultObjPtr = NULL;

  return;
} 

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RemoveNurbsFromeMesh --
           Removes nurbs objec from mesh.
 Syntax:          
      nurbsObj  mesh -remove
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        static void
RemoveNurbsFromMesh (ClientData clientData, 
                  Tcl_Interp *interp, 
                  int objc, 
                  Tcl_Obj *CONST objv[],
                  int *result,
                  Tcl_Obj **resultObjPtr)
{
  register int i;
  int nnum = -1;
  int rest;
  NurbsObj *nurbsObjPtr = (NurbsObj *) clientData; 
  NURBSCURVE *nurbs = nurbsObjPtr->nurbs; /* convinience variable */
  NURBSCURVE **nurbsArray; 

  for (i=0; i<meshCollect->Geom.NNurbsCurves; i++)
  {
    if (nurbs == nurbsCollect[i])
    {
       nnum = i;
       break;
    }
  }
  if (nnum > -1 && nnum < meshCollect->Geom.NNurbsCurves)
  {
      *result = TCL_OK;
      sprintf(errorMessage, "Warning: NURBS not in mesh"); 
      *resultObjPtr = Tcl_NewStringObj (errorMessage, -1); 
  }

  nurbsArray =(NURBSCURVE **)e_alloc_vp_vect(meshCollect->Geom.NNurbsCurves-1);
  if (nnum > 0)
  {
    memcpy (nurbsArray, nurbsCollect, nnum*sizeof(NURBSCURVE*));
  }
  rest = meshCollect->Geom.NNurbsCurves - nnum-1;
  if (rest > 0)
  {
   memcpy (nurbsArray+nnum, 
           nurbsCollect+nnum+1, rest*sizeof(NURBSCURVE*)); 
  } 
  e_free  (nurbsCollect); 
  meshCollect->Geom.NNurbsCurves--;
  
  nurbsCollect = nurbsArray;

  *result = TCL_OK;
  *resultObjPtr = NULL;

  return;
} 
/* Modifications:
 * $Id: etliCmdNurbs.c,v 1.1.1.1 2002/08/04 13:19:11 peteri Exp $
 */

