 /********************************************************************
  mesh_supp.c 
            MESH interface routines 
 
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University Edinburgh    
 ********************************************************************/

 /* $Id: mesh_supp.c,v 1.1.1.1 2002/08/04 13:19:11 peteri Exp $ */

#include "mesh_supp.h"
#include <assert.h>

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RP_MeshPointsBBox
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define LOW_X  (*low)
#define LOW_Y  (*(low+1))
#define LOW_Z  (*(low+2))
#define HIGH_X  (*high)
#define HIGH_Y  (*(high+1))
#define HIGH_Z  (*(high+2))
#define MESHPOINT_X(ii) (*mesh->MeshPointCoords[ii])
#define MESHPOINT_Y(ii) (*(mesh->MeshPointCoords[ii]+1))
#define MESHPOINT_Z(ii) (*(mesh->MeshPointCoords[ii]+2))

	int
RP_MeshPointsBBox (MESH *mesh, double *low, double *high)
{
  register int i;
  int result = 1; 

  /* check validity of mesh pointer */
  assert (mesh != NULL);
  
  /* check validity of other pointers */
  assert (low != NULL);
  assert (high != NULL); 

  /* initialize bounding box points */
  LOW_X = MESHPOINT_X(0); 
  LOW_Y = MESHPOINT_Y(0);
  LOW_Z = MESHPOINT_Z(0);
  HIGH_X = MESHPOINT_X(0);
  HIGH_Y = MESHPOINT_Y(0);
  HIGH_Z = MESHPOINT_Z(0);

 /* loop through all mesh points */
  for (i=1; i<mesh->NMeshPoints; i++) /* we skip point 0 as obvious */
     {
       if (LOW_X > MESHPOINT_X(i)) LOW_X = MESHPOINT_X(i); 
       if (LOW_Y > MESHPOINT_Y(i)) LOW_Y = MESHPOINT_Y(i); 
       if (LOW_Z > MESHPOINT_Z(i)) LOW_Z = MESHPOINT_Z(i); 
       if (HIGH_X < MESHPOINT_X(i)) HIGH_X = MESHPOINT_X(i); 
       if (HIGH_Y < MESHPOINT_Y(i)) HIGH_Y = MESHPOINT_Y(i); 
       if (HIGH_Z < MESHPOINT_Z(i)) HIGH_Z = MESHPOINT_Z(i); 
     }
  return result; 
}    
#undef LOW_X  
#undef LOW_Y  
#undef LOW_Z 
#undef HIGH_X 
#undef HIGH_Y 
#undef HIGH_Z  
#undef MESHPOINT_X
#undef MESHPOINT_Y
#undef MESHPOINT_Z

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RP_MeshElementType -- 
         Search over FirstNetIndex to find element type.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        int
RP_MeshElementType (MESH *mesh, int netIndex)
{
  register int i;
  register int low;
  register int newlow;
  register int type; 

  assert (mesh != NULL);

  /* check if netIndex is valid */
  if (netIndex <1 || netIndex>mesh->TotalNElems)
    {
      return -1; 
    }
  
  low = 1;
  type = -1; 
  for (i=0; i<E_NELEM_TYPES; i++)
    { 
      newlow=mesh->FirstNetIndex[i]; 
      if (newlow <= netIndex && newlow >= low) 
        {
          type = i;
          low = newlow;  
        }
    }
  return type;   
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RP_MeshElementNodes --
          For documentation see mesh_supp.h
          Trick is that the nodeList must not be modified of  freed
          as it points to memory managed by mesh object 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define _CASE_ELEM(name1, name2) \
        {\
          *llength = E_NNPE_##name1; \
          element = netIndex - mesh->FirstNetIndex[E_ELEM_TYPE_##name1]; \
          *nodeList = mesh->Nodes##name2[element]; \
          break; \
        } 

        int
 RP_MeshElementNodes (MESH *mesh, int netIndex, 
                      int **nodeList, int *llength)
{
  int elementType, element;
  int result=1;
  char error[80];
  
  assert (mesh != NULL); 

  elementType = RP_MeshElementType (mesh, netIndex);

  switch (elementType)
    {
      case E_ELEM_TYPE_LINK1: 
          _CASE_ELEM(LINK1, Link1)
      case E_ELEM_TYPE_LINK2: 
          _CASE_ELEM(LINK2, Link2)
      case E_ELEM_TYPE_LINK3: 
          _CASE_ELEM(LINK3, Link3)
      case E_ELEM_TYPE_LINK4: 
          _CASE_ELEM(LINK4, Link4)
      case E_ELEM_TYPE_LINK5: 
          _CASE_ELEM(LINK5, Link5)
      case E_ELEM_TYPE_TRIANG1: 
          _CASE_ELEM(TRIANG1, Triang1)
      case E_ELEM_TYPE_TRIANG2: 
          _CASE_ELEM(TRIANG2, Triang2)
      case E_ELEM_TYPE_TRIANG3: 
          _CASE_ELEM(TRIANG3, Triang3)
      case E_ELEM_TYPE_TRIANG4: 
          _CASE_ELEM(TRIANG4, Triang4)
      case E_ELEM_TYPE_TRIANG5: 
          _CASE_ELEM(TRIANG5, Triang5)
      case E_ELEM_TYPE_QUAD1: 
          _CASE_ELEM(QUAD1, Quad1)
      case E_ELEM_TYPE_QUAD3: 
          _CASE_ELEM(QUAD3, Quad3)
      case E_ELEM_TYPE_TETRAH1: 
          _CASE_ELEM(TETRAH1, Tetrah1)
      case E_ELEM_TYPE_BLOCK1: 
          _CASE_ELEM(BLOCK1, Block1)
      case -1:          
        {
          llength = 0;
          nodeList = NULL; 
          result = 0;
          break;
        }
      default: 
        {
          sprintf (error,
               "RP_MeshElementNodes: Invalid element type index: %d. %s",
                elementType, "Has no switch entry.");
          e_FileWarning(__FILE__,error);
          result = 0;
        }
    }/* end of switch(elementType) */
  return result;
}
#undef _CASE_ELEM

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RP_MeshElementVertices --
          For documentation see mesh_supp.h
          Trick is that the nodeList must not be modified of  freed
          as it points to memory managed by mesh object 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define _CASE_ELEM(name1, name2) \
        {\
          *llength = E_NVPE_##name1; \
          element = netIndex - mesh->FirstNetIndex[E_ELEM_TYPE_##name1]; \
          *vertexList = mesh->Nodes##name2[element]; \
          break; \
        } 

        int
 RP_MeshElementVertices (MESH *mesh, int netIndex, 
                      int **vertexList, int *llength)
{
  register int elementType;
  register int element;
  int result = 1;
  
  assert (mesh != NULL); 

  elementType = RP_MeshElementType (mesh, netIndex);

  switch (elementType)
    {
      case E_ELEM_TYPE_LINK1: 
          _CASE_ELEM(LINK1, Link1)
      case E_ELEM_TYPE_LINK2: 
          _CASE_ELEM(LINK2, Link2)
      case E_ELEM_TYPE_LINK3: 
          _CASE_ELEM(LINK3, Link3)
      case E_ELEM_TYPE_LINK4: 
          _CASE_ELEM(LINK4, Link4)
      case E_ELEM_TYPE_LINK5: 
          _CASE_ELEM(LINK5, Link5)
      case E_ELEM_TYPE_TRIANG1: 
          _CASE_ELEM(TRIANG1, Triang1)
      case E_ELEM_TYPE_TRIANG2: 
          _CASE_ELEM(TRIANG2, Triang2)
      case E_ELEM_TYPE_TRIANG3: 
          _CASE_ELEM(TRIANG3, Triang3)
      case E_ELEM_TYPE_TRIANG4: 
          _CASE_ELEM(TRIANG4, Triang4)
      case E_ELEM_TYPE_TRIANG5: 
          _CASE_ELEM(TRIANG5, Triang5)
      case E_ELEM_TYPE_QUAD1: 
          _CASE_ELEM(QUAD1, Quad1)
      case E_ELEM_TYPE_QUAD3: 
          _CASE_ELEM(QUAD3, Quad3)
      case E_ELEM_TYPE_TETRAH1: 
          _CASE_ELEM(TETRAH1, Tetrah1)
      case E_ELEM_TYPE_BLOCK1: 
          _CASE_ELEM(BLOCK1, Block1)
      case -1:          
        {
          llength = 0;
          vertexList = NULL;
          result = 0;
          break;
        }
      default: 
        {
          char error[80];
          sprintf (error, "Invalid element type index: %d. %s",
                          elementType, "Has no switch entry.");
          e_FileWarning(__FILE__,error);
          result = 0;
        }
    }/* end of switch(elementType) */
  return result;
}
#undef _CASE_ELEM


/* Modifications:
 * $Log: mesh_supp.c,v $
 * Revision 1.1.1.1  2002/08/04 13:19:11  peteri
 * Imported using TkCVS
 *
 */ 
