#include "e_lib.h"
#include "adv_edge.h"
#include "adv_mpar.h"    /* mesh parameter determination */
#include "adv_config.h"  /* parameters for the advancing front algorithm */
#include "getvertices.h"
#include "nodelem.h"
#include "boundary.h"
#include "eg_lib.h"

static void RefineEdge(MESH *mesh, BG_DATA *bgd,
                       int ctrl, int geomset, int type, int n1, int n2,
                       int *nnodes, double ***coords, double **mpr, double h,
                       int nbgedge, EDGE *bgedges, EDGE *edge);

static void CollectEdges(MESH *mesh, BG_DATA *bgd,
                         int ctrl, int geomset, int force_even,
                         int NElem, int NNodes, int **Nodes, int StartIndex,
                         int *no_node_elem, int **node_elem, EDGE **edges,
                         int *newnodes, double ***newcoords, double **newmpr,
                         double h);


/* //////////////////////////////////////////////////////////////////////////
Function:     RefineEdge
Descriptions: refines an edge given by two point, store the coordinates of the
              new nodes and the nodes numbers in the edge data structure
Arguments:
              mesh: coarse mesh
              bgd: all information of background mesh and sources
              ctrl: mesh parameter control type, MPR or BACKGROUND
              geomset: whether geometric information is available
              type: even or odd number of nodes to generate, or does not matter
              n1, n2: end nodes of the edges
              nnodes: number of nodes in coords
              coords: coordinates of the points
              mpr: mesh parameters of the points
              nbgedge: number of edges to use for mesh parameter interpolation
              bgedges: edges to use for mesh parameter interpolation
              edge: the generated edge
Return:       None
///////////////////////////////////////////////////////////////////////////// */
static void RefineEdge(MESH *mesh, BG_DATA *bgd,
                       int ctrl, int geomset, int type, int n1, int n2,
                       int *nnodes, double ***coords, double **mpr, double h,
                       int nbgedge, EDGE *bgedges, EDGE *edge)
{
  double       d1, d2;
  int          nurbs, n;
  double      *local_mpr, *par_t;
  double     **local_coords;
  NURBSCURVE  *curve;
  double       t1, t2;
  int          start, end, ii;

  /* mesh parameter at one end */
  d1 = AdvGetMeshParam(mesh->MeshPointCoords[n1], ctrl, bgd,
                       nbgedge, bgedges, *coords, *mpr, h);

  /* mesh parameter at the other end */
  d2 = AdvGetMeshParam(mesh->MeshPointCoords[n2], ctrl, bgd,
                       nbgedge, bgedges, *coords, *mpr, h);

  nurbs = 0;
  /* is there nurbs in the mesh */
  if(geomset)
  {
    nurbs = e_GetCurve(mesh, n1+1, n2+1, &curve, &t1, &t2);
  }

  if(nurbs)
  {
    /* curved boundary */
    n = e_CurveRefine(curve, t1, d1, t2, d2,
                      &local_coords, &local_mpr, &par_t, TRUE, type);
    start = 1;
    end   = n - 1;
  }
  else
  {
    /* straight line */
    n = e_SegmentRefine(mesh->MeshPointCoords[n1], d1,
                        mesh->MeshPointCoords[n2], d2,
                        &local_coords, &local_mpr, TRUE, type);
    start = 0;
    end   = n;
  }


  edge->en = e_alloc_i_vect(end-start + 2);
  edge->en[0] = n1+1;
  if(n > 0)
  {
    /* FIXME: do it with memcpy ??? */
    /* store the coordinates and the mesh parameters */
    for(ii = 0; ii < end - start; ii++)
    {
      (*coords) = e_incr_cont_d_matr(*nnodes, E_NDIMS, *coords, INCREMENT);
      (*coords)[*nnodes][0] = local_coords[ii+start][0];
      (*coords)[*nnodes][1] = local_coords[ii+start][1];

      if(ctrl == CTRL_MPR)
      {
        (*mpr) = e_incr_d_vect(*nnodes, *mpr, INCREMENT);
        (*mpr)[*nnodes] = local_mpr[ii+start];
      }

      (*nnodes)++;
      edge->en[ii+1] = (*nnodes);
    }

    e_free_cont_d_matr(n, local_coords);
    e_free(local_mpr);
    if(nurbs)
    {
      e_free(par_t);
    }

  }

  edge->en[end-start+1] = n2+1;
  edge->nen = end-start+2;

}

/* //////////////////////////////////////////////////////////////////////////
Function:     CollectEdges
Descriptions: for each element of one type it generates the nodes
              on all sides,
              when a side is shared between two elements, the node numbers
              are the same, but in a reverse order
Arguments:
              mesh: coarse mesh
              bgd: all data for the background mesh and sources
              ctrl: mesh parameter control
              geomset: whether geometric information is available
              force_even: even number of nodes on the _mesh_ boundary
              NElem: number of elements of this type
              NNodes: number of nodes in this element type
              Nodes: node definition of the element type
              StartIndex: FirstNetIndex of the element type
              no_node_elem: node element connectivity data
              node_elem: node element connectivity data
              edges: matrix of edges per elements
              newnodes: number of nodes in newcoords
              newcoords: coordinates of nodes
              newmpr: new mesh parameters on the boundary
Return:       None
///////////////////////////////////////////////////////////////////////////// */
static void CollectEdges(MESH *mesh, BG_DATA *bgd,
                         int ctrl, int geomset, int force_even,
                         int NElem, int NNodes, int **Nodes, int StartIndex,
                         int *no_node_elem, int **node_elem, EDGE **edges,
                         int *newnodes, double ***newcoords, double **newmpr,
                         double h)
{
  int i, j, ii, jj;
  int ns, nx, n1, n2, smaller, n, bnd;
  int GIndex, index1, index2;
  int nvertices, *vertices;
  EDGE bgedges[4]; /* 4 because max quad element */
  int  bgnodes[4][2];

  for(i = 0; i < NElem; i++)
  {
    GIndex = StartIndex + i;

    /* closed polygons, therefore no of nodes equal no of sides */
    if(edges[GIndex-1] == NULL)
    {
      edges[GIndex-1] = (EDGE *)e_calloc(NNodes, sizeof(EDGE) );
    }

    /* create edge structure for interpolation */
    ns = 0;
    for(j = 0; j < NNodes; j++)
    {
      nx = NNodes - 1 - j;
      bgnodes[j][0] = Nodes[i][ns];
      bgnodes[j][1] = Nodes[i][nx];
      bgedges[j].en = bgnodes[j];
      bgedges[j].nen = 2;

      ns = nx;
    }

    ns = 0;
    for(j = 0; j < NNodes; j++)
    {
      nx = NNodes - 1 - j;

      n1 = Nodes[i][ns]-1;
      n2 = Nodes[i][nx]-1;

      /* check for adjacent element */
      smaller = -1;
      bnd = TRUE;
      for(ii = 0; ii < no_node_elem[n1] && smaller == -1; ii++)
      {
        for(jj = 0; jj < no_node_elem[n2] && smaller == -1; jj++)
        {
          if(node_elem[n1][ii] == node_elem[n2][jj])
          {
            /* we found an adjacent element, not boundary */
            if(GIndex != node_elem[n1][ii])
            {
              bnd = FALSE;
            }
            if(node_elem[n1][ii] < GIndex)
            {
              smaller = node_elem[n1][ii];
            }
          }
        }
      }

      /* find an empty slot */
      for(index1 = 0; index1 < NNodes; index1++)
      {
        if(edges[GIndex-1][index1].nen == 0)
          break;
      }
      if(index1 == NNodes)
        e_GeneralError("CollectEdges: internal error 1");

      /* if we need even number of nodes on the boundary of the whole mesh */
      if(force_even && bnd)
      {
        RefineEdge(mesh, bgd, ctrl, geomset, E_SEGREF_ODD, n2, n1,
                   newnodes, newcoords, newmpr, h, NNodes, bgedges,
                   &(edges[GIndex-1][index1]) );
      }
      /* boundary of the smaller adjacent element */
      else if(smaller == -1)
      {
        RefineEdge(mesh, bgd, ctrl, geomset, E_SEGREF_NONE, n2, n1,
                   newnodes, newcoords, newmpr, h, NNodes, bgedges,
                   &(edges[GIndex-1][index1]) );
      }
      /* smaller element already stored the edge, here we store it reversed */
      else
      {
        smaller--;
        e_GetElementVertices(smaller, mesh, &nvertices, &vertices);

        if(edges[smaller] == NULL)
        {
          edges[smaller] = (EDGE *)e_calloc(nvertices, sizeof(EDGE) );
        }

        for(index2 = 0; index2 < nvertices; index2++)
        {
          n = edges[smaller][index2].nen;
          if(   n1+1 == edges[smaller][index2].en[0]
             && n2+1 == edges[smaller][index2].en[n-1])
          {
            break;
          }
        }
        if(index2 == nvertices)
          e_GeneralError("CollectEdges: internal error 2");

        /* store it reversed */
        edges[GIndex-1][index1].nen = n;
        edges[GIndex-1][index1].en = e_alloc_i_vect(edges[GIndex-1][index1].nen);
        for(ii = 0; ii < edges[GIndex-1][index1].nen; ii++)
        {
          edges[GIndex-1][index1].en[ii] = edges[smaller][index2].en[n-1-ii];
        }
      }

      ns = nx;
    }
  }
}


/* //////////////////////////////////////////////////////////////////////////
Function:     CreateElemEdges
Descriptions: in the coarse mesh for all element types generate the refined
              edges, which can be passed for meshing
Arguments:
              mesh: coarse mesh
              bgmesh: background mesh
              ParamCntlSets: mesh parameter sources
              geomset: whether geometric information is available
              nnodes: number of nodes in coords
              coords: coordinates of new points
              edges: the edge list per element
Return:       None
///////////////////////////////////////////////////////////////////////////// */
void CreateElemEdges(MESH *mesh, BG_DATA *bgd,
                     int ctrl, int geomset, int force_even,
                     int *nnodes, double ***coords, double **mpr, double h,
                     EDGE ***edges)
{
  int  *no_node_elem;
  int **node_elem;
  int   alloc;

  /* store all edge end points */
  alloc = (int)(mesh->NMeshPoints / INCREMENT)+1;
  *coords = e_alloc_cont_d_matr(INCREMENT*alloc, E_NDIMS);
  memcpy((void *)((*coords)[0]),
         (void *)(mesh->MeshPointCoords[0]),
         sizeof(double) * E_NDIMS * mesh->NMeshPoints);
  /* mesh parameters */
  if(ctrl == CTRL_MPR)
  {
    *mpr = e_alloc_d_vect(INCREMENT*alloc);
    memcpy((void *)(*mpr),
           (void *)(mesh->MeshParam.NodalMeshParams),
           sizeof(double) * mesh->NMeshPoints);
  }
  else
  {
    *mpr = NULL;
  }
  *nnodes = mesh->NMeshPoints;

  (*edges) = (EDGE **)e_init_frag_vp_matr(mesh->TotalNElems);

  /* node element connectivity */
  no_node_elem = e_alloc_i_vect(mesh->NMeshPoints);
  node_elem = e_init_frag_i_matr(mesh->NMeshPoints);
  e_InitMeshPointElemConnection(mesh, no_node_elem, node_elem);

  /* organize, refine edges */
  CollectEdges(mesh, bgd, ctrl, geomset, force_even,
               mesh->NElemsTriang1, E_NNPE_TRIANG1, mesh->NodesTriang1,
               mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1],
               no_node_elem, node_elem, *edges, nnodes, coords, mpr, h);

  CollectEdges(mesh, bgd, ctrl, geomset, force_even,
               mesh->NElemsQuad1, E_NNPE_QUAD1, mesh->NodesQuad1,
               mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1],
               no_node_elem, node_elem, *edges, nnodes, coords, mpr, h);

  e_free(no_node_elem);
  e_free_frag_i_matr(mesh->NMeshPoints, node_elem);

}



/* //////////////////////////////////////////////////////////////////////////
Function:     CreateMeshEdges
Descriptions: extract the boundary from the coarse mesh and each segment is
              refined and converted into an edge
Arguments:
              mesh: coarse mesh
              bgd: all data for the background mesh and sources
              geomset: whether geometric information is available
              force_even: generate even number of nodes (for Quad fusion)
              nnodes: number of nodes in coords
              coords: coordinates of new points
              nedge: number of edges in the list
              edges: the edge list per element
Return:       None
///////////////////////////////////////////////////////////////////////////// */
void CreateMeshEdges(MESH *mesh, BG_DATA *bgd,
                     int ctrl, int geomset, int force_even,
                     int *nnodes, double ***coords, double **mpr, double h,
                     int *nedge, EDGE **edges)
{
  int  *no_node_elem;
  int **node_elem;
  int   alloc;
  int   i, j, n1, n2, nn, iedge;
  int   NBoundaries;
  BND  *Boundaries;
  int   imax, totalbnode, type;
  double lmax, l;
  int   nbgedge;
  EDGE *bgedges;

  /* store all edge end points */
  alloc = (int)(mesh->NMeshPoints / INCREMENT)+1;
  *coords = e_alloc_cont_d_matr(INCREMENT*alloc, E_NDIMS);
  memcpy((void *)((*coords)[0]),
         (void *)(mesh->MeshPointCoords[0]),
         sizeof(double) * E_NDIMS * mesh->NMeshPoints);
  /* mesh parameters */
  if(ctrl == CTRL_MPR)
  {
    *mpr = e_alloc_d_vect(INCREMENT*alloc);
    memcpy((void *)(*mpr),
           (void *)(mesh->MeshParam.NodalMeshParams),
           sizeof(double) * mesh->NMeshPoints);
  }
  else
  {
    *mpr = NULL;
  }
  *nnodes = mesh->NMeshPoints;


  /* node element connectivity */
  no_node_elem = e_alloc_i_vect(mesh->NMeshPoints);
  node_elem = e_init_frag_i_matr(mesh->NMeshPoints);
  e_InitMeshPointElemConnection(mesh, no_node_elem, node_elem);

  /* determine mesh boundary */
  Gen2DBoundary(mesh, no_node_elem, node_elem,
                &NBoundaries, &Boundaries);
  if(NBoundaries != 1)
  {
    e_GeneralError("adv_main: more than one external boundary in the mesh !");
  }

  /* determine number of edges */
  (*nedge) = Boundaries[0].oe.nen;
  for(i = 0; i < Boundaries[0].nie; i++)
  {
    (*nedge) += Boundaries[0].ie[i].nen;
  }
  /* allocate the edge list */
  (*edges) = (EDGE *)e_calloc(*nedge, sizeof(EDGE) );

  /* setup for MPR interpolation */
  nbgedge = Boundaries[0].nie + 1;
  bgedges = (EDGE *)e_calloc(nbgedge, sizeof(EDGE) );
  bgedges[0] = Boundaries[0].oe;
  for(i = 0; i < Boundaries[0].nie; i++)
  {
    bgedges[i+1] = Boundaries[0].ie[i];
  }

  /* determine the longest segment on the external boundary if necessary */
  if(force_even)
  {
    imax = -1;
    lmax = 0.0;
    nn = Boundaries[0].oe.nen;
    n1 = Boundaries[0].oe.en[nn-1]-1;
    for(i = 0; i < nn; i++)
    {
      n2 = Boundaries[0].oe.en[i]-1;
      l = eg_PointDist3D((*coords)[n1], (*coords)[n2] );
      if(l > lmax || imax == -1)
      {
        lmax = l;
        imax = i;
      }

      n1 = n2;
    }
  }

  /* edge index */
  iedge = 0;
  /* number of boundary nodes,
     may not be the same as the total number of nodes */
  totalbnode = 1;
  /* first refine the internal holes */
  for(j = 0; j < Boundaries[0].nie; j++)
  {
    nn = Boundaries[0].ie[j].nen;
    n1 = Boundaries[0].ie[j].en[nn-1]-1;
    for(i = 0; i < nn; i++)
    {
      n2 = Boundaries[0].ie[j].en[i]-1;

      RefineEdge(mesh, bgd, ctrl, geomset, E_SEGREF_NONE, n1, n2,
                 nnodes, coords, mpr, h, nbgedge, bgedges,
                 &((*edges)[iedge]) );

      totalbnode += (*edges)[iedge].nen-1;
      n1 = n2;
      iedge++;
    }
  }


  /* external boundary segment refinements */
  nn = Boundaries[0].oe.nen;
  n1 = Boundaries[0].oe.en[nn-1]-1;
  for(i = 0; i < nn; i++)
  {
    n2 = Boundaries[0].oe.en[i]-1;

    if(   !force_even
       || (force_even && i != imax))
    {
      RefineEdge(mesh, bgd, ctrl, geomset, E_SEGREF_NONE, n1, n2,
                 nnodes, coords, mpr, h, nbgedge, bgedges,
                 &((*edges)[iedge]) );

      totalbnode += (*edges)[iedge].nen-1;
      iedge++;
    }

    n1 = n2;
  }

  /* refine the longest as the last to have even number of nodes on the boundary */
  if(force_even)
  {
    n1 = Boundaries[0].oe.en[(imax-1+nn) % nn]-1;
    n2 = Boundaries[0].oe.en[imax]-1;

    type = (totalbnode % 2) ? E_SEGREF_ODD : E_SEGREF_EVEN;

    RefineEdge(mesh, bgd, ctrl, geomset, type, n1, n2,
               nnodes, coords, mpr, h, nbgedge, bgedges,
               &((*edges)[iedge]) );

    iedge++;
  }

  Free2DBoundary(NBoundaries, Boundaries);

  e_free(no_node_elem);
  e_free_frag_i_matr(mesh->NMeshPoints, node_elem);

  e_free(bgedges);
}



