#include "bgdata.h"
#include "eg_lib.h"
#include "pcs_int.h"
#include "mgn.h"
#include "dtr.h"
#include "mgn_aux.h"

static double mpr_of_point_in_elem_mpr_mesh(int bt,int s,int **tr,int ntr,double *mpr);

void get_bg_data(BG_DATA **bgd, MESH *bgmesh, PARAMCNTLSETS *pcs)
{
  if (bgmesh == NULL)
    e_GeneralError("get_bg_data: Background mesh not set");
  if (!bgmesh->SubStructSet[E_SS_NUM_MESHPARAM])
    e_GeneralError("get_bg_data: Background mesh parameters not set");
  
  *bgd=(BG_DATA *)e_calloc(1,sizeof(BG_DATA));
  parse_full_mesh(bgmesh, &((*bgd)->bgcoords), &((*bgd)->bgnp), 
                  &((*bgd)->bgtr), &((*bgd)->bgntr), FALSE);
  
  if (!CCW_check((*bgd)->bgtr, (*bgd)->bgntr, (*bgd)->bgcoords))
    e_GeneralError("get_bg_data: not all elements in your background mesh are CCW!\n");

  if (bgmesh->MeshParam.NodalMeshParamsSet)
  {
    (*bgd)->bgmpr=e_alloc_d_vect(bgmesh->NMeshPoints);
    memcpy(&((*bgd)->bgmpr[0]),&(bgmesh->MeshParam.NodalMeshParams[0]),
      		 bgmesh->NMeshPoints*sizeof(double));
    (*bgd)->bgwmpr=NODEMPRSET;
  }
  else /* if (bgmesh->MeshParam.ElemMeshParamsSet) */
  {
    (*bgd)->bgmpr=e_alloc_d_vect(bgmesh->NElemsTriang1);
    memcpy(&((*bgd)->bgmpr[0]),&(bgmesh->MeshParam.ElemMeshParamsTriang1[0]),
      		 bgmesh->NElemsTriang1*sizeof(double));
    (*bgd)->bgwmpr=ELEMMPRSET;
  }
  if (pcs!=NULL)
    (*bgd)->pcs=pcs;
}

void free_bg_data(BG_DATA *bgd)
{
  e_free_cont_d_matr(bgd->bgnp,bgd->bgcoords);
  e_free_cont_i_matr(bgd->bgntr,bgd->bgtr);
  e_free(bgd->bgmpr);
  e_free(bgd);
}

double get_bg_mpr(BG_DATA *bgd, double *pnt, int ctrl, MTRF *mtrf)
{
  double a,area;		  /* area for area based interpolation */
  double dpc;					/* the return value */
  double ldpc;				/* result of sources interpolation */
  int s;							/* counters */
  int sres;						/* search result */
  int bt;							/* background triangle pnt lies in */
  int onside;					/* search result */

  if ((ctrl!=CTRL_BACKGROUND)&&(ctrl!=CTRL_SOURCES))
    e_GeneralError("get_bg_mpr: only for CTRL_BACKGROUND and CTRL_SOURCES");
    
  /* for background or sources controlled remeshing lookup the mpr in
     the underlying mesh */
  sres = exh_tr_search(pnt, bgd->bgcoords, bgd->bgtr, bgd->bgntr, 
                       &bt, &onside, mtrf);
  if (sres == 0)
  {
    printf("\npoint with coords (%1.2f,%1.2f) falls outside bg mesh\n",pnt[0],pnt[1]);
    e_GeneralError("get_bg_mpr: Background mesh doesn't cover main mesh area");
  }
  else if (sres == -1)
  {
    if (bgd->bgwmpr == NODEMPRSET)
      dpc = bgd->bgmpr[onside];
    else if (bgd->bgwmpr == ELEMMPRSET)
    {
      /* this is the nasty case. what is the mpr of a triangle corner
         in a element mesh parameter mesh */
      dpc=mpr_of_point_in_elem_mpr_mesh(bt,onside,bgd->bgtr,bgd->bgntr,bgd->bgmpr);
    } 
  }
  else /* if (sres == 1) */
  {
    if (bgd->bgwmpr == ELEMMPRSET)
    {
      /* the mesh parameter is inherited from the element parameter
         of the background element the point falls in */
      if (onside == 0)
        dpc = bgd->bgmpr[bt];
      else /* point falls on a side somewhere */
        dpc = (bgd->bgmpr[bt] + bgd->bgmpr[bgd->bgtr[bt][3 + onside]]) / 2;
    }
    else if (bgd->bgwmpr == NODEMPRSET)
    {
      /* triangle coordinate system (area) interpolated mesh parameter */
      dpc = 0.0;
      area = 0.0;
      for (s = 0; s < 3; s++)
      {
        a = eg_TriangArea3D(bgd->bgcoords[bgd->bgtr[bt][(s + 1) % 3]],
                            bgd->bgcoords[bgd->bgtr[bt][(s + 2) % 3]], pnt);
        area += a;
        dpc += a * bgd->bgmpr[bgd->bgtr[bt][s]];
      }
      dpc /= area;
    }
  }
  
  if (ctrl == CTRL_SOURCES)
  {
    /* local sources interpolation */
    ldpc = -1.0;
    ldpc = localinterpolation(dpc, bgd->pcs, pnt[0], pnt[1]);
    if (ldpc > 0.0)
      dpc = ldpc;
  }
  
  return dpc;
}

static double mpr_of_point_in_elem_mpr_mesh(int bt,int s,int **tr,int ntr,double *mpr)
{
  /* returns the mpr of the s corner node of triangle bt in mesh tr
     with element mesh parameters mpr */
  
  int n=tr[bt][s];
  double dpc=0.0;
  int i,j;
  double ne=0.0;
  
  for (i=0;i<ntr;i++)
    for (j=0;j<3;j++)
      if (tr[i][j]==n)
      {
        dpc+=mpr[i];
        ne+=1.0;
      }
  
  return dpc/ne;
}


