/********************************************************************
 gentrianadj.c --
        Generalized adjaceny for Triang1 elements mesh.

 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University, Edinburgh
 ********************************************************************/

/* $Id: gentriangadj.c,v 1.2 2001/11/12 13:22:03 jelle Exp $ */

#include "gentriangadj.h"
#include "nodelem.h"

#ifndef __GNUC__
#  define __FUNCTION__  __FILE__
#endif

#define TRIANG1_ELEM_HAS_NODE(adjelem, node, found) \
{ \
  int i=0; \
  found = FALSE; \
  for (i=0; i<3; i++) \
  { \
    if (*(mesh->NodesTriang1[adjelem]+i) == node) \
    { \
      found = TRUE; \
      break; \
    } \
  } \
}

/* loog for minimum in vector of integers */
#define INT_VECT_MIN(vec, length, min) \
{ \
  int ivmi; \
  min =(vec)[0]; \
  for (ivmi=1; ivmi<(length); ivmi++) \
      if (min > (vec)[ivmi]) min = (vec)[ivmi]; \
}

/* definition of generic number of nodes on side */
static int sideDef[3][2] = {{0, 1},
                            {1, 2},
                            {2, 0}};

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_Triang1ElemsAdjToSide --
        For given Triang1 element 'elem' return all elements
        adjacent  to side 'side'. It returns number of adjacent
        elemetns and fills the vector of adjacent element numbers.

 Arguments:
        mesh - pointer to mesh of Triang1 elements   (IN)
        elem - element number                        (IN)
        side - side number (0, 1 or 2)               (IN)
        nodeElemAdjLen - vector (of lenght NMeshPoints) of lengths
                of rows of nodeElemAdj matrix        (IN)
        nodeElemAdj - nodes:->elements adjacency matrix (IN)
        sideAdj                                         (OUT)

 Return:
       Number of elements adjacent to side.

 Note:
        sideAdj should be a NULL pointer otherwise the memory pointed
        by it will be lost.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int Tos_Triang1ElemsAdjToSide (MESH *mesh,
                       int elem,
                       int side,
                       int *nodeElemAdjLen,
                       int **nodeElemAdj,
                       int **sideAdj)
{
  int node1, node2; /* nodes of the face */
  int found = FALSE;
  int adjelem, i;
  int adjcount=0;

  /* get the numbers of nodes on face */
  node1 = *(mesh->NodesTriang1[elem]+sideDef[side][0]);
  node2 = *(mesh->NodesTriang1[elem]+sideDef[side][1]);

  for (i=0; i < nodeElemAdjLen[node1-1]; i++)
  { /* for each element adjacent to node1 */
    adjelem = *(nodeElemAdj[node1-1]+i)-1;
    TRIANG1_ELEM_HAS_NODE (adjelem, node1, found);
    if (found == TRUE)
    {
      TRIANG1_ELEM_HAS_NODE (adjelem, node2, found);
      if (found == TRUE)
      {
        if (adjelem != elem)
        {
          (*sideAdj) = e_ext_i_vect (adjcount, (*sideAdj), 1);
          (*sideAdj)[adjcount] = adjelem+1;
          adjcount++;
        }
      }
    }
  }
  return adjcount;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_GenTriang1Adj
        Build adjacency matrix for triangular (Triang1) element mesh.
 Two elements are considered adjacent if they share  at least two
 nodes. This function can be safely called for meshes in which
 a side is shared by more than two elements. IF 'flag' is set
 to E_COMP_ADJ this function behaves like e_SetAdjacency (in case
 when more than two triangles share a side the adjacent triangle with
 the lowest number is the only one considered on adjacency list).
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        int
Tos_GenTriang1Adj (MESH *mesh, char flag, int **Degree, int ***AdjElems)
{
  int *nodeElemAdjLen; /* length of node->elements adjacency list */
  int **nodeElemAdj;   /* array of node->elements adjacency lists */
  int *sideAdj[3] = {NULL, NULL, NULL};  /* adjacency vector of each side */
  int sideAdjLen[3] = {0, 0, 0};    /* length of adjacency side vector */
  int i, side;

  if (!e_IsOfOneElementType(mesh))
  {
    e_GeneralError("%s: mesh contains more than one element type",
                    __FUNCTION__);
    return 0;
  }
  else if ( e_GetElementType(mesh) != E_ELEM_TYPE_TRIANG1)
  {
    e_GeneralError("%s: Mesh is not Triang1 mesh", __FUNCTION__);
    return 0;
  }

  /* allocate memory */
  *AdjElems = e_init_frag_i_matr (mesh->NElemsTriang1);
  if (*AdjElems == NULL)
    e_GeneralError ("%s: Cannot alocate memory (line %d)",
                    __FUNCTION__, __LINE__);

  *Degree = e_alloc_i_vect (mesh->NElemsTriang1);
  if (*Degree == NULL)
    e_GeneralError ("%s: Cannot alocate memory (line: %d)",
                    __FUNCTION__, __LINE__);

  nodeElemAdj = e_init_frag_i_matr (mesh->NNodes);
  if (nodeElemAdj == NULL)
    e_GeneralError ("%s: Cannot alocate memory (line %d)\n",
                    __FUNCTION__,  __LINE__);

  nodeElemAdjLen = e_alloc_i_vect (mesh->NNodes);
  if (nodeElemAdjLen == NULL)
    e_GeneralError ("%s: Cannot alocate memory (line %d)\n",
                    __FUNCTION__, __LINE__);

  /* set the nodes to elements adjacency */
  e_InitMeshPointElemConnection (mesh, nodeElemAdjLen, nodeElemAdj);

  for (i=0; i<mesh->NElemsTriang1; i++)
  {
    for (side=0; side<3; side++)
    {
      if (sideAdj[side] != NULL)
      {
        e_free ((void *) sideAdj[side]);
        sideAdjLen[side] = 0;
        sideAdj[side] = NULL;
      }
      sideAdjLen[side] = Tos_Triang1ElemsAdjToSide (mesh, i, side,
                     nodeElemAdjLen, nodeElemAdj, &sideAdj[side]);
    }
    if (flag == E_COMP_ADJ)
    {
      int k=0;
      for (side=0; side<3; side++)
      {
        if (sideAdjLen[side] != 0)
        {
          (*AdjElems)[i] = e_ext_i_vect ((*Degree)[i], (*AdjElems)[i], 1);
          INT_VECT_MIN(sideAdj[side], sideAdjLen[side], *((*AdjElems)[i]+k));
          k++;
          (*Degree)[i]++;
        }
      }
    }
    else if (flag == TOS_COMP_ADJ)
    {
     int len=sideAdjLen[0]+sideAdjLen[1]+sideAdjLen[2];
     (*AdjElems)[i] = e_alloc_i_vect (len);
     if (sideAdj[0] != NULL)
        memcpy ((*AdjElems)[i], sideAdj[0], sideAdjLen[0]*sizeof(int));
     if (sideAdj[1] != NULL)
          memcpy ((*AdjElems)[i]+sideAdjLen[0], sideAdj[1],
                     sideAdjLen[1]*sizeof(int));
     if (sideAdj[2] != NULL)
          memcpy ((*AdjElems)[i]+sideAdjLen[0]+sideAdjLen[1], sideAdj[2],
                     sideAdjLen[2]*sizeof(int));
     (*Degree)[i] = len;
    }
    else
    {
      e_GeneralError ("%s: Unknown flag %d", __FUNCTION__, flag);
    }
  }

  /* free memory */
  e_free_frag_i_matr (mesh->NNodes, nodeElemAdj);
  e_free ((void *) nodeElemAdjLen);
  for (side=0; side<3; side++)
  {
    if (sideAdj[side] != NULL)
    {
      e_free ((void *) sideAdj[side]);
    }
  }
  return 1;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Tos_FreeGenTriang1Adj --
        Free memory allocated by Tos_GenTriang1Adj.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        void
Tos_FreeGenTriang1Adj (MESH *mesh, int *Degree, int **AdjElems)
{
   e_free ((void *) Degree);
   e_free_frag_i_matr (mesh->NElemsTriang1, AdjElems);

   return;
}

/* Modifications:
 * $Log: gentriangadj.c,v $
 * Revision 1.2  2001/11/12 13:22:03  jelle
 * major revision, made in France
 *
 * Revision 1.1.1.1  2001/04/09 22:50:37  peteri
 * Imported using TkCVS
 *
 * Revision 1.1  2000/02/08 19:07:24  romek
 * Initial revision
 *
 */
