#include "e_lib.h"
#include "mgn_aux.h"
#include "eg_lib.h"

int IsCCW(double *A, double *B, double *C, 
          MTRF *mtrf, int *IsValidTriangle)
{
  /* is triangle ABC CCW ? and is it a valid triangle? */
  
  double normal[3];
  double d;
  int res;
  
  if (!mtrf->used)
    d = eg_Orient2D(A, B, C);
  else
  {
    d = eg_TriangArea3D(A, B, C);
    res=eg_TriangNormal3D(A, B, C, normal);
    if (res==EG_INVALID)
      d = 0.0;
    else if (e_VectDotProd(3, mtrf->w, normal) < 0.0)
      d *= -1.0;
  }
  
  if (E_ABS(d)<eg_GetTolerance(EG_ZERO_AREA_TOL))
    *IsValidTriangle=FALSE;
  else
    *IsValidTriangle=TRUE;
  
  
  if (d < eg_GetTolerance(EG_ZERO_AREA_TOL))
    return FALSE;
  else 
    return TRUE;
}

int ColinearTriangle(double *A, double *B, double *C, 
          MTRF *mtrf, double *mu)
{
  /* sortout what the situation is with with 3 collinear points
     return values:
     FALSE: if A and B are equal or ABC not collinear
     TRUE: else

     if TRUE the mu coordinate of C on the base AB is returned
  */
  
  double dAB;		/* distance(A,B) */
  double pd;		/* perpendicular distance from c to AB */
  
  if (mtrf->used)
    dAB=eg_PointDist3D(A,B);
  else
    dAB=eg_PointDist2D(A,B);
    
  if (dAB<eg_GetTolerance(EG_ZERO_DIST_TOL))
    return FALSE;
  
  if (mtrf->used)
    pd=eg_PointLineDist3D(A,B,C,mu);
  else
    pd=eg_PointLineDist2D(A,B,C,mu);
    
  /* I commented this sanity check out because the tolerance 
     isn't properly fitted with the tolerance in IsCCW 
  if (pd>eg_GetTolerance(EG_ZERO_DIST_TOL))
    return FALSE; */
    
  return TRUE;
}

int j_incircle(double *A, double *B, double *C, double *D, MTRF * mtrf)
{
  double d;
  double *At, *Bt, *Ct, *Dt;

  if (mtrf->used)
  {
    At = e_alloc_d_vect(3);
    Bt = e_alloc_d_vect(3);
    Ct = e_alloc_d_vect(3);
    Dt = e_alloc_d_vect(3);
    if (!mtrf_map(A, mtrf, At))
      e_GeneralError("mapping problem");
    if (!mtrf_map(B, mtrf, Bt))
      e_GeneralError("mapping problem");
    if (!mtrf_map(C, mtrf, Ct))
      e_GeneralError("mapping problem");
    if (!mtrf_map(D, mtrf, Dt))
      e_GeneralError("mapping problem");
    d = eg_InCircle(At, Bt, Ct, Dt);
    e_free(At);
    e_free(Bt);
    e_free(Ct);
    e_free(Dt);
  }
  else
  {
    d = eg_InCircle(A, B, C, D);
  }

  return (d > 0.0);
}

void centroid(double **cntr, double *A, double *B, double *C)
{
  /* computes the centroid of a CCW triangle ABC */

  int i;

  for (i = 0; i < 3; i++)
    (*cntr)[i] = (A[i] + B[i] + C[i]) / 3.0;
  if (eg_TriangArea3D(A,B,C)<eg_GetTolerance(EG_ZERO_AREA_TOL))
    e_GeneralWarning("centroid: centroid of zero area triangle requested");
}

int neighbouringside(int **tr, int t, int nt, int *n, int *s, int relax_adj)
{
  /* in triangle matrix tr, t is a neighbour of nt.
   * it has nt on its side n and
   * nt has t on its side s.
   * 
   * the function return false if t and nt are not neighbours.
   * with relax_adj=TRUE there is no requirement for shared
   * edges to have oposite orientation (e.g when having non-uniform
   * orientation of 3D triangles)
   */

  int i, j;                     /* counters */
  int found;                    /* a match was made */

  found = FALSE;
  for (i = 0; i < 3; i++)
  {
    for (j = 0; j < 3; j++)
    {
      if (tr[t][i] == tr[nt][j])
      {
        if ((tr[t][(i + 1) % 3] == tr[nt][(j + 2) % 3])||
           ((relax_adj)&&(tr[t][(i + 1) % 3] == tr[nt][(j + 1) % 3])))
        {
          found = TRUE;
          break;
        }
      }
    }
    if (found)
      break;
  }

  if (found)
  {
    (*n) = 3 + i;
    if (tr[t][(i + 1) % 3] == tr[nt][(j + 2) % 3])
      (*s) = 3 + ((j + 2) % 3);
    else
      (*s) = 3 + j;
    return TRUE;
  }
  
  return FALSE;
}

int CCW_check(int **tr, int ntr, double **coords)
{
  int t;
  MTRF mtrf;
  int valid;

  mtrf.used=FALSE;

  for (t = 0; t < ntr; t++)
  {
    if (!IsCCW(coords[tr[t][0]], coords[tr[t][1]], coords[tr[t][2]],
              &mtrf,&valid))
      break;
    if (!valid)
      break;
  }

  if (t!=ntr)
  {
    if (!valid)
      printf("element %d is not valid\n", t+1);
    else
      printf("element %d is not CCW\n", t + 1);
    return FALSE;
  }

  return TRUE;
}


/* end of file */
