#include "e_lib.h"
#include "eg_lib.h"
#include "pcs_int.h"

static double point_source_interpolation(double bgmpr,
  PARAMCNTL *pc,int s,double *pnt);
static double line_source_interpolation(double bgmpr,
  PARAMCNTL *pc,double *pnt);

double localinterpolation(double bgmpr, PARAMCNTLSETS *pcs,
                          double x, double y)
{
  /* modify the background mesh parameter to account for sources.
     only take into account the first source that applies! */
     
  /* fatal flaw: the bgmpr is assumed to apply over the whole
     area of the source */

  int s;                /* set counter */
  double mpr;           /* interpolated mesh parameter */
  double pnt[3];
  
  /* sanity check */
  if (bgmpr<=0.0)
    e_GeneralError("localinterpolation: bgmpr<=0.0");
  
  /* except for this top function (easily update-able) all will be 3D */
  pnt[0]=x;
  pnt[1]=y;
  pnt[2]=0.0;
  
  /* go through all sets */
  for (s=0;s<pcs->NCntlSets;s++)
  {
    if (pcs->ParamCntls[s].NCntlPoints==1)
    {
      mpr=point_source_interpolation(bgmpr,&(pcs->ParamCntls[s]),0,pnt);
      if (mpr>0.0)
        return mpr;
    }
    else
    {
      mpr=line_source_interpolation(bgmpr,&(pcs->ParamCntls[s]),pnt);
      if (mpr>0.0)
        return mpr;
    }
  }
  
  /* if no sets apply just return the background mpr */
  return bgmpr;
}

static double point_source_interpolation(double bgmpr,
  PARAMCNTL *pc,int s,double *pnt)
{
  /* interpolation routine for point source s 
     negative return value means source does not apply */
  
  double d;							/* distance from point to source center */
  double mpr;						/* the resultant interpolated mesh parameter */
  
  d=eg_PointDist3D(pnt,pc->Coords[s]);
  
  /* point lies outside source */
  if (d>pc->OuterRadius[s])
    return -1.0;
  
  /* point lies in inner core of source */
  if (d<pc->InnerRadius[s])
    return pc->Param[s];
  
  /* point lies in interpolation zone of source */
  mpr = pc->Param[s]+(bgmpr- pc->Param[s])*(d-pc->InnerRadius[s])/
    		(pc->OuterRadius[s]-pc->InnerRadius[s]);
  return mpr;
}

static double line_source_interpolation(double bgmpr,
  PARAMCNTL *pc,double *pnt)
{
  /* interpolation routine for line sources
     negative return value means source does not apply */
  
  double mpr;			/* the resultant interpolated mesh parameter */
  int nas;				/* number of applicable line segments */
  int seg;				/* segment counter */
  double pd;			/* perpendicular distance from point to segment */
  double mu;			/* the location of the point drawn on the base of the seg */
  double ir,or;		/* inner and outer radius */
  double sm;			/* segment interpolated mesh parameter */
  int s;					/* source counter */
  
  mpr=0.0;
  nas=0;
  
  /* check the influence region of each line segment */
  for (seg=0;seg<pc->NCntlPoints-1;seg++)
  {
    pd=eg_PointLineDist3D(pc->Coords[seg],pc->Coords[seg+1],pnt,&mu);
    if ((mu<0.0)||(mu>1.0))
      continue;
    or=pc->OuterRadius[seg]+mu*(pc->OuterRadius[seg+1]-pc->OuterRadius[seg]);
    if (pd>or)
      continue;
    ir=pc->InnerRadius[seg]+mu*(pc->InnerRadius[seg+1]-pc->InnerRadius[seg]);
    sm=pc->Param[seg]+mu*(pc->Param[seg+1]-pc->Param[seg]);
    if (pd<ir)
      mpr+=sm;
    else
      mpr+=sm+(bgmpr-sm)*(pd-ir)/(or-ir);
    nas++;
  }
  if (mpr>0.0)
    return (mpr/nas);
  
  /* else check the individual point sources */
  for (s=0;s<pc->NCntlPoints;s++)
  {
    mpr=point_source_interpolation(bgmpr,pc,s,pnt);
    if (mpr>0.0)
      return mpr;
  }
  
  /* no applicable sources found */
  return -1.0;
}
