#include "eg_lib.h"
#include "mgn.h"
#include "mgn_graphics.h"
#include "dtr.h"
#include "bgdata.h"
#include "mgn_aux.h"

#define DEL_CHECK FALSE

static double get_mesh_parameter(double *cntr,int ct,int **tr,
                                 double **crds, double *mpr);
static int get_opt_ins_pos(double **cntr, int **tr, int t, int *ct,
                    double **coords, MTRF *mtrf);

/* experimental delaunay refinement functions */
#include "dtr_f2.c"


#ifndef WIN32
/* the randomize function doesn't work on PC's */

void f_randomize_mesh(MESH *mesh)
{
  int i, p, pt, r, t, q;
  int *a;
  double **d;

  if (DebugOutput)
    printf("\nRandomization of the input nodes\n\n");
  a = e_alloc_i_vect(mesh->NNodes);
  for (i = 0; i < mesh->NNodes; i++)
    a[i] = i + 1;
  a[mesh->NNodes - 1] = 0;
  t = 0;
  for (i = 0; i < mesh->NNodes; i++)
  {
    r = (int) rint(drand48() * (mesh->NNodes - i));
    for (p = 0; p < r + 1; p++)
    {
      pt = t;
      t = a[t];
    }
    if (DebugOutput)
      printf("r=%d a[%d]=%d\n", r, t, i);
    a[pt] = a[t];
    q = t;
    t = a[t];
    a[q] = i;
  }
  d = e_alloc_cont_d_matr(mesh->NNodes, 3);
  for (i = 0; i < mesh->NNodes; i++)
  {
    d[a[i]][0] = mesh->MeshPointCoords[i][0];
    d[a[i]][1] = mesh->MeshPointCoords[i][1];
    d[a[i]][2] = mesh->MeshPointCoords[i][2];
  }
  for (i = 0; i < mesh->NNodes; i++)
  {
    mesh->MeshPointCoords[i][0] = d[i][0];
    mesh->MeshPointCoords[i][1] = d[i][1];
    mesh->MeshPointCoords[i][2] = d[i][2];
  }
  for (i = 0; i < mesh->NElemsTriang1; i++)
  {
    mesh->NodesTriang1[i][0] = a[mesh->NodesTriang1[i][0] - 1] + 1;
    mesh->NodesTriang1[i][1] = a[mesh->NodesTriang1[i][1] - 1] + 1;
    mesh->NodesTriang1[i][2] = a[mesh->NodesTriang1[i][2] - 1] + 1;
  }
  e_free(a);
  e_free_cont_d_matr(mesh->NNodes, d);
}

#endif

int f_compute_convex_hull(double **coords, int nnodes, int **h, MTRF *mtrf,
                          int meth)
{
  /* a convex hull creation algorithm in 2D using the algorithm
   *
   * meth = 1 -> Incremental
   * meth = 2 -> Gift wrap
   * meth = 3 -> Divid and conquer not yet implemented
   * meth = 4 -> Quickhull  not yet implemented
   *
   */

  int Nhull;                    /* number of elements on the hull */
  int *hull;                    /* the hull */
  int i, j;                     /* counters */
  int ily;                      /* index of lowest y coord node yet */
  int *oh;                      /* already on the hull vector */
  int cn;                       /* candidate of next hull point */
  double dx, dy, dx2, dy2, d1, d2, p; /* decission variables */
  int ccw, valid;               /* IsCCW results */
  double mu;                    /* ColinearTriangle result */
  int end;                      /* termination of the hull searching */
  int outside;                  /* whether an inserted point is outside the current hull */
  int ossi, osei;               /* outside start and end hull vector indices */
  int falsestart;               /* indicates that ossi motght be wrong */

  if (DebugOutput)
    printf("\nDetermination of the convex hull started\n\n");

  Nhull = 0;
  oh = NULL;

  if (meth == 1)                /* Incremental algorithm */
  {
    if (DebugOutput)
      printf("\nMethod: Incremental algorithm\n\n");

    hull = e_alloc_i_vect(INCR_SIZE);
    Nhull = 3;
    hull[0] = 0;

    /* find a starting triangle */
    cn = 1;
    valid=FALSE;
    while (!valid)
    {
      cn++;                     /* so it starts with cn=2 */
      ccw=IsCCW(coords[0], coords[1], coords[cn], mtrf, &valid);
    }
    if (ccw)
    {
      hull[1] = 1;              /* build a starting CCW triangle */
      hull[2] = cn;
    }
    else
    {
      hull[1] = cn;
      hull[2] = 1;
    }

    /* and add all other points */
    for (i = 2; i < nnodes; i++)
    {
      if (i != cn)
      {
        ossi = -1;
        osei = -1;
        falsestart = FALSE;
        for (j = 0; j < Nhull; j++)
        {
          /* i is outside current hull if it is really outside or on it */
          outside = !IsCCW(coords[hull[j]], coords[hull[(j + 1)%Nhull]], coords[i], mtrf, &valid);
          if (!valid)
          {
            if (!ColinearTriangle(coords[hull[j]], coords[hull[(j + 1)%Nhull]], coords[i],
                                  mtrf, &mu))
              e_GeneralError("f_compute_convex_hull: improper call to ColinearTriangle");
            if ((mu<0.0)||(mu>1.0))
              outside=TRUE;
          }
          if (outside)
          {
            if ((falsestart) && (osei != -1))
            {
              ossi = j;
              falsestart = FALSE;
            }
            else if (ossi == -1)
            {
              ossi = j;
              if (j == 0)
                falsestart = TRUE;
            }
          }
          else if ((ossi != -1) && (osei == -1))
            osei = j;
        }
        if (ossi != -1)
        {

          if (osei == -1)
            osei = 0;
          update_hull(&hull, &Nhull, ossi, osei, i);
        }
      }
    }
  }
  else if (meth == 2)           /* Gift wrap algorithm */
  {
    if (DebugOutput)
      printf("\nMethod: Gift wrap algorithm\n\n");

    ily = 0;
    for (i = 0; i < nnodes; i++)
      if (coords[i][1] < coords[ily][1])
        ily = i;
    hull = e_alloc_i_vect(1);
    hull[0] = ily;
    oh = e_alloc_i_vect(nnodes);
    for (i = 0; i < nnodes; i++)
      oh[i] = 0;
    oh[ily] = 1;
    end = FALSE;

    while (!end)
    {
      cn = 0;
      while (oh[cn] == 1)
        cn++;
      for (i = 0; i < nnodes; i++)
        if ((oh[i] == 0) && (i != cn))
        {
          ccw=IsCCW(coords[i], coords[cn], coords[hull[Nhull]], mtrf, &valid);
          if (ccw && valid)        /* i right of line through hull[Nhull] and cn */
            cn = i;
          else if (!valid)         /* i on line through hull[Nhull] and cn */
          {
            dy = (coords[i][1] - coords[hull[Nhull]][1]);
            dx = (coords[i][0] - coords[hull[Nhull]][0]);
            d1 = dx * dx + dy * dy;
            dy2 = (coords[cn][1] - coords[hull[Nhull]][1]);
            dx2 = (coords[cn][0] - coords[hull[Nhull]][0]);
            d2 = dx2 * dx2 + dy2 * dy2;
            if (E_ABS(dx2) < E_ABS(dy2))
              p = dy / dy2;
            else
              p = dx / dx2;
            if ((d1 < d2) && (p > 0)) /* i closer to hull[Nhull] than cn */
              cn = i;
          }
        }
      if (Nhull > 0)            /* searching for closing the hull */
      {
        ccw=IsCCW(coords[hull[0]], coords[cn], coords[hull[Nhull]], mtrf, &valid);
        if (valid)          /* hull[0] right of line through hull[Nhull] and cn */
        {
          cn = hull[0];
          end = TRUE;
        }
      }
      Nhull++;
      hull = e_incr_i_vect(Nhull, hull, 1);
      hull[Nhull] = cn;
      oh[cn] = 1;
    }
  }
  else
    e_GeneralError("\nnot implemented hull creation algorithm chosen\n");
  if (PrintProgress)
    printf("Convex hull created\n\n");
  *h = e_alloc_i_vect(Nhull);
  for (i = 0; i < Nhull; i++)
    (*h)[i] = hull[i];
  e_free(hull);
  e_free(oh);
  return Nhull;
}

void f_connect_mesh(double **coords, int Np, int ***triangles,
                    int *ntriangles, int **convhull, int *nhullelems,
                    int **tpnl, MTRF *mtrf)
{
  /* This function triangulates the Np nodes of which coordinates
   * are to be found in coords. It initializes triangle matrix triangles.
   * Creates a convex hull, border table and tpnl list
   */

  int i, j, t, k;               /* counters */
  int Ntr;                      /* number of triangles to be created */
  int c0, c1, c2;               /* the nodes of the first triangle */
  int **tr;                     /* triangle matrix */
  int onside;                   /* index of side point lies on */
  int cn;                       /* candidate node for first triangle */
  int ccw,valid;                /* IsCCW results */

  int *efwl;                    /* edge flip wait list */
  int nwl;                      /* length of the wait list */
  int twl;                      /* free spaces in the efwl */

  int last;                     /* last found triangle */
  int sres;                     /* result of the search operation */

  int Nhull;                    /* number of points on the convex hull */
  int *hull;                    /* vector containing the node indices that form the hull */

  /* debug variables */
  char ch;

  /* compute the hull, then we know how many triangles we will make */
  Nhull = f_compute_convex_hull(coords, Np, &hull, mtrf, 1);
  if (DebugOutput)
  {
    printf("\nNhull=%d\n\n", Nhull);
    for (i = 0; i < Nhull + 1; i++)
      printf("hull[%d]=%d\n", i, hull[i]);
  }
  e_free(hull);
  Ntr = 2 * (Np - 1) - Nhull;   						/* according to D.T. Lee, Two Alg..., p221 */
  if (!(tr = e_alloc_cont_i_matr(Ntr, 11))) /* 3 nodes + 3 indices of adjacent triangles */
    e_GeneralError("memory error\n"); 			/* + 3 indices of the side in the adjacent triangle */
  																					/* + 1 active/inactive flag for refining */
  																					/* and 1 areacode */

  /* initialize the first triangle */
  Ntr = 1;

  tr[0][0] = 0;
  cn = 1;
  valid = FALSE;
  while (!valid)
  {
    cn++;
    ccw=IsCCW(coords[0], coords[1], coords[cn], mtrf, &valid);
  }
  if (ccw)
  {
    tr[0][1] = 1;
    tr[0][2] = cn;
  }
  else
  {
    tr[0][1] = cn;
    tr[0][2] = 1;
  }
  tr[0][9] = 1;
  tr[0][10] = 0;

  /* reinitialize the hull */
  if (!(hull = e_alloc_i_vect(INCR_SIZE)))
    e_GeneralError("memory error\n");
  Nhull = 3;
  c0 = tr[0][0];
  c1 = tr[0][1];
  c2 = tr[0][2];
  hull[0] = c0;
  hull[1] = c1;
  hull[2] = c2;

  /* initialize the tpnl */
  addtr2tpnlist(tr, 0, tpnl);

  /* initialize the border */
  for (i = 0; i < 3; i++)
  {
    tr[0][i + 3] = -1;          /* initialize adjacency info */
    /* indices 6,7 & 8 n/a if [3],[4] & [5]==-1 */
  }
  twl = INCR_SIZE;              /* initialize the edge flip wait list */
  efwl = e_alloc_i_vect(twl);   /* edge flip wait list side 4 to be checked */
  nwl = 0;

  last = 0;                     /* last found triangle */

  for (i = 0; i < Np; i++)
  {
    if ((i != c0) && (i != c1) && (i != c2))
    {
      sres = trianglesearch(coords[i], coords, tr, last, &t, &onside, mtrf);
      if (sres == 1)            /* triangle found */
      {
        last = t;
        insert_node(i, t, &Ntr, tr, &efwl, &nwl, &twl, onside,
                    &hull, &Nhull, tpnl);
      }
      else if (sres == 0)
      {
        if (DebugOutput)
          printf("\noutside connection !!\n");
        connect_outside_node(i, coords, &Ntr, tr, &efwl, &nwl, &twl, &hull,
                             &Nhull, tpnl, mtrf);
      }
      edge_flipping(i, coords, &efwl, &nwl, &twl, tr, tpnl, mtrf);
    }

    /* some debug output */
    if (DebugOutput)
    {
      for (k = 0; k < Ntr; k++)
        printf("\ntr[%d]: %d   %d   %d  /  %d   %d   %d  /  %d   %d   %d",
               k, tr[k][0], tr[k][1], tr[k][2], tr[k][3], tr[k][4],
               tr[k][5], tr[k][6], tr[k][7], tr[k][8]);
      printf("\n\n");
      for (k = 0; k < Np; k++)
      {
        printf("tpnl[%d]=", k);
        for (j = 0; j < tpnl[k][0]; j++)
          printf("%d ", tpnl[k][2 + j]);
        printf("\n");
      }
      printf("\n\n");
      if ((ch = getchar()) != '\n')
        break;
    }

    /* a serious but very time-consuming Delaunay check */
    if (DEL_CHECK)
      if (!delaunay_check(tr, Ntr, coords, i + 1))
      {
        printf("\n\nwhile inserting point %d\n", i);
        printf("not Delaunay triangles found\n");
        e_GeneralError("\n");
      }
  }

  if (DebugOutput)
  {
    for (k = 0; k < Nhull; k++)
      printf("hull[%d]=%d\n", k, hull[k]);
  }

  /* pass all computed information */
  (*triangles) = e_alloc_cont_i_matr(Ntr, 11);
  (*convhull) = e_alloc_i_vect((int)ceil((double)Nhull/(double)INCR_SIZE)*INCR_SIZE);
  for (i = 0; i < Ntr; i++)
    for (j = 0; j < 11; j++)
      (*triangles)[i][j] = tr[i][j];
  for (i = 0; i < Nhull; i++)
    (*convhull)[i] = hull[i];
  (*ntriangles) = Ntr;
  (*nhullelems) = Nhull;

  /* free up all consumed memory */
  e_free_cont_i_matr(Ntr, tr);
  e_free(hull);
  e_free(efwl);
}

void f_refine_del_mesh(double ***coords, int *np, int *tcs, int ***triangles,
                       int *ntriangles, int *tts, int **convhull,
                       int *nhullelems, int ctrl, double alfa,
                       double beta, BG_DATA *bgd, double **mpr, double h,
                       int ***tpnl, BND *bnd, int dln, MTRF *mtrf)
{
  /* This function refines an edge node only mesh by inserting new
   * nodes.
   * These new nodes are made according to the delaunay option passed in dln.
   *
   * The mesh density is controlled by the controltype set in the ctrl argument.
   * Depending on this other arguments passed or not.
   */

  int i, s;                     /* counters */
  int onside;                   /* onside index of the triangle search */
  int t;                        /* index of an existing triangle */
  double *cntr;                 /* centre of an existing triangle */
  double dpc;                   /* the mesh parameter in the prospective node */
  double tdpc[3];               /* the mesh parameter of the corner nodes
                                   of the originating triangle */
  double ds;                    /* distance from centr to node */
  int include;                  /* wether or not to include this point */
  int sres;                     /* result of the search operation */
  double **inslist;             /* contains nodes accepted for insertion */
  int ninsl;                    /* number of nodes in the insertion list */
  int tinsl;                    /* total allocated space in the inslist */
  int refine;                   /* keep on refining? yes/no */
  int n;                        /* index of a new point while inserting */
  int *efwl;                    /* edge flip wait list */
  int nwl;                      /* length of the wait list */
  int twl;                      /* total allocated space in the efwl */
  int bgstarttr;                /* start triangle for a search of background triangles */
  double smallestmpr;           /* for tracking purposes */
  int ct;												/* triangle the candidate point belongs to */

  int nsteps;                   /* number of refining steps */

	if (dln==DLN_CHEW1)
		return delref_chew1(coords, np, tcs, triangles, ntriangles, tts, convhull,
         					      nhullelems, ctrl, h, tpnl, bnd, mtrf);

  if (DebugOutput)
    printf("refining with alfa=%1.1f beta=%1.1f\n", alfa, beta);

  cntr = e_alloc_d_vect(3);
  ninsl = 0;                    /* initializing the insertion list */
  tinsl = INCR_SIZE;
  inslist = e_alloc_cont_d_matr(tinsl, 5); /* the three coords ,the pdf,
                                              and the originating triangle */
  twl = INCR_SIZE;              /* initialize the edge flip wait list */
  efwl = e_alloc_i_vect(twl);   /* edge flip wait list side 4 to be checked */
  nwl = 0;

  /* initialising the active/passive flags
   * inactive means triangle passed the alfa criterion (enough refined)
   * active means triangle is new, either created or from flipping
   */

  nsteps = 0;
  refine = TRUE;
  /* as long as there are triangles to be refined */
  while (refine)
  {
    nsteps++;
    if (PrintProgress)
      printf("refining step number: %d ", nsteps);

    bgstarttr = 0;              /* initializing the search variable for all triangle searches */
    if (ShowGraphics)           /* a little bit of graphics output */
    {
      mgnDrawAll(*np, *coords, *ntriangles, *triangles, 0, NULL, 0,
                 NULL, NULL, "refining");
    }

    for (t = 0; t < (*ntriangles); t++)
    {
      if (((*triangles)[t][9] == 1) && ((*triangles)[t][10] != -1))
      /* for those triangles that are active and inside the domain */
      {
        /* first determine the location of the candidate insertion point */
        if ((dln == DLN_CENTR) || (dln == DLN_CENTR_RAD))
        {
          /* for these options it is just the centroid of the originating triangle */
          centroid(&cntr, (*coords)[(*triangles)[t][0]],
                          (*coords)[(*triangles)[t][1]],
                          (*coords)[(*triangles)[t][2]]);
          ct = t;
        }
        else if (dln == DLN_OPT)
        {
          /* for the optimal insertion position */
          get_opt_ins_pos(&cntr, *triangles, t, &ct, *coords, mtrf);
        }

        /* now determine the mesh parameter */
        if (ctrl == CTRL_MPR)
          dpc=get_mesh_parameter(cntr,ct,*triangles,*coords,*mpr);
        else if (ctrl == CTRL_BACKGROUND || ctrl == CTRL_SOURCES)
          dpc=get_bg_mpr(bgd,cntr,ctrl,mtrf);
        else if (ctrl == CTRL_UNIFORM)
        	dpc=h;

        /* and at the same time derive the mesh parameter of the three corner
           nodes of the originating triangle */
        for (s=0;s<3;s++)
        {
          if (ctrl == CTRL_MPR)
            tdpc[s] = (*mpr)[(*triangles)[ct][s]];
          else if (ctrl == CTRL_BACKGROUND || ctrl == CTRL_SOURCES)
            tdpc[s]=get_bg_mpr(bgd,(*coords)[(*triangles)[ct][s]],ctrl,mtrf);
          else if (ctrl == CTRL_UNIFORM)
          	tdpc[s]=h;
        }

        /* the criteria for candidate nodes are checked */
        /* setting the initial values */
        if (dln == DLN_CENTR_RAD)
          include = TRUE;
        else
          include = FALSE;
        /* checking the so called alfa criterion */
        for (s = 0; s < 3; s++)
        {
          if (dln == DLN_CENTR_RAD)
          {
            /* if any of the three connecting lines from candidate
               to corners is too big: candidate passes */
            ds = eg_PointDist3D((*coords)[(*triangles)[ct][s]], cntr);
                        /***** if (ds<alfa*dpc) *****/
            if (ds < alfa * tdpc[s])
            {
              include = FALSE;
              break;
            }
          }
          else                  /* dln option DLN_CENTROID and DLN_OPT */
          {
            /* if triangle it belongs to has a side which is not ok: the candidate passes */
            ds = eg_PointDist3D((*coords)[(*triangles)[ct][s]],
                                (*coords)[(*triangles)[ct][(s + 1) % 3]]);
                        /***** if (ds>alfa*dpc) *****/
            if (ds > alfa * E_MIN(tdpc[s],tdpc[(s+1)%3]))
            {
              include = TRUE;
              break;
            }
          }
        }

        /* if a triangle passes the alfa criterion it's set to passive ! */
        if (!include)
          (*triangles)[ct][9] = 0;

        /* for the remaining candidates: the second round */
        if (include)
          for (s = 0; s < ninsl; s++)
          {
            ds = eg_PointDist3D(inslist[s], cntr);
            /* candidates which are too close to EACH OTHER are rejected in second round */
                        /***** if (ds<dpc/beta) ******/
            if (ds < beta * dpc)
            {
              include = FALSE;
              break;
            }
          }

        /* for those who survived both criteria */
        if (include)
        {
          /* ok put candidate in list to get inserted */
          if (tinsl == ninsl)
          {
            /*  make some more space in the list */
            inslist = e_incr_cont_d_matr(tinsl, 5, inslist, tinsl);
            tinsl += ninsl;
          }
          for (s = 0; s < 3; s++)
            inslist[ninsl][s] = cntr[s]; /* the new coordinates */
          inslist[ninsl][3] = dpc; /* the new mpr value */
          inslist[ninsl][4] = t; /* the originating triangle */
          ninsl++;
        }
      }
    }

    /* so all new nodes are determined */
    if (PrintProgress)
      printf(" inserted centres: %d\n", ninsl);

    /* if there are no new nodes you can terminate the refinement procedure here */
    if (ninsl == 0)
      refine = FALSE;
    else
    {
      /* otherwise insert them into the triangulation */
      while ((*tcs) - (*np) < ninsl)
      {
        /* make more space in the arrays (coords etc..) */
        (*coords) = e_incr_cont_d_matr((*tcs), 3, (*coords), (*tcs));
        if (ctrl == CTRL_MPR)
          (*mpr) = e_incr_d_vect((*tcs), (*mpr), (*tcs));
        (*tpnl) = (int **) e_incr_vp_vect((*tcs), (void **) (*tpnl), (*tcs));
        (*tcs) += (*tcs);
      }
      /* it's good to monitor the smallest mpr */
      smallestmpr = inslist[0][3];

      for (n = 0; n < ninsl; n++)
      {
        /* update the coordinates array */
        for (i = 0; i < 3; i++)
          (*coords)[(*np) + n][i] = inslist[n][i];
        /* initialize the new entries in the tpnl */
        (*tpnl)[(*np) + n] = e_alloc_i_vect(2);
        (*tpnl)[(*np) + n][0] = 0;
        (*tpnl)[(*np) + n][1] = 0;
        if (ctrl == CTRL_MPR)
        {
          /* update the mpr array */
          (*mpr)[(*np) + n] = inslist[n][3];
          if ((*mpr)[(*np) + n] < smallestmpr)
            smallestmpr = (*mpr)[(*np) + n];
        }
        /* insert the new node */
        if ((*tts) - (*ntriangles) < 4)
        {
          /* make more space for new triangles (maximum 4 new) */
          (*triangles) = e_incr_cont_i_matr((*tts), 11, (*triangles), (*tts));
          (*tts) += (*tts);
        }
        sres =
          trianglesearch((*coords)[(*np) + n], (*coords), (*triangles),
                         inslist[n][4], &t, &onside, mtrf);
        if (sres == 0)
          e_GeneralError("refinement node can't be outside the domain!");
        if (sres == -1)
          e_GeneralError
            ("refinement node is duplicate and so far undetected ?!?");
        else                    /* the normal case! */
          insert_node((*np) + n, t, ntriangles, (*triangles), &efwl, &nwl,
                      &twl, onside, convhull, nhullelems, (*tpnl));
        edge_flipping((*np) + n, (*coords), &efwl, &nwl, &twl, (*triangles),
                      (*tpnl), mtrf);
      }
      /* monitoring of the smallest mpr */
      if ((DebugOutput) && (ctrl == CTRL_MPR))
        printf(" smallest mpr=%f\n", smallestmpr);
      (*np) += ninsl;
      ninsl = 0;
    }
  }

  /* free up that memory */
  e_free_cont_d_matr(tinsl, inslist);
  e_free(cntr);
  e_free(efwl);
}

static double get_mesh_parameter(double *cntr,int ct,int **tr,
                                 double **crds, double *mpr)
{
  double a,area;		  /* area for area based interpolation */
  double dpc;					/* the return value */
  int s;							/* counter */

  dpc = 0.0;
  area = 0.0;
  for (s = 0; s < 3; s++)
  {
    a = eg_TriangArea3D(crds[tr[ct][(s + 1) % 3]],
                        crds[tr[ct][(s + 2) % 3]], cntr);
    area += a;
    dpc += a * mpr[tr[ct][s]];
  }
  dpc /= area;

  return dpc;
}

static int get_opt_ins_pos(double **cntr, int **tr, int t, int *ct,
                    double **coords, MTRF *mtrf)
{
  /* determine the optimal position for point insertion
     using projection with as base triangle t.

     the result is put in cntr and lies in triangle ct.
     unless the return value is false in which case the
     generated point accidentaly resulted in an already existing point.
     or the base triangle t satisfies criteria.

     this function is NOT 3D compatible! */

  int nsides_ok;
  int sides_ok[3];
  int s,i;
  int sres;
  int onside;

  double ds,l;
  double v1[3],v2[3],v3[3];

  *cntr=e_alloc_d_vect(3);
  nsides_ok = 0;
  for (s = 0; s < 3; s++)
    sides_ok[s] = FALSE;
  for (s = 0; s < 3; s++)
  {
    if (tr[t][3 + s] == -1)
    /* side is boundary */
    {
      sides_ok[s] = TRUE;
      nsides_ok++;
    }
    else if (tr[tr[t][3 + s]][10] == -1)
    /* side is shared by non-active triangle */
    {
      sides_ok[s] = TRUE;
      nsides_ok++;
    }
    else if (tr[tr[t][3 + s]][9] == 0)
    /* side contains a switched off element */
    {
      sides_ok[s] = TRUE;
      nsides_ok++;
    }
  }
  if (nsides_ok == 3)
    return FALSE;

  /* initial position for the candidate point */
  (*cntr)[0] = 0.0;
  (*cntr)[1] = 0.0;
  (*cntr)[2] = 0.0;

  /* the candidate point position is the average of the projective
     positions originating from each side which is OK */
  for (s = 0; s < 3; s++)
    if (sides_ok[s])
    {
      /* once more: this part is not 3D compliant !!! */
      ds = eg_PointDist2D(coords[tr[t][s]],
                          coords[tr[t][(s + 1) % 3]]);
      v1[0] = coords[tr[t][(s + 1) % 3]][0] -
              coords[tr[t][s]][0];
      v1[1] = coords[tr[t][(s + 1) % 3]][1] -
              coords[tr[t][s]][1];
      v1[2] = 0.0;
      v2[0] = 0.0;
      v2[1] = 0.0;
      v2[2] = 1.0;
      /* assembling the perpendicular direction for projection */
      e_VectCrossVect(v1, v2, v3);
      l = 0.0;
      for (i = 0; i < 3; i++)
        l += v3[i] * v3[i];
      for (i = 0; i < 3; i++)
        v3[i] /= sqrt(l);
      /* adding the resulting projected position to the sum */
      (*cntr)[0] += (coords[tr[t][(s + 1) % 3]][0] +
                     coords[tr[t][s]][0]) / 2.0;
      (*cntr)[0] -= ds * v3[0] * 0.866;
      (*cntr)[1] += (coords[tr[t][(s + 1) % 3]][1] +
                     coords[tr[t][s]][1]) / 2.0;
      (*cntr)[1] -= ds * v3[1] * 0.866;
    }
  if (nsides_ok > 0)    /* taking the average */
    for (i = 0; i < 3; i++)
      (*cntr)[i] /= nsides_ok;
  else
  {
    /* or just taking the centroid if none of the sides were ok */
    centroid(cntr, coords[tr[t][0]],
                   coords[tr[t][1]],
                   coords[tr[t][2]]);
  }

  /* find out to which current triangle it belongs
     starting the walking search at triangle t */
  sres = trianglesearch(*cntr, coords, tr, t, ct, &onside, mtrf);
  if (sres == 0)
  {
    /* the optimal projection fell outside the domain */
    e_GeneralWarning("optimal projection failed, centroid assumed");
    centroid(cntr, coords[tr[t][0]],
                   coords[tr[t][1]],
                   coords[tr[t][2]]);
    *ct = t;
  }
  else if (sres == -1)
    return FALSE;

  return TRUE;
}

/* end of file */
