#include "mgn.h"
#include "mgn_graphics.h"
#include "dtr.h"
#include "timing.h"
#include "mgn_aux.h"
#include "gentriangadj.h"

/* #define USE_TOS */

void parse_initial_mesh(MESH *mesh, double ***coords, int *np)
{
  /* This function parses the initial mesh and takes out the
   * points and their number.
   */

  int i;                        /* counter */

  *np = mesh->NNodes;
  if (!((*coords) = e_alloc_cont_d_matr((*np), 3)))
    e_GeneralError("memory error\n");
  for (i = 0; i < (*np); i++)
  {
    (*coords)[i][0] = mesh->MeshPointCoords[i][0];
    (*coords)[i][1] = mesh->MeshPointCoords[i][1];
    (*coords)[i][2] = mesh->MeshPointCoords[i][2];
  }
}

void parse_boundary_of_mesh(MESH *mesh, double ***coords, int *np)
{
  /* This function parses the mesh and takes out the boundary
   * points and their number.
   */

  int i;                        /* counter */
  int *hull;                    /* convex hull of the mesh */
  MTRF *mtrf;

  mtrf = (MTRF *) e_calloc(1, sizeof(MTRF));
  mtrf->used = FALSE;

  hull = NULL;
  *np =
    f_compute_convex_hull(mesh->MeshPointCoords, mesh->NNodes, &hull, mtrf,
                          1);

  if (!((*coords) = e_alloc_cont_d_matr((*np), 3)))
    e_GeneralError("memory error\n");
  for (i = 0; i < (*np); i++)
  {
    (*coords)[i][0] = mesh->MeshPointCoords[hull[i]][0];
    (*coords)[i][1] = mesh->MeshPointCoords[hull[i]][1];
    (*coords)[i][2] = mesh->MeshPointCoords[hull[i]][2];
  }
}

void parse_full_mesh(MESH *bgmesh, double ***bgcoords, int *bgnp,
                     int ***bgtr, int *bgntr, int relax_adj)
{
  /* This function parses a background mesh so that it can be used for
   * searching with the trianglesearch algorithm.
   *
   * when relax_adj is enabled, neighbouringside doesn't require
   * oposite orientations for shared edges
   */

  int t, i, j;                  /* counters */
  int **adj;                    /* the adjacency matrix */
  int *adjdeg;                  /* the adjacency degree matrix */
  int n, s;                     /* side indicators */

  adj = NULL;
  adjdeg = NULL;

  (*bgntr) = bgmesh->NElemsTriang1;
  (*bgtr) = e_alloc_cont_i_matr((*bgntr), 11);
  (*bgnp) = bgmesh->NNodes;
  (*bgcoords) = e_alloc_cont_d_matr((*bgnp), 3);

#ifndef USE_TOS
  if (!e_SetAdjacency(bgmesh, &adjdeg, &adj))
    e_GeneralError("Problems with parsing the background mesh\n");
#else
  if (!Tos_GenTriang1Adj(bgmesh, E_COMP_ADJ, &adjdeg, &adj))
    e_GeneralError("Roman's function failed !!!");
#endif

  for (i = 0; i < (*bgnp); i++)
  {
    for (j = 0; j < 3; j++)
      (*bgcoords)[i][j] = bgmesh->MeshPointCoords[i][j];

  }
  for (t = 0; t < (*bgntr); t++)
    for (j = 0; j < 3; j++)
    {
      (*bgtr)[t][j] = bgmesh->NodesTriang1[t][j] - 1;
      (*bgtr)[t][j + 3] = -1;
    }
  for (t = 0; t < (*bgntr); t++)
  {
    for (i = 0; i < adjdeg[t]; i++)
    {
      if (neighbouringside((*bgtr), t, adj[t][i] - 1, &n, &s, relax_adj))
      {
        (*bgtr)[t][n] = adj[t][i] - 1;
        (*bgtr)[t][3 + n] = s;
        (*bgtr)[t][9] = 1;      /* set the active flag */
      }
    }
  }
#ifndef USE_TOS
  e_FreeAdjacency(bgmesh, adjdeg, adj);
#else
  Tos_FreeGenTriang1Adj(bgmesh, adjdeg, adj);
#endif
}

void create_outputmesh(double **coords, int np, int **triangles,
                       int ntriangles, MESH *outputmesh)
{
  /* This function outputs the info stored in trangles and coords
   * in the output mesh structure.
   */

  int i, j, t;                  /* counters */
  int nt;                       /* the real number of triangles */

  outputmesh->Title = e_alloc_c_vect(20);
  strcpy(outputmesh->Title, "output mesh");
  outputmesh->KeywordSet[E_KW_NUM_TITLE] = TRUE;
  outputmesh->NNodes = np;
  outputmesh->NMeshPoints = np;
  outputmesh->KeywordSet[E_KW_NUM_NNODES] = TRUE;
  outputmesh->KeywordSet[E_KW_NUM_NMESHPOINTS] = TRUE;
  outputmesh->MeshPointCoords = e_alloc_cont_d_matr(np, 3);
  outputmesh->GNIs = e_alloc_i_vect(np);
  outputmesh->NodalDOF = e_alloc_i_vect(np);
  outputmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] = 1;
  for (i = 0; i < np; i++)
  {
    outputmesh->GNIs[i] = i + 1;
    outputmesh->NodalDOF[i] = 3;
    for (j = 0; j < 3; j++)
      outputmesh->MeshPointCoords[i][j] = coords[i][j];
  }
  outputmesh->TotalNodalDOF = np*3;
  
  nt = 0;
  for (t = 0; t < ntriangles; t++)
    if (triangles[t][10] != -1)
      nt++;
  outputmesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS] = TRUE;
  outputmesh->NodesTriang1 = e_alloc_cont_i_matr(nt, 3);
  outputmesh->GEIs = e_alloc_i_vect(nt);
  outputmesh->GEIsTriang1 = outputmesh->GEIs;
  outputmesh->KeywordSet[E_KW_NUM_NODES_TRIANG1] = TRUE;
  outputmesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1] = TRUE;
  nt = 0;
  for (t = 0; t < ntriangles; t++)
    if (triangles[t][10] != -1)
    {
      for (i = 0; i < 3; i++)
        outputmesh->NodesTriang1[nt][i] = triangles[t][i] + 1;
      outputmesh->GEIs[nt] = nt + 1;
      nt++;
    }
  outputmesh->NElemsTriang1 = nt;
  outputmesh->TotalNElems = nt;

}

void parse_mesh(int *task, int meth, MESH *mesh, double ***coords, int *np,
                int ***tr, int *ntr, int *ttr, int ***tpnl,
                int *nhull, int **convhull, BND **bnd)
{
  int parsed;                   /* for rerouting code */
  char c[E_LINE_MAX];           /* user input answer */
  char d[E_LINE_MAX];           /* user input answer */
  char line[E_LINE_MAX];        /* user input answer */
  int e;                        /* counters */
  MTRF mtrf;

  mtrf.used = FALSE;

  if (*task == TASK_REFINE)
  {
        /*** always CCW check ***/
    parse_full_mesh(mesh, coords, np, tr, ntr, FALSE);
    parsed = TRUE;
    if (ShowGraphics)
    {
      mgnDrawAll(*np, *coords, 0, NULL, 0, NULL, 0, NULL, NULL,
                 "initial nodes");
      mgnDrawAll(*np, *coords, *ntr, *tr, 0, NULL, 0, NULL, NULL,
                 "initial coarse mesh");
    }

    if (!CCW_check(*tr, *ntr, *coords))
      e_GeneralError("not all elements in your mesh are CCW!\n");
    if (meth == METH_DELAUNAY)
    {
      temptime = p_Timer();
      printf("\nAre you sure the coarse mesh is Delaunay? yes/no : ");
      fgets(line, E_LINE_MAX, stdin);
      p_StartTimer();
      timerStart -= temptime;
      sscanf(line, "%s", c);
      if ((!strcmp(c, "n")) || (!strcmp(c, "N")))
      {
        printf("\nThe mesh will be checked. This may take some time\n");
        if (!delaunay_check(*tr, *ntr, *coords, *np))
        {
          temptime = p_Timer();
          printf("The coarse mesh was not Delaunay! The task is therefore\n");
          printf("set to REMESH instead of REFINE! Do you agree? yes/no: ");
          fgets(line, E_LINE_MAX, stdin);
          p_StartTimer();
          timerStart -= temptime;
          sscanf(line, "%s", d);
          if ((!strcmp(d, "y")) || (!strcmp(d, "Y")))
          {
            e_free_cont_i_matr(*ntr, *tr);
            e_free_cont_d_matr(*np, *coords);
            *task = TASK_REMESH;
            parsed = FALSE;
          }
          else
          {
            printf("too bad!\n");
            e_exit(1);
          }
        }
        else
        {
          temptime = p_Timer();
          printf("The coarse mesh is Delaunay! Press enter to continue.\n");
          fgets(line, E_LINE_MAX, stdin);
          p_StartTimer();
          timerStart -= temptime;
          strcpy(c, "y");
          parsed = TRUE;
        }
      }
      if ((!strcmp(c, "y")) || (!strcmp(c, "Y")))
      {
        /**** parse triangles and coords ****/
        if (!parsed)
          parse_full_mesh(mesh, coords, np, tr, ntr, FALSE);
        if (!CCW_check(*tr, *ntr, *coords))
          e_GeneralError("not all elements in your mesh are CCW!\n");
        /**** make the convex hull ****/
        *nhull = f_compute_convex_hull(*coords, *np, convhull, &mtrf, 1);
        /**** create tpnl table ****/
        createtpnl(tpnl, *tr, *ntr, *np);
        /**** retrieve the bnd ****
        *bnd = (BND *) e_calloc(1, sizeof(BND));
        if (!setupbnd(*tr, *ntr, (*convhull)[0], *tpnl, *bnd))
          e_GeneralError("bnd generation failed\n"); */
        setup_2D_bnd(mesh,bnd);
        /**** compare and patch-in non-convex parts ****/
        *ttr = *ntr;
        patchnonconv(*coords, tr, ntr, ttr, *tpnl, *nhull, *convhull, *bnd);
        /**** patch-in holes ****/
        for (e = 0; e < (*bnd)->nie; e++)
          patchholes(*coords, tr, ntr, ttr, *tpnl, (*bnd)->ie[e]);
      }
    }
  }
  else if (*task == TASK_REMESH)
  {
        /*** always CCW check ***/
    parse_full_mesh(mesh, coords, np, tr, ntr, FALSE);
    if (ShowGraphics)
    {
      mgnDrawAll(*np, *coords, 0, NULL, 0, NULL, 0, NULL, NULL,
                 "initial nodes");
      mgnDrawAll(*np, *coords, *ntr, *tr, 0, NULL, 0, NULL, NULL,
                 "initial coarse mesh");
    }
    if (!CCW_check(*tr, *ntr, *coords))
      e_GeneralError("not all elements in your mesh are CCW!\n");
    if (meth == METH_DELAUNAY)
    {
      *nhull = f_compute_convex_hull(*coords, *np, convhull, &mtrf, 1);
      createtpnl(tpnl, *tr, *ntr, *np);
      /* *bnd = (BND *) e_calloc(1, sizeof(BND));
      if (!setupbnd(*tr, *ntr, (*convhull)[0], *tpnl, *bnd))
        e_GeneralError("bnd generation failed\n"); */
      setup_2D_bnd(mesh,bnd);
      e_free(*convhull);
      *nhull = 0;
      freetpnl(*tpnl, *np);
      e_free_cont_i_matr(*ntr, *tr);
      *ntr = 0;
      inittpnl(tpnl, *np);
      f_connect_bnd_nodes(*coords, *np, *bnd, tr, ntr, convhull,
                          nhull, *tpnl, &mtrf);
      /* f_connect_mesh(*coords, *np, tr, ntr, convhull,
         nhull, *tpnl, &mtrf); */
      *ttr = *ntr;
    }
  }
  else if (*task == TASK_EBE)
  {
    parse_full_mesh(mesh, coords, np, tr, ntr, TRUE);
    if (ShowGraphics)
    {
      mgnDrawAll(*np, *coords, 0, NULL, 0, NULL, 0, NULL, NULL,
                 "initial nodes");
      mgnDrawAll(*np, *coords, *ntr, *tr, 0, NULL, 0, NULL, NULL,
                 "initial coarse mesh");
    }
    inittpnl(tpnl, *np);
    *ttr = *ntr;
    *nhull = f_compute_convex_hull(*coords, *np, convhull, &mtrf, 1); /* allocated but not used */
    *bnd = (BND *) e_calloc(1, sizeof(BND)); /* allocated but not used */
  }
}

/* end of file */
