#include "mgn.h"
#include "dtr.h"

void patchnonconv(double **coords, int ***tr, int *ntr, int *ttr, int **tpnl,
                  int nh, int *hull, BND *bnd)
{
  /* this function triangulates all area's which fall out of the domain
   * but inside the convex hull.
   * this is done to ensure correct behaviour of the trianglesearch during
   * a refinement of an existing mesh.
   * the new triangles are patched into the triangle matrix.
   */

  int bst;                      /* start of the overlap between hull and bnd */
  int b;                        /* node on an outer edge */
  int h;                        /* node on convhull */
  int **etr;                    /* new triangle matrix for the patch area */
  int entr;                     /* number of elements in etr */
  int *ehull;                   /* convex hull of patch area */
  int enh;                      /* number of elems in ehull */
  int **etpnl;                  /* tpnl table for the patch area */
  BND *ebnd;                    /* dummy bnd for the patch area */
  double **ecoords;             /* dummy coords matrix */
  int enp;                      /* number of nodes in the patch area */
  int etp;                      /* total space allocated in ecoords */
  int stntr;                    /* start index of the new triangles */
  int nnewtr;                   /* number of new patched-in triangles */
  int *etrtable;                /* translation table if triangle are removed */
  int *enntable;                /* translation table for the nodes in ecoords */
  int first;                    /* first patch found */
  int a, lt, rt;                /* updating of the adjacency info variables */
  int i, j, s, t, n;            /* counters */

  MTRF mtrf;
  mtrf.used = FALSE;

  ecoords = NULL;
  etrtable = NULL;
  enntable = NULL;
  etp = 0;
  etr = NULL;
  ehull = NULL;
  etpnl = NULL;
  ebnd = NULL;
  first = TRUE;

  /* non-convex patches can be found where the boundary and the hull are not equal */

  /* first find the offset between hull and boundary */
  for (bst = 0; bst < bnd->oe.nen; bst++)
    if (hull[0] == bnd->oe.en[bst])
      break;
  if (bst == bnd->oe.nen)
    e_GeneralError("hull and bnd have no common nodes\n");

  /* walk through the hull and compare with bnd */
  for (h = 0; h < nh; h++)
  {
    if (bnd->oe.en[(bst + 1) % bnd->oe.nen] == hull[(h + 1)%nh]) /* bnd and hull match */
      bst++;
    else                        /* discrepancy between hull and bnd found */
    {
      /* find where hull and bnd match again */
      for (n = 0; n < bnd->oe.nen - 2; n++)
        if (hull[(h + 1)&nh] == bnd->oe.en[(bst + 2 + n) % bnd->oe.nen])
          break;
      if (n == bnd->oe.nen - 2)
        e_GeneralError("hull and bnd don't close\n");
      /* and setup a meshing operation for the non-convex patch */
      enp = n + 3;
      if (first)
      {
        /* only allocate for the first patch */
        ecoords = e_alloc_cont_d_matr(enp, 3);
        enntable = e_alloc_i_vect(enp);
        etp = enp;
      }
      else
        while (etp < enp)
        {
          /* increment for all later patches (if necessary) */
          ecoords = e_incr_cont_d_matr(etp, 3, ecoords, etp);
          enntable = e_incr_i_vect(etp, enntable, etp);
          etp += etp;
        }
      /* copy the node coordinates of the patch */
      for (j = 0; j < 3; j++)
      {
        ecoords[0][j] = coords[hull[h]][j];
        ecoords[1][j] = coords[hull[(h + 1)%nh]][j];
        for (i = 0; i < n + 1; i++)
          ecoords[2 + i][j] =
            coords[bnd->oe.en[(bst + 2 + (n - 1) - i) % bnd->oe.nen]][j];
      }
      /* setup a translation table */
      enntable[0] = hull[h];
      enntable[1] = hull[(h + 1)%nh];
      for (i = 0; i < n + 1; i++)
        enntable[2 + i] = bnd->oe.en[(bst + 2 + (n - 1) - i) % bnd->oe.nen];
      /* initialize a tpnl */
      inittpnl(&etpnl, enp);
      /* and triangulate the patch */
      f_connect_mesh(ecoords, enp, &etr, &entr, &ehull, &enh, etpnl,
                     &mtrf);
      /* setup the boundary of the patch */
      if (first)
      {
        ebnd = (BND *) e_calloc(1, sizeof(BND));
        ebnd->nie = 0;
        ebnd->oe.nen = n + 3;
        ebnd->oe.en =
          e_alloc_i_vect(n + 3 +
                         ((((n + 3) % EDGE_INCR) ==
                           0) ? 0 : (EDGE_INCR - ((n + 3) % EDGE_INCR))));
      }
      else
      {
        e_free(ebnd->oe.en);
        ebnd->oe.nen = n + 3;
        ebnd->oe.en =
          e_alloc_i_vect(n + 3 +
                         ((((n + 3) % EDGE_INCR) ==
                           0) ? 0 : (EDGE_INCR - ((n + 3) % EDGE_INCR))));
        ebnd->nie = 0;
      }
      for (i = 0; i < ebnd->oe.nen; i++)
        ebnd->oe.en[i] = i;
      /* setup the area code for the triangles in the non-convex part of the patch */
      setareacode(ebnd, etpnl, etr, entr);
      /* count the reamining number of patch triangles */
      nnewtr = 0;
      for (t = 0; t < entr; t++)
        if (etr[t][10] == +1)
          nnewtr++;
      stntr = (*ntr);
      /* if necessary enlarge the triangle matrix */
      while ((*ttr) - (*ntr) < nnewtr)
      {
        (*tr) = e_incr_cont_i_matr((*ttr), 11, (*tr), (*ttr));
        (*ttr) += (*ttr);
      }
      /* as the patch itself might be non-convex, some patch triangles
         * will not be copied in. the triangle numbers therefore need to be 
         * translated.   */
      /* setup the translation table */
      etrtable = e_alloc_i_vect(entr);
      for (t = 0; t < entr; t++)
        etrtable[t] = stntr + t;
      for (t = 0; t < entr; t++)
        if (etr[t][10] == -1)
        {
          etrtable[t] = -1;
          for (s = t + 1; s < entr; s++)
            if (etrtable[s] != -1)
              etrtable[s]--;
        }
      /* and change the triangle definitions according toi the translation table */
      for (t = 0; t < entr; t++)
        for (s = 0; s < 3; s++)
          if (etr[t][3 + s] != -1)
            etr[t][3 + s] = etrtable[etr[t][3 + s]];
      /* and copy the triangles into the general tr matrix */
      for (t = 0; t < entr; t++)
      {
        if (etr[t][10] == +1)
        {
          for (i = 0; i < 3; i++)
          {
            (*tr)[(*ntr)][i] = enntable[etr[t][i]];
            (*tr)[(*ntr)][3 + i] = etr[t][3 + i];
            (*tr)[(*ntr)][6 + i] = etr[t][6 + i];
          }
          (*tr)[(*ntr)][10] = -1;
          (*ntr)++;
        }
      }
      /* update the tpnl */
      for (t = stntr; t < (*ntr); t++)
        addtr2tpnlist((*tr), t, tpnl);
      /* now correct the adjacency of the triangles on both sides of the edge. */
      for (b = 0; b < enp - 1; b++)
      {
        a =
          leftandrighttr((*tr), tpnl, bnd->oe.en[bst + b],
                         bnd->oe.en[(bst + b + 1) % bnd->oe.nen], &lt, &rt);
        /* watch out: left now means in the domain and right means in the patch area */
        if (a != 3)
          e_GeneralError("patching of the non convex area failed\n");
        /* find out which sides lt and rt share */
        for (s = 0; s < 3; s++)
          if ((*tr)[lt][s] == bnd->oe.en[bst + b])
            break;
        for (t = 0; t < 3; t++)
          if ((*tr)[rt][t] == bnd->oe.en[bst + b])
            break;
        /* and adjust the adjacency accordingly */
        (*tr)[lt][3 + s] = rt;
        (*tr)[rt][3 + (t + 2) % 3] = lt;
        (*tr)[lt][6 + s] = 3 + (t + 2) % 3;
        (*tr)[rt][6 + (t + 2) % 3] = 3 + s;
      }
      if (DebugOutput)
        for (t = 0; t < *ntr; t++)
        {
          printf("\ntriangle %d : ", t);
          for (i = 0; i < 11; i++)
            printf("%d ", (*tr)[t][i]);
        }
      /* free up some memory unfortunately, just hope there are not too many patches */
      e_free(etrtable);
      e_free_cont_i_matr(entr, etr);
      e_free(ehull);
      freetpnl(etpnl, enp);
      first = FALSE;
      bst += enp - 1;
    }
  }
  /* free up all memory */
  e_free_cont_d_matr(etp, ecoords);
  if (!first)
  {
    e_free(ebnd->oe.en);
    e_free(ebnd);
  }
}

void patchholes(double **coords, int ***tr, int *ntr, int *ttr, int **tpnl,
                EDGE ie)
{
  /* this function triangulates all holes in the domain.
   * this is done to ensure correct behaviour of the trianglesearch during
   * a refinement of an existing mesh.
   * the new triangles are patched into the triangle matrix.
   */

  int b;                        /* node on an inner edge */
  int **etr;                    /* new triangle matrix */
  int entr;                     /* number of elements in etr */
  int *ehull;                   /* convex hull of the hole */
  int enh;                      /* number of elems in ehull */
  int **etpnl;                  /* tpnl table for the inner edge */
  BND *ebnd;                    /* dummy bnd for the hole */
  double **ecoords;             /* dummy coords matrix */
  int stntr;                    /* start index of the new triangles */
  int nnewtr;                   /* number of new patched-in triangles */
  int *etrtable;                /* translation table if triangle are removed */
  int a, lt, rt;                /* updating of the adjacency info variables */
  int i, j, s, t;               /* counters */

  MTRF mtrf;
  mtrf.used = FALSE;

  ecoords = NULL;
  etr = NULL;
  ehull = NULL;
  etpnl = NULL;
  ebnd = NULL;
  inittpnl(&etpnl, ie.nen);

  /* copy the coordinates of the points on the boundary of the hole */
  ecoords = e_alloc_cont_d_matr(ie.nen, 3);
  for (i = 0; i < ie.nen; i++)
    for (j = 0; j < 3; j++)
      ecoords[i][j] = coords[ie.en[i]][j];

  /* triangulate the hole */
  f_connect_mesh(ecoords, ie.nen, &etr, &entr, &ehull, &enh, etpnl,
                 &mtrf);
  /* the hole itself can be not-convex so setup a boundary definition */
  ebnd = (BND *) e_calloc(1, sizeof(BND));
  ebnd->nie = 0;
  ebnd->oe.nen = ie.nen;
  for (i = -0; i < ie.nen; i++)
    ebnd->oe.en = e_incr_i_vect(i, ebnd->oe.en, EDGE_INCR);
  for (i = 0; i < ebnd->oe.nen; i++)
    ebnd->oe.en[ebnd->oe.nen - i] = i;
  /* determine the non-convex parts of the hole */
  setareacode(ebnd, etpnl, etr, entr);
  /* count the remaining hole triangles */
  nnewtr = 0;
  for (t = 0; t < entr; t++)
    if (etr[t][10] == +1)
      nnewtr++;
  /* patch the new triangles in */
  stntr = (*ntr);
  /* if necessary enlarge the triangle matrix */
  while ((*ttr) - (*ntr) < nnewtr)
  {
    (*tr) = e_incr_cont_i_matr((*ttr), 11, (*tr), (*ttr));
    (*ttr) += (*ttr);
  }
  /* as the hole itself might be non-convex, some hole triangles
   * will not be patched in. the triangle numbers therefore need to be 
   * translated.   */
  /* build the translation tables */
  etrtable = e_alloc_i_vect(entr);
  for (t = 0; t < entr; t++)
    etrtable[t] = stntr + t;
  for (t = 0; t < entr; t++)
    if (etr[t][10] == -1)
    {
      etrtable[t] = -1;
      for (s = t + 1; s < entr; s++)
        if (etrtable[s] != -1)
          etrtable[s]--;
    }
  for (t = 0; t < entr; t++)
    for (s = 0; s < 3; s++)
      if (etr[t][3 + s] != -1)
        etr[t][3 + s] = etrtable[etr[t][3 + s]];
  /* and copy the hole triangles into the general tr matrix */
  for (t = 0; t < entr; t++)
  {
    if (etr[t][10] == +1)
    {
      for (i = 0; i < 3; i++)
      {
        (*tr)[(*ntr)][i] = ie.en[etr[t][i]];
        (*tr)[(*ntr)][3 + i] = etr[t][3 + i];
        (*tr)[(*ntr)][6 + i] = etr[t][6 + i];
      }
      (*tr)[(*ntr)][10] = -1;
      (*ntr)++;
    }
  }
  /* update the tpnl */
  for (t = stntr; t < (*ntr); t++)
    addtr2tpnlist((*tr), t, tpnl);
  /* now correct the adjacency of the triangles on both sides of the edge */
  for (b = 0; b < ie.nen; b++)
  {
    a = leftandrighttr((*tr), tpnl, ie.en[b], ie.en[b + 1], &lt, &rt);
    /* watch out: left now means in the domain and right means in the hole */
    if (a != 3)
      e_GeneralError("patching of the hole failed\n");
    /* find out which sides lt and rt share */
    for (s = 0; s < 3; s++)
      if ((*tr)[lt][s] == ie.en[b])
        break;
    for (t = 0; t < 3; t++)
      if ((*tr)[rt][t] == ie.en[b])
        break;
    /* and adjust adjacency accordingly */
    (*tr)[lt][3 + s] = rt;
    (*tr)[rt][3 + (t + 2) % 3] = lt;
    (*tr)[lt][6 + s] = 3 + (t + 2) % 3;
    (*tr)[rt][6 + (t + 2) % 3] = 3 + s;
  }
  if (DebugOutput)
    for (t = 0; t < *ntr; t++)
    {
      printf("\ntriangle %d : ", t);
      for (i = 0; i < 11; i++)
        printf("%d ", (*tr)[t][i]);
    }

  /* free up that memory */
  e_free_cont_d_matr(ie.nen, ecoords);
  e_free_cont_i_matr(entr, etr);
  e_free(ehull);
  freetpnl(etpnl, ie.nen);
  free(ebnd->oe.en);
  e_free(ebnd);
}

/* end of file */
