/* include files */
#include "e_lib.h"
#include "eg_lib.h"

#define MGNI_C
#include "mgn.h"
#undef MGNI_C

#include "mgn_graphics.h"
#include "postpro.h"
#include "fusion.h"
#include "split.h"
#include "cntlprm.h"
#include "clean.h"
#include "e_copy.h"

void mgni_init(MGNI ** mgni)
{
  int i;

  (*mgni) = (MGNI *) e_calloc(1, sizeof(MGNI));
  (*mgni)->task = TASK_NONE;
  (*mgni)->meth = METH_DELAUNAY;
  (*mgni)->alfa = 1.0;
  (*mgni)->beta = 1.0;
  (*mgni)->ctrl = CTRL_MPR;
  (*mgni)->opt = CONV_OPT_NONE;
  (*mgni)->dln = DLN_CENTR;
  (*mgni)->smtype = SM_NODAL_AVG;
  (*mgni)->smiter = 3;
  (*mgni)->mesh = NULL;
  (*mgni)->geomset = FALSE;
  (*mgni)->outputmesh = NULL;
  (*mgni)->bgmesh = NULL;
  (*mgni)->pcs = NULL;
  (*mgni)->h = 1.0;
  (*mgni)->gra = FALSE;
  (*mgni)->doa = FALSE;
  (*mgni)->pps = TRUE;
  (*mgni)->outf = OUTF_E_LIB;
  (*mgni)->loadfiles = FALSE;
  for (i = 0; i < MGJ_NFI; i++)
    strcpy((*mgni)->fname[i], "");

}

void mgni_free(MGNI * mgni)
{
  e_FreeMeshData(mgni->mesh);
#ifndef PC_VER
  e_FreeMeshData(mgni->outputmesh);
#endif
  if ((mgni->ctrl == CTRL_BACKGROUND) || (mgni->ctrl == CTRL_SOURCES))
    e_FreeMeshData(mgni->bgmesh);
  if (mgni->ctrl == CTRL_SOURCES)
    FreeCntlMeshParams(mgni->pcs);
}

void mgni_check(MGNI * mgni)
{
  int OK = TRUE;

  /* check for validity of the mgni */
  if (   (mgni->task != TASK_MESHGEN)
      && (mgni->task != TASK_REMESH)
      && (mgni->task != TASK_REFINE)
      && (mgni->task != TASK_EBE)
      && (mgni->task != TASK_POSTPROC) )
  {
    fprintf(stderr, "\nunrecognized task description");
    OK = FALSE;
  }
  if (   (mgni->meth != METH_ADV_FRONT)
      && (mgni->meth != METH_DELAUNAY)
      && (mgni->meth != METH_PAVING)
      && (mgni->meth != METH_REG_GRID)
      && (mgni->meth != METH_NONE) )
  {
    fprintf(stderr, "\nunrecognized method description");
    OK = FALSE;
  }
  if (mgni->meth == METH_DELAUNAY)
  {
    if (mgni->alfa < 0.1)
    {
      fprintf(stderr, "\nalfa should be bigger than 0.1");
      OK = FALSE;
    }
    if (mgni->beta < 0.1)
    {
      fprintf(stderr, "\nbeta should be bigger than 0.1");
      OK = FALSE;
    }
    if ((mgni->dln == DLN_CENTR) && (mgni->alfa < 1.0))
    {
      fprintf(stderr, "\nfor delaunay option CENTROID, alfa must be > 1.0 (e.g. 1.1)");
      OK = FALSE;
    }
    if ((mgni->dln == DLN_CHEW1) && (mgni->ctrl != CTRL_UNIFORM))
    {
		      fprintf(stderr, "\nfor delaunay option CHEW1, the ctrl method must be uniform");
		      OK = FALSE;
		}
    if ((mgni->dln != DLN_CENTR) &&
        (mgni->dln != DLN_OPT) &&
        (mgni->dln != DLN_CENTR_RAD) &&
        (mgni->dln != DLN_CHEW1))
    {
      fprintf(stderr, "\nunrecognized delaunay option");
      OK = FALSE;
    }
  }
  if (mgni->meth == METH_REG_GRID)
  {
    if (mgni->alfa < 0.1)
    {
      fprintf(stderr, "\nalfa should be bigger than 0.1");
      OK = FALSE;
    }
    if (mgni->beta < 0.1)
    {
      fprintf(stderr, "\nbeta should be bigger than 0.1");
      OK = FALSE;
    }
  }
  if ((mgni->ctrl != CTRL_BACKGROUND) &&
      (mgni->ctrl != CTRL_SOURCES) &&
      (mgni->ctrl != CTRL_MPR) &&
      (mgni->ctrl != CTRL_UNIFORM))
  {
    fprintf(stderr, "\nunrecognized control type description");
    OK = FALSE;
  }
  if (   (mgni->opt != CONV_OPT_NONE)
      && (mgni->opt != CONV_OPT_FISSION_TO_QUAD)
      && (mgni->opt != CONV_OPT_RANK_TO_QUAD)
      && (mgni->opt != CONV_OPT_HALF_TO_QUAD)
      && (mgni->opt != CONV_OPT_SPLIT_TO_TRIANG) )
  {
    fprintf(stderr, "\nunrecognized mesh option description");
    OK = FALSE;
  }
  if ((mgni->smtype != SM_LAPLACIAN) &&
      (mgni->smtype != SM_LAPLACIAN_NDE) &&
      (mgni->smtype != SM_NODAL_AVG) &&
      (mgni->smtype != SM_PAVING) && (mgni->smtype != SM_NONE))
  {
    fprintf(stderr, "\nunrecognized smoothing type description");
    OK = FALSE;
  }
  if (mgni->smtype != SM_NONE)
  {
    if (mgni->smiter < 1)
    {
      fprintf(stderr, "\nnumber of smoothing iterations should be > 0");
      OK = FALSE;
    }
  }
  if ((mgni->outf != OUTF_E_LIB) &&
    	(mgni->outf != OUTF_SMSH)    )
  {
    fprintf(stderr, "\nunrecognized output format description");
    OK = FALSE;
  }

  /* check for limits to the current implementation */
  if (mgni->task == TASK_MESHGEN)
  {
    if (mgni->meth != METH_PAVING)
    {
      fprintf(stderr, "\nmesh generation from scratch only for paving");
      OK = FALSE;
    }
  }
  if (mgni->task == TASK_REFINE)
  {
    if (mgni->meth != METH_DELAUNAY)
    {
      fprintf(stderr, "\nrefinement keeping internal nodes works only for Delaunay");
      OK = FALSE;
    }
  }
  if (mgni->meth == METH_PAVING)
  {
    if (mgni->smtype != SM_PAVING)
    {
      fprintf(stderr, "\npaving method requires paving smoothing");
      OK = FALSE;
    }
  }
  if (mgni->smtype == SM_NODAL_AVG)
    if (mgni->meth != METH_DELAUNAY)
    {
      fprintf(stderr, "\nnodal averaging only with Delaunay meshing");
      OK = FALSE;
    }
  if (mgni->smtype == SM_PAVING)
    if (mgni->meth != METH_PAVING)
    {
      fprintf(stderr, "\npaving smoothing only for paving method");
      OK = FALSE;
    }
  if (mgni->meth == METH_PAVING)
    if (mgni->opt!=CONV_OPT_NONE)
    {
      fprintf(stderr, "\npaving cannot have any option");
      OK = FALSE;
    }
  if (mgni->meth == METH_REG_GRID)
  {
    if (mgni->opt!=CONV_OPT_NONE)
    {
      fprintf(stderr, "\nregular grid method cannot have any options");
      OK = FALSE;
    }
    if (mgni->task != TASK_REMESH)
    {
      fprintf(stderr, "\nregular grid method only with remesh task");
      OK = FALSE;
    }
    if (   (mgni->ctrl != CTRL_UNIFORM)
        && (mgni->ctrl != CTRL_MPR))
    {
      fprintf(stderr, "\nregular grid method only with uniform or mpr control");
      OK = FALSE;
    }
  }
  if (   (mgni->task==TASK_POSTPROC)
      && (mgni->meth!=METH_NONE) )
  {
    fprintf(stderr, "\npost-processing can have only NONE as method");
    OK = FALSE;
  }

  if (   (mgni->task==TASK_POSTPROC)
      && (mgni->opt == CONV_OPT_NONE)
      && (mgni->smtype == SM_NONE) )
  {
    fprintf(stderr, "\npost-processing must have an option or smoothing other than NONE");
    OK = FALSE;
  }

  if (mgni->meth==METH_NONE)
    if (mgni->task!=TASK_POSTPROC)
    {
      fprintf(stderr, "\nNONE method can be used only with post-processing");
      OK = FALSE;
    }

  if (   (mgni->opt == CONV_OPT_SPLIT_TO_TRIANG)
      && (mgni->task != TASK_POSTPROC)
      && (mgni->meth != METH_PAVING) )
  {
    fprintf(stderr, "\nsplitting can be used only with post-processing or paving");
    OK = FALSE;
  }

  if (!OK)
    e_GeneralError("mgni_check found errors in your mgni structure!");

}

void mgni_loadfiles(MGNI *mgni)
{
  if (!mgni->loadfiles)
    e_GeneralError("mgni_loadfiles: no filenames specified");

  /* load and check the coarse mesh */
  if (!load_check_crsmesh(mgni, &(mgni->mesh)))
    e_GeneralError("Coarse mesh does not meet criteria\n");

  /* load geometry data if available */
  if (mgni->geomset)
  {
    if (strcmp(mgni->fname[MGJ_FI_GEOM], ""))
    {
      if (!e_GetGeomModel(mgni->mesh, mgni->fname[MGJ_FI_GEOM]))
        e_GeneralError
          ("geometric model contains faults; execution halted!\n");
    }
    else
      e_GeneralError("no geometric model filename specified");
  }

  /* load background mesh */
  if ((mgni->ctrl == CTRL_BACKGROUND) || (mgni->ctrl == CTRL_SOURCES))
  {
    mgni->bgmesh = NULL;
    mgni->bgmesh = (MESH *) e_calloc(1, sizeof(MESH));
    if (!e_GetMeshData(&(mgni->bgmesh), mgni->fname[MGJ_FI_BKGRMESH]))
      e_GeneralError("Background mesh contains faults; execution halted!");
    if (!e_GetMeshParams(mgni->bgmesh, mgni->fname[MGJ_FI_BKGRPARAM]))
      e_GeneralError("Can't read the mesh parameter file; execution halted!");
    if (mgni->ctrl == CTRL_SOURCES)
      GetCntlMeshParams(&(mgni->pcs), mgni->fname[MGJ_FI_SRC]);
  }

  /* load mesh parameters */
  else if (mgni->ctrl == CTRL_MPR)
  {
    if (!e_GetMeshParams(mgni->mesh, mgni->fname[MGJ_FI_CRSPARAM]))
      e_GeneralError("Can't read the mesh parameter file; execution halted!");
    if (!(mgni->mesh)->MeshParam.NodalMeshParamsSet)
      e_GeneralError
        ("For the moment no element mesh parameters are accepted! Sorry!");
  }

  mgni->loadfiles = FALSE;
}

void mgni_main(MGNI * mgni)
{
  /* check the validity of the MGNI structure */
  mgni_check(mgni);

  /* initialise eg_lib */
  eg_Init();

  /* global variables */
  DebugOutput = mgni->doa;
  ShowGraphics = mgni->gra;
  PrintProgress = mgni->pps;

  /* set up graphics (paving has it's own due to data structs) */
  if (ShowGraphics)
  {
    mgnInitGraphics(G2_WINDOW_X, G2_WINDOW_Y,
                    mgni->mesh->NMeshPoints, mgni->mesh->MeshPointCoords);
  }

  /* main functions */
  if (mgni->meth == METH_DELAUNAY)
  {
    /* pass everything to the Delaunay mesh generator */
    dtr_main(mgni);
  }
  else if (mgni->meth == METH_ADV_FRONT)
  {
    /* pass everything to the advancing front mesh generator */
    adv_main(mgni);
  }
  else if (mgni->meth == METH_PAVING)
  {
    /* pass everything to the paving code */
    pav_main(mgni);
  }
  else if (mgni->meth == METH_REG_GRID)
  {
    /* pass everything to the regular grid code */
    rgr_main(mgni);
  }
  else if (mgni->meth == METH_NONE)
  {
    /* just copy the mesh */
    e_MeshCopy(mgni->mesh, &(mgni->outputmesh));
  }

  /* removing unconnected nodes */
  remove_unconnected_nodes(mgni->outputmesh);

  /* conversion of elements if required */
  if (mgni->opt != CONV_OPT_NONE)
  {
    /* first some smoothing of the triangular mesh */
    /* PostProcessMesh(mgni->outputmesh, 3, SM_LAPLACIAN_NDE); */

    /* conversion to a different element */
    if(mgni->opt == CONV_OPT_FISSION_TO_QUAD)
    {
      fuse2quad(mgni->outputmesh, 1);
    }
    else if(mgni->opt == CONV_OPT_RANK_TO_QUAD)
    {
      fuse2quad(mgni->outputmesh, 0);
      split2quad(mgni->outputmesh);
    }
    else if(mgni->opt == CONV_OPT_HALF_TO_QUAD)
    {
      split2quad(mgni->outputmesh);
    }
    else if(mgni->opt == CONV_OPT_SPLIT_TO_TRIANG)
    {
      split2triang(mgni->outputmesh);
    }
    
    
    if (ShowGraphics)
      mgnDrawAll((mgni->outputmesh)->NMeshPoints,
                 (mgni->outputmesh)->MeshPointCoords,
                 0,NULL,
                 (mgni->outputmesh)->NElemsQuad1,
                 (mgni->outputmesh)->NodesQuad1, -1, NULL, NULL,
                 "quad mesh conversion");
  }

  /* perform the necessary smoothing */
  if (mgni->smtype != SM_NONE && mgni->smtype != SM_NODAL_AVG)
    PostProcessMesh(mgni->outputmesh, mgni->smiter, mgni->smtype);

  if (ShowGraphics)
  {
    mgnDrawAll((mgni->outputmesh)->NMeshPoints,
               (mgni->outputmesh)->MeshPointCoords,
               (mgni->outputmesh)->NElemsTriang1,
               (mgni->outputmesh)->NodesTriang1,
               (mgni->outputmesh)->NElemsQuad1,
               (mgni->outputmesh)->NodesQuad1, -1, NULL, NULL,
               "mesh after smoothing");
    mgnCloseGraphics();
  }
}

