#include "pav_main.h"

static void InsertWedge(PAVEDMESH *mesh, LOOP *loop);
static int WedgePositions(PAVEDMESH *mesh, LOOP *loop, 
                          int nexpside, double expangle, NODE *startexp);

static int FormTuck(PAVEDMESH *mesh, LOOP *loop);
static int TuckPositions(PAVEDMESH *mesh, LOOP *loop, 
                         int nredside, double redangle, NODE *startred);

/* ######################################################################### */
/*                       Wedge and tuck insertion                            */
/* ######################################################################### */

/*
 * insert a wedge element at the loop head 
 */
static void InsertWedge(PAVEDMESH *mesh, LOOP *loop)
{
  int     curr, next, prev;
  double  xc, yc, xp, yp, xn, yn, dx, dy, x, y, l1;
  int     wn1, wn3, wn4, quad, rq;
  NODE   *newnode;


  curr = loop->Nodes->Node;
  next = loop->Nodes->Next->Node;
  prev = loop->Nodes->Prev->Node;

  l1   = loop->Nodes->ReqLength;

  xc = mesh->NodeCoords[curr][0];
  yc = mesh->NodeCoords[curr][1];

  xp = mesh->NodeCoords[prev][0];
  yp = mesh->NodeCoords[prev][1];

  xn = mesh->NodeCoords[next][0];
  yn = mesh->NodeCoords[next][1];

  /* move the current node towards next */
  dx = ((2*xc+xn)/3.0) - xc;
  dy = ((2*yc+yn)/3.0) - yc;
  ChangeNodeCoord(curr, dx, dy, mesh, FALSE);
  
  /* insert node towards previous */
  x = (xp+2*xc)/3.0;
  y = (yp+2*yc)/3.0;
  CreateNode(x, y, NODE_FRONT, mesh, FALSE);
  wn1           = mesh->NNodes-1;
  newnode       = GetNewNode(mesh);
  newnode->Node = wn1;
#ifdef GLOBAL_SIZE_CONTROL
  newnode->ReqLength = ProjLenCalc(mesh, newnode->Node);
#else
  newnode->ReqLength = l1;
#endif

  InsertBehindFrontNode(mesh, newnode, loop, FALSE);

  /* find third node of wedge */
  quad = FindQuadOnSide(mesh, curr, next);
  wn3 = FindNodeInQuad(mesh->QuadNodes[quad], curr, 1);

  if(CHK_FRONT(mesh->NodeFlag[wn3]))
  {
    e_GeneralError("InsertWedge: internal error 1!");
  }

  /* insert node above */
  x = (x + xc - mesh->NodeCoords[wn3][0]);
  y = (y + yc - mesh->NodeCoords[wn3][1]);
  CreateNode(x, y, NODE_FRONT, mesh, FALSE);
  wn4           = mesh->NNodes-1;
  newnode       = GetNewNode(mesh);
  newnode->Node = wn4;
#ifdef GLOBAL_SIZE_CONTROL
  newnode->ReqLength = ProjLenCalc(mesh, newnode->Node);
#else
  newnode->ReqLength = l1;
#endif

  InsertBehindFrontNode(mesh, newnode, loop, FALSE);

  /* change quad node numbers */
  while(mesh->NNodeElem[curr] > 1)
  {
    rq = mesh->NodeElem[curr][0];
    if(quad == rq)
    {
      rq = mesh->NodeElem[curr][1];
    }
    ChangeElementNode(rq, curr, wn1, mesh, FALSE);
  }
  
  /* create wedge element */
  CreateElement(wn1, wn3, curr, wn4, mesh, FALSE);
}


/*
 * Determine how many wedges and to where they should be inserted
 */
static int WedgePositions(PAVEDMESH *mesh, LOOP *loop, 
                          int nexpside, double expangle, NODE *startexp)
{
  int i, n, m, div, travdiv, output;

  /* the travelled angle contains how many 90 degrees */
  n = ((int)(expangle / ANGLE_90))+1;
  m = n;
  for(i = 0; i < m; i++)
  {
    div = nexpside / (n+1);
    if(div > 2)
      break;
    n--;
  }

  if(n != 0)
  {
#ifdef PAVING_DEBUG
    e_GeneralWarning("!!!!! Inserting %d wedge(s) against expansion !!!!!", n);
#endif
    SetLoopHead(loop, startexp, FALSE);
    for(i = 0; i < n; i++)
    {
      travdiv = 0;
      do
      {
        travdiv++;
        StepForward(loop, FALSE);
      }
      while(travdiv < div);
 
      InsertWedge(mesh, loop);
    }
    output = P_YES;
  }
  else
  {
#ifdef PAVING_DEBUG
    e_GeneralWarning("WedgePositions: No wedge could be inserted !", n);
#endif
    output = P_NO;
  }

  return(output);
}

/*
 * Form a tuck at loop head
 */
static int FormTuck(PAVEDMESH *mesh, LOOP *loop)
{
  int curr, next, prev;
  int wn2, wn3, wn4, quad;

  curr = loop->Nodes->Node;
  next = loop->Nodes->Next->Node;
  prev = loop->Nodes->Prev->Node;
  
  /* If there are more element connected to the nodes I do not know
     how to do tuck, so just ignore the request this time */
  if(   (mesh->NNodeElem[curr] != 2)
     || (mesh->NNodeElem[next] != 2) )
  {
#ifdef PAVING_DEBUG
    e_GeneralWarning("InsertTuck: tuck formation is not possible !");
#endif
    return(P_NO);
  }

  /* find the nodes which will be used */
  quad = FindQuadOnSide(mesh, curr, prev);
  wn2 = FindNodeInQuad(mesh->QuadNodes[quad], curr, 2);
  wn3 = FindNodeInQuad(mesh->QuadNodes[quad], curr, 3);

  quad = FindQuadOnSide(mesh, curr, next);
  wn4 = FindNodeInQuad(mesh->QuadNodes[quad], curr, 2);

  /* the nodes should not be on a mesh boundary, because in that case bad
     elements will be formed, ignore tuck formation */
  if(   CHK_BND(mesh->NodeFlag[wn2])
     || CHK_BND(mesh->NodeFlag[wn3])
     || CHK_BND(mesh->NodeFlag[wn4]) )
  {
    return(P_NO);
  }

  /* delete two elements */  
  DeleteElement(mesh->NodeElem[curr][1], mesh, FALSE);
  DeleteElement(mesh->NodeElem[curr][0], mesh, FALSE);
  
  if(mesh->NNodeElem[next] != 1)
  {
    e_GeneralError("InsertTuck: internal error 1 !");
  }
  
  /* change node number of the quad */
  ChangeElementNode(mesh->NodeElem[next][0], next, prev, mesh, FALSE);
  /* create a new element */
  CreateElement(prev, wn2, wn3, wn4, mesh, FALSE);
  
  /* delete the nodes from the loop and the mesh */
  StepForward(loop, FALSE);
  StepForward(loop, FALSE);
  RemoveBehindFrontNode(mesh, loop, FALSE);
  RemoveBehindFrontNode(mesh, loop, FALSE);
  DeleteNode(mesh, curr, FALSE);
  DeleteNode(mesh, next, FALSE);
  
  return(P_YES);
}


/*
 * Determine the number of tucks and their positions
 */
static int TuckPositions(PAVEDMESH *mesh, LOOP *loop, 
                         int nredside, double redangle, NODE *startred)
{
  int i, n, m, div, travdiv, output;

  n = ((int)(redangle / ANGLE_90))+1;
  m = n;
  for(i = 0; i < m; i++)
  {
    div = nredside / (n+1);
    if(div > 4)
      break;
    n--;
  }

  if(n != 0)
  {
#ifdef PAVING_DEBUG
    e_GeneralWarning("!!!!! Forming %d tuck(s) against reduction !!!!!", n);
#endif
    output = P_NO;

    SetLoopHead(loop, startred, FALSE);        
    for(i = 0; i < n; i++)
    {
      travdiv = 0;
      do
      {
        travdiv++;
        StepForward(loop, FALSE);
      }
      while(travdiv < div);
    
      /* we might ignore certain tuck requests */
      if(FormTuck(mesh, loop) == P_YES)
      {
        output = P_YES;
      }
    }
  }
  else
  {
#ifdef PAVING_DEBUG
    e_GeneralWarning("TuckPositions: No tuck could be formed !");
#endif
    output = P_NO;
  }

  return(output);
}


/* 
 * Check whether wedge or tuck should be inserted 
 * to keep the required element size
 */
int WedgeTuckInsert(PAVEDMESH *mesh, LOOP *loop)
{
  NODE   *head, *start, *curr, *startexp, *startred;
  int     nexpside, nredside, output;
  double  expangle, redangle;

  /* go to a non-side node if it is possible, 
     in this way every side node series is properly considered */
  head = loop->Nodes;
  do
  {
    if(!CHK_BIT(loop->Nodes->Flag, NODE_SIDE))
      break;
    StepForward(loop, FALSE);
  }
  while(head != loop->Nodes);

  start = loop->Nodes;
  nexpside = 0;
  nredside = 0;
  expangle = 0.0;
  redangle = 0.0;
  startexp = NULL;
  startred = NULL;
  output = P_NO;
  /* go around the loop and check for expansion and reduction sides */
  do
  {
    curr = loop->Nodes;

    /* if it is not mesh boundary and it is expanding */
    if(   ((curr->Flag & ALL_NODE_TYPE) == NODE_SIDE)
       && ( !CHK_BND(mesh->NodeFlag[curr->Node]) )
       && (curr->Angle > EXPAND_ANGLE) )
    {
      if(nexpside == 0)
      {
        startexp = curr;
      }
      nexpside++;
      expangle += curr->Angle - ANGLE_180;
    }
    else
    {
      if((nexpside > 4) && (expangle > ANGLE_90))
      {
        if(WedgePositions(mesh, loop, nexpside, expangle, startexp) == P_YES)
        {
          output = P_YES;
        }
      }
      nexpside = 0;
      expangle = 0.0;
    }

    /* if it is not mesh boundary and it is shrinking */
    if(   ((curr->Flag & ALL_NODE_TYPE) == NODE_SIDE)
       && ( !CHK_BND(mesh->NodeFlag[curr->Node]) )
       && (curr->Angle < REDUCE_ANGLE) )
    {
      if(nredside == 0)
      {
        startred = curr;
      }
      nredside++;
      redangle += ANGLE_180 - curr->Angle;
    }
    else
    {
      if((nredside > 4) && (redangle > ANGLE_90))
      {
        /* tuck formation is not always allowed */
        if(TuckPositions(mesh, loop, nredside, redangle, startred) == P_YES)
        {
          output = P_YES;
        }
      }
      nredside = 0;
      redangle = 0.0;
    }
    StepForward(loop, FALSE);
  }
  while(start != loop->Nodes);

  /* maybe we went around, so check again */
  if((nexpside > 4) && (expangle > ANGLE_90))
  {
    if(WedgePositions(mesh, loop, nexpside, expangle, startexp) == P_YES)
    {
      output = P_YES;
    }
  }

  if((nredside > 4) && (redangle > ANGLE_90))
  {
    /* tuck formation is not always allowed */
    if(TuckPositions(mesh, loop, nredside, redangle, startred) == P_YES)
    {
      output = P_YES;
    }
  }

  /* if no wedge or tuck is inserted restore original position,
     otherwise it does not matter */
  if(output == P_NO)
  {
    SetLoopHead(loop, head, FALSE);
  }

  return(output);
}



