#include "pav_main.h"
#include "pav_operation.h"
#include "inside.h"
#include "eg_lib.h"
#include "common.h"
#include "bgdata.h"
#include "mtrf.h"

#ifdef PAVING_GRAPHICS
#include "pav_graphics.h"
#endif


static double SegmentPerpendicular(double **Coords, int n1, int n2, int n3, int n4);
static double SegmentDist(double **Coords, int n1, int n2, int n3, int n4);
static int FactoryMonitor(PAVEDMESH *mesh, LOOP *loop, int nsegment,
                          int backmem, int formem);


/* ############################################################################ */
/*                              Basic functions                                 */
/* ############################################################################ */

/*
 * Find a quad element on the side
 */
int FindQuadOnSide(PAVEDMESH *mesh, int n1, int n2)
{
  int found, i, j, quad;

  found = FALSE;
  quad  = -1;
  for(i = 0; i < mesh->NNodeElem[n1] && !found; i++)
  {
    for(j = 0; j < mesh->NNodeElem[n2]; j++)
    {
      if(mesh->NodeElem[n1][i] == mesh->NodeElem[n2][j])
      {
        quad  = mesh->NodeElem[n2][j];
        found = TRUE;
        break;
      }
    }
  }
  
  if(found == FALSE)
    e_GeneralError("FindQuad: internal error !");
  
  return(quad);
}


/* 
 * Given a node in a quad, find it, and return the node which
 * is 'incr' positions away from it.
 */
int FindNodeInQuad(int *Nodes, int n, int incr)
{
  int i;

  for(i = 0; i < 4; i++)
  {
    if(Nodes[i] == n)
      break;
  }
  if(i == 4)
    e_GeneralError("FindNodeInQuad: Node is not in quad !");
  
  return(Nodes[(i+incr) % 4]);
}


/* 
 * The function calculates the internal angle of two vectors.
 * The vectors are n1->n2 and n2->n3 . Assuming a counterclockwise
 * direction, the internal angle of the two vectors are calculated.
 */
double AngleCalc(double xn1, double yn1, double xn2, double yn2,
                 double xn3, double yn3)
{
  double  x1, x2, y1, y2, sign, angle;
  double  preang;

  x1 = xn2 - xn1;
  y1 = yn2 - yn1;

  x2 = xn3 - xn2;
  y2 = yn3 - yn2;

  /* determine what kind of joint we have at the current node,
     it can be convex or non-convex */
  sign = x1 * y2 - y1 * x2;

  /* determine the angle between the two vector, 
     it is always between 0 and 180 degree */
  preang = (x1 * x2 + y1 * y2) / (sqrt(x1*x1+y1*y1)*sqrt(x2*x2+y2*y2));
  /* limit the values for cos,
     this is necessary because of precision problem */
  if(preang < -1.0)
    preang = -1.0;
  else if(preang > 1.0)
    preang = 1.0;

  angle = acos(preang);

  /* convex angle */
  if((sign > 0) || (E_ABS(sign) < ZERO))
  {
    angle = M_PI - angle;
  }
  else if(sign < 0) /* non-convex angle */
  {
    angle = M_PI + angle;
  }

  /* check for almost zero angle */
  if(   (E_ABS(angle - M_PI*2.0) < ZERO)
     || (E_ABS(angle) < ZERO) )
    angle = 0.0;
  
  return(angle);
}

/* 
 * The function calculates a division of an internal angle.
 * The division angle is counted from n1->n2 vector.
 * The division angle is corrected by the inclination of the
 * angle of the n1->n2 vector.
 */
double AngleDivider(double xn1, double yn1, double xn2, double yn2,
                    double div, double angle)
{
  double x1, y1, d1, gamma, preang;

  x1 = xn2 - xn1;
  y1 = yn2 - yn1;

  d1 = sqrt(x1*x1+y1*y1);

  /* angle of vector pre->cur */
  preang = x1 / d1;
  if(preang < -1.0)
    preang = -1.0;
  else if(preang > 1.0)
    preang = 1.0;

  gamma = acos(preang);

  if(y1 < 0)
    gamma *= -1.0;

  return(gamma + M_PI + angle * (div - 1.0));
}


#ifdef GLOBAL_SIZE_CONTROL
double ProjLenCalc(PAVEDMESH *mesh, int ncurr)
{

  int i;
  double x, y, xc, yc, dx, dy, d, length, weight;
  MTRF  mtrf;
  
  /* mesh parameter control */
  if(mesh->ctrl == CTRL_MPR)
  {
    if(CHK_BND(mesh->NodeFlag[ncurr]))
    {
      return(mesh->BndLenParam[ncurr]);
    }
  
    length = 0.0;
    weight = 0.0;
    xc = mesh->NodeCoords[ncurr][0];
    yc = mesh->NodeCoords[ncurr][1];
    for(i = 0; i < mesh->NBnd; i++)
    {
      x = mesh->NodeCoords[i][0];
      y = mesh->NodeCoords[i][1];
    
      dx = x-xc;
      dy = y-yc;
    
      d = 1.0 / (dx*dx + dy*dy);
      length += d * mesh->BndLenParam[i];
      weight += d;
    }
    length /= weight;
  }
  /* background control with possible sources */
  else if(mesh->ctrl == CTRL_BACKGROUND || mesh->ctrl == CTRL_SOURCES)
  {
    mtrf.used = FALSE;
    length = get_bg_mpr(mesh->bgd, mesh->NodeCoords[ncurr], mesh->ctrl, &mtrf);
  }
  else if(mesh->ctrl == CTRL_UNIFORM)
  {
    length = mesh->uniform_h;
  }
  else
  {
    e_GeneralError("ProjLenCalc: invalid control type !");
  }

  if(E_ABS(length) < ZERO)
  {
    e_GeneralError("ProjLenCalc: zero projected length at node %d!", ncurr+1);
  }

  return(length);
}

#else /* local element size control, we trust the all mighty ... */
/* 
 * Calculates the projected distance from the two adjacent side length
 * of the new node. It takes into consideration the angle and the 
 * node type at the "middle" node.
 */
double ProjLenCalc(PAVEDMESH *mesh, int nprev, int ncurr, int nnext,
                   double angle, int type)
{

  /* at the length determination, what value to use for "smoothing" the bumps out */
  double bumps[NO_NODE_CLASSES] = {1.0, 2.0, 3.0, 4.0};
  double x1, x2, y1, y2, d1, d2, length;

  /* length of the vector pointing to the new node from cur */
  x1 = mesh->NodeCoords[ncurr][0] - mesh->NodeCoords[nprev][0];
  y1 = mesh->NodeCoords[ncurr][1] - mesh->NodeCoords[nprev][1];

  x2 = mesh->NodeCoords[nnext][0] - mesh->NodeCoords[ncurr][0];
  y2 = mesh->NodeCoords[nnext][1] - mesh->NodeCoords[ncurr][1];

  d1 = sqrt(x1*x1+y1*y1);
  d2 = sqrt(x2*x2+y2*y2);

  length = ((d1 + d2) * 0.5) / (sin(angle / bumps[type]));

  if(E_ABS(length) < ZERO)
  {
    e_GeneralError("ProjLenCalc: zero projected length at node %d!", ncurr+1);
  }

  return(length);
}

#endif



int GetNodeType(int nodeclass)
{
  if(CHK_BIT(nodeclass, NODE_END))
  {
    return 0;
  }
  else if(CHK_BIT(nodeclass, NODE_SIDE))
  {
    return 1;
  }
  else if(CHK_BIT(nodeclass, NODE_CORNER))
  {
    return 2;
  }
  else if(CHK_BIT(nodeclass, NODE_REVERSAL))
  {
    return 3;
  }
  else /* unknown type */
  {
    e_GeneralError("GetNodeType: unknown type !");
    return -1;
  }
}


/* ############################################################################ */
/*                                    Creation                                  */
/* ############################################################################ */

/* 
 * Abstract: Generates nodes and one element (if requested) at the current node 
 *           at the current stage.
 *           Stages are used at nodes where more than one element should be generated.
 *           Returns the number of generated segments, counting from the current node.
 *           The returned value can be zero !
 */
int CreateEntities(PAVEDMESH *mesh, LOOP *loop, int genelem)
{
  /* the divisions of the internal angle */
  double divider[NO_NODE_CLASSES][MAX_GEN_NODE] = { 
    {0.0, 0.0, 0.0, 0.0, 0.0},
    {0.5, 0.0, 0.0, 0.0, 0.0},
    {0.666666666666666, 0.5, 0.333333333333333, 0.0, 0.0},
    {0.75, 0.625, 0.5, 0.375, 0.25}
  };
  /* at what stage to generate the node, -1 nothing to generate */
  int nodestage[NO_NODE_CLASSES][MAX_GEN_NODE] = {
    {-1, -1, -1, -1, -1},
    { 0, -1, -1, -1, -1},
    { 0,  1,  1, -1, -1},
    { 0,  1,  1,  2,  2}
  };
  NODE   *curr, *newnode;
  int     start, type;
  int     currstage;
  double  angle, projangle, length;
  int     ncurr, nprev, nnext;
  double  xcurr, ycurr, xprev, yprev, xnew, ynew;
  int     n1, n2, n3, n4;
  int     i, ngen = 1;


  curr  = loop->Nodes;
  type = GetNodeType(curr->Flag);
  currstage = curr->Stage;

  /* special treatment for end nodes, 
     no node is generated,
     element is generated from already existing nodes */
  if(type == 0)
  {
    /* nothing should happen at this end node */
    if(currstage != 0)
      return 0;
    
    n1 = curr->Prev->Prev->Node;
    n2 = curr->Prev->Node;
    n3 = curr->Node;
    n4 = curr->Next->Node;
    
    /* delete previous */
    RemoveBehindFrontNode(mesh, loop, TRUE);
    /* set head to the next */
    StepForward(loop, TRUE);
    /* delete previous again */
    RemoveBehindFrontNode(mesh, loop, TRUE);
  }
  else
  {
    /* store the previous element, because it may be deleted*/
    start = curr->Prev->Node;
    angle = curr->Angle;
    ncurr = curr->Node;
    
    nprev = curr->RefPrev;
    nnext = curr->RefNext;
    
    /* where to put the new node */
#ifdef GLOBAL_SIZE_CONTROL
    length = ProjLenCalc(mesh, ncurr);
#else
    length = ProjLenCalc(mesh, nprev, ncurr, nnext, angle, type);
#endif

    xcurr = mesh->NodeCoords[ncurr][0];
    ycurr = mesh->NodeCoords[ncurr][1];
    xprev = mesh->NodeCoords[nprev][0];
    yprev = mesh->NodeCoords[nprev][1];
    for(i = 0; i < MAX_GEN_NODE; i++)
    {
      if(nodestage[type][i] == -1)
        break;
    
      /* node should be generated at this stage */
      if(nodestage[type][i] == currstage)
      {
        /* count generated segments */
        ngen++;
        
        /* divide the angle at the current node */
        projangle = AngleDivider(xprev, yprev, xcurr, ycurr, divider[type][i], angle);
 
        /* only every second node is on a diagonal vector, then use sqrt(2.0) length */
        xnew = xcurr + cos(projangle) * length * ((i % 2) ? sqrt(2.0) : 1.0);
        ynew = ycurr + sin(projangle) * length * ((i % 2) ? sqrt(2.0) : 1.0);
 
        CreateNode(xnew, ynew, NODE_FRONT, mesh, TRUE);
        
        /* first stage, remove previous node */
        if((i == 0) && (genelem == TRUE))
        {
          RemoveBehindFrontNode(mesh, loop, TRUE);
        }
        
        /* create a new node to insert into the loop */
        newnode            = GetNewNode(mesh);
        newnode->Node      = mesh->NNodes-1;
        /* mark the node as new ????? */
        ADD_BIT(newnode->Flag, NODE_NEW);
        newnode->ReqLength = length;
        newnode->ProjNode  = curr->Node;
        
        /* insert node into the loop */
        InsertBehindFrontNode(mesh, newnode, loop, TRUE);
      }
    }

    /* increase the stage level */
    StageIncr(loop->Nodes, TRUE);

    n1 = start;
    n2 = loop->Nodes->Node;
    n3 = loop->Nodes->Prev->Node;
    n4 = loop->Nodes->Prev->Prev->Node;
  }

  /* create one element */
  if(genelem == TRUE)
  {
    CreateElement(n1, n2, n3, n4, mesh, TRUE);
  }

  /* return the number of generated segments */
  return(ngen);
}

/* 
 * Generate the requested nodes with coordinates and then the quad elements.
 * Returns: P_FACE_ACCEPT    -> continue the generation of the row of elements
 *          P_FACE_INTERSECT -> there was an intersection resolution, stop
 *          P_FACE_UNDO      -> changes are undo, due to error, stop
 *
 * !!!!! Reset loop->Primitive to 0, when face is undone or resolved !!!!!
 */
int FaceFactory(PAVEDMESH *mesh, LOOP *loop)
{
  int nsegment, output, stagemax, stagecur;
  int created;
  int stages[NO_NODE_CLASSES] = {1, 1, 2, 3};
  static int   backwardmemory = 0;
  static LOOP *loopmem = NULL;
  int          forwardmemory;
  NODE        *ln1, *ln2, *ln3;

  /* current stage level */
  stagecur = loop->Nodes->Stage;
  /* max no of stages to carry out */
  stagemax = stages[GetNodeType(loop->Nodes->Flag)];
  
  /* check whether we are processing the same loop, 
     otherwise the backward memory is not valid */
  if(loop != loopmem)
  {
    backwardmemory = 0;
    loopmem = loop;
  }

  created = FALSE;
  while((stagecur < stagemax) && (loop->NNodes > 6))
  {
    UndoMarker();

    /* if we are on an a mesh boundary, remember that we are coming from an end node */
    ln1 = loop->Nodes;
    ln2 = ln1->Prev;
    if(   CHK_BND(mesh->NodeFlag[ln2->Node]) 
       && CHK_BIT(ln2->Flag, NODE_END) 
       && ln2->Angle < (M_PI / 3.0) )
    {
      backwardmemory = 1;
    }
    if(backwardmemory)
    {
      backwardmemory++;
    }
    if(backwardmemory > 3)
    {
      backwardmemory = 0;
    }
    
    /* if we are on a mesh boundary, look ahead for an end node */
    forwardmemory = FALSE;
    ln2 = ln1->Next;
    ln3 = ln2->Next;
    if(   (   CHK_BND(mesh->NodeFlag[ln1->Node]) 
           && CHK_BIT(ln1->Flag, NODE_END) 
           && ln1->Angle < (M_PI / 3.0) )
       || (   CHK_BND(mesh->NodeFlag[ln2->Node]) 
           && CHK_BIT(ln2->Flag, NODE_END) 
           && ln2->Angle < (M_PI / 3.0) )
       || (   CHK_BND(mesh->NodeFlag[ln3->Node]) 
           && CHK_BIT(ln3->Flag, NODE_END) 
           && ln3->Angle < (M_PI / 3.0) ) )
    {
      forwardmemory = TRUE;
    }

    /* generate one element and required amount of nodes */
    nsegment = CreateEntities(mesh, loop, TRUE);

    /* check for quality, intersection, etc... */
    output = FactoryMonitor(mesh, loop, nsegment, 
             backwardmemory, forwardmemory);

    /* if everything is well and good */
    if(output & P_FACE_ACCEPT)
    {
      created = TRUE;
      /* intersections have been resolved, so adjust row */
      if(output & P_FACE_RESOLV)
      {
        /* be sure that the loop is re-evaluated */
        loop->Primitive = 0;
        return(P_FACE_RESOLV);
      }
    }
    else if(output & P_FACE_UNDO) /* undo last step */
    {
      if(backwardmemory || forwardmemory)
      {
#ifdef PAVING_DEBUG
        e_GeneralWarning("!!! End is too close !!!");
#endif
        output = P_FACE_ACCEPT;
      }
      else
      {
        UndoUntilMarker(mesh);
        /* be sure that the loop is re-evaluated */
        loop->Primitive = 0;
        if(created == TRUE)
        {
          return(P_FACE_ACCEPT | P_FACE_RESOLV);
        }
        else
        {
          return(P_FACE_UNDO);
        }
      }
    }
    stagecur++;
  }
  
  if(loop->NNodes > 6)
  {
    return(P_FACE_ACCEPT);
  }
  else
  {
    return(P_FACE_ACCEPT | P_FACE_RESOLV);
  }
}


/*
 * Initialise the row putting down the first element, 
 * this allows to define the other end point where the row can end
 */
int InitElement(PAVEDMESH *mesh, LOOP *loop)
{
  int    output, end;
  NODE  *newnode;

  /* if something goes wrong */
  UndoMarker();

  /* for non-end nodes generate the extra node and insert it */
  if(!CHK_BIT(loop->Nodes->Flag, NODE_END))
  {
    /* duplicate starting point */
    newnode       = GetNewNode(mesh);
    newnode->Node = loop->Nodes->Node;

    /* create the first node projection, but no element ! */
    CreateEntities(mesh, loop, FALSE);
    
    /* move the head back, so */
    StepBackward(loop, TRUE);
    /* we insert the duplicated node here */
    InsertBehindFrontNode(mesh, newnode, loop, TRUE);
    /* restore loop head */
    StepForward(loop, TRUE);
    if(CHK_BIT(loop->Nodes->Flag, NODE_SIDE))
    {
      StepForward(loop, TRUE);
    }

    /* reclassify the duplicated node, since it should be an end node */
    ClassifyANode(mesh->NodeCoords, newnode->Prev->Node, newnode->Node, 
                  newnode->Next->Node, &(newnode->Flag), &(newnode->Angle));

    RemoveTypeAmbiguouity(newnode, mesh->NNodeElem[newnode->Node]);

    /* since new node is inserted, we need projected length for smoothing */
#ifdef GLOBAL_SIZE_CONTROL
    newnode->ReqLength = ProjLenCalc(mesh, newnode->Node);
#else
    newnode->ReqLength = 
      ProjLenCalc(mesh, newnode->Prev->Node, newnode->Node, newnode->Next->Node,
                  newnode->Angle, GetNodeType(newnode->Flag) );
#endif

  }
  /* at the end node nothing to generate, but after it */
  else
  {
    StepForward(loop, TRUE);
  }

  /* generate the initial element(s) */
  end = (CHK_BIT(loop->Nodes->Flag, NODE_END)) ? 1 : 0;
  output = FaceFactory(mesh, loop);

  /* even the initial element could not be created due to some error */
  if(output & P_FACE_UNDO)
  {
    UndoUntilMarker(mesh);
    StepForward(loop, FALSE);
    return(output);
  }
  if(output & P_FACE_RESOLV)
  {
    return(output);
  }
  if(end)
  {
    return(P_FACE_RESOLV | P_FACE_ACCEPT);
  }

  /* delete undo information, everything is accepted */
  EmptyUndoQueue();

  StepForward(loop, FALSE);
  
  return(output);
}

/* ############################################################################ */
/*                Quality Checking, intersection resolution                     */
/* ############################################################################ */

/*
 * Calculates the scalar product of the two unit vectors. 
 * Returns: if they are perpendicular -> 1.0
 *          if they are parallel      -> 0.0
 */
static double SegmentPerpendicular(double **coords, int n1, int n2, int n3, int n4)
{
  double x1, y1, x2, y2, x3, y3, x4, y4, s;
  
  x1 = coords[n1][0];
  y1 = coords[n1][1];
  x2 = coords[n2][0];
  y2 = coords[n2][1];
  x3 = coords[n3][0];
  y3 = coords[n3][1];
  x4 = coords[n4][0];
  y4 = coords[n4][1];

  s = fabs(ScalarProd(x1, y1, x2, y2, x3, y3, x4, y4));
  
  /* scalar product */
  return(1.0 - s);
}


static double SegmentDist(double **coords, int n1, int n2, int n3, int n4)
{
  double x1, y1, x2, y2, x3, y3, x4, y4;
  double dx, dy, d, l1, l2, min, S;

  x1 = coords[n1][0];
  y1 = coords[n1][1];
  x2 = coords[n2][0];
  y2 = coords[n2][1];
  x3 = coords[n3][0];
  y3 = coords[n3][1];
  x4 = coords[n4][0];
  y4 = coords[n4][1];

  /* shortest distance */
  dx = (x1 - x3);
  dy = (y1 - y3);
  d = sqrt((dx*dx) + (dy*dy));
  min = d;
  
  dx = (x1 - x4);
  dy = (y1 - y4);
  d = sqrt((dx*dx) + (dy*dy));
  min = (min > d) ? d : min;

  dx = (x2 - x3);
  dy = (y2 - y3);
  d = sqrt((dx*dx) + (dy*dy));
  min = (min > d) ? d : min;

  dx = (x2 - x4);
  dy = (y2 - y4);
  d = sqrt((dx*dx) + (dy*dy));
  min = (min > d) ? d : min;

  /* sticky space */
  dx = (x1 - x2);
  dy = (y1 - y2);
  l1 = sqrt((dx*dx) + (dy*dy));

  dx = (x3 - x4);
  dy = (y3 - y4);
  l2 = sqrt((dx*dx) + (dy*dy));

  S = 0.45 * (l1 + l2) / 2.0;
  
  return(min / S);
}

/*
 * Determines, whether the two segment intersect.
 * Returns: P_YES -> in case of itersection 
 *          P_NO  -> in all other cases
 */
static int SegmentIntersect(double **coords, int n1, int n2, int n3, int n4)
{
  int res;
  double mu1, mu2;
  
  res = eg_LineLineIntersect2D(coords[n1], coords[n2], 
                               coords[n3], coords[n4], 
                               &mu1, &mu2);

  if(res == EG_INVALID)
  {
    e_GeneralError("SegmentIntersect: geometry error !");
  }

  if(   (res == EG_INTERSECT)
     && (mu1 >= 0.0 && mu1 <= 1.0)
     && (mu2 >= 0.0 && mu2 <= 1.0) )
  {
    return(P_YES);
  }

  return(P_NO);
}


/*
 * Determines, whether the segment is inside the quad element.
 * Returns: P_YES -> if segment is inside
 *          P_NO  -> in all other cases
 */
static int SegmentInsideQuad(double **coords, int *nodes, int n1, int n2)
{
  double x1, y1, x2, y2;
  int    i1, i2;

  x1 = coords[n1][0];
  y1 = coords[n1][1];
  x2 = coords[n2][0];
  y2 = coords[n2][1];

  i1 = Check2DNodeInside(coords, 4, nodes, n1+1, x1, y1);
  i2 = Check2DNodeInside(coords, 4, nodes, n2+1, x2, y2);
  
  if(i1 == 1 && i2 == 1)
    return(P_YES);
  else
    return(P_NO);
}



static int CheckSegmentWithLoop(double **coords, LOOP *loop, 
                                int n1, int n2, int *quadnodes,
                                int *ncand, NODE **cand, LOOP **cloop,
                                int *mustresolv)
{
  int i, n3, n4;
  NODE *curr, *start;
  double fdist, fangle, proximity;

  curr = start = loop->Nodes;
  do
  {
    n3 = curr->Node;
    n4 = curr->Next->Node;
    /* touching segment are not good, they are handled by seaming */
    if(!(   (n1 == n3) || (n2 == n4)
         || (n1 == n4) || (n2 == n3) ))
    {

      if(SegmentInsideQuad(coords, quadnodes, n3, n4) == P_YES)
      {
#ifdef PAVING_DEBUG
        e_GeneralWarning("Segment %d - %d is inside quad !", n3, n4);
#endif
        return(P_ERROR);
      }

      fdist  = SegmentDist(coords, n1, n2, n3, n4);
      fangle = SegmentPerpendicular(coords, n1, n2, n3, n4);
      proximity = 1.0 * fdist + 0.25 * fangle;

      if(SegmentIntersect(coords, n1, n2, n3, n4) == P_YES)
      {
        *mustresolv = TRUE;
        proximity = 0.0;
      }

      if(proximity < 1.0)
      {
        for(i = 0; i < *ncand; i++)
        {
          if(cand[i] == curr)
            break;
        }
        if(i == *ncand)
        {
          if(*ncand >= P_MAX_CAND)
          {
#ifdef PAVING_DEBUG
            e_GeneralWarning("Too many candidates for one segment !");
#endif
            return(P_ERROR);
          }
          cand[*ncand]  = curr;
          cloop[*ncand] = loop;
          (*ncand)++;
        }
#ifdef PAVING_DEBUG        
        e_GeneralWarning("e1=%3d %3d | e2-%3d %3d ----- prox: %f mustresolv: %d", 
                         n1+1, n2+1, n3+1, n4+1, proximity, *mustresolv);
#endif
      }
    }
    
    curr = curr->Next;
  }
  while(curr != start);
  
  return(P_SUCCESS);
}


static int CheckAllSegments(PAVEDMESH *mesh, LOOP *loop, int nsegment,
                            int ncand[P_MAX_SEGMENT],
                            NODE *cand[P_MAX_SEGMENT][P_MAX_CAND],
                            LOOP *cloop[P_MAX_SEGMENT][P_MAX_CAND], 
                            int *mustresolv)
{
  int    nproblem;
  int    i, n1, n2, output, quad;
  LOOP  *cl;
  NODE  *node;
  int    qn[4];

  /* last created quad element */
  quad = mesh->NQuads-1;

  for(i = 0; i < 4; i++)
  {
    qn[i] = mesh->QuadNodes[quad][i]+1;
  }

  /* 
   * all new segment should be checked for 
   *   - intersection, 
   *   - proximity, 
   * etc ...
   */
  nproblem = 0;
  node = loop->Nodes;
  for(i = 0; i < nsegment; i++)
  {
    ncand[i] = 0;

    n1 = node->Node;
    n2 = node->Prev->Node;

    /* check nodes are floating */
    if(!(   CHK_BND(mesh->NodeFlag[n1])
         || CHK_BND(mesh->NodeFlag[n2]) ))
    {

      /* go through all fronts */
      cl = mesh->Fronts;
      do
      {
      
        output = CheckSegmentWithLoop(mesh->NodeCoords, cl, n1, n2, qn,
                                      &(ncand[i]), cand[i], cloop[i], 
                                      mustresolv);
      
        /* if something is wrong, quit */
        if(output == P_ERROR)
        {
          return(P_FACE_UNDO);
        }
 
        cl = cl->Next;
      }
      while(cl != mesh->Fronts);
 
      nproblem += ncand[i];
    }
    
    node = node->Prev;
  }

  return(nproblem);
}


/* 
 * Checks the internal angles and the shape of the quad element.
 * Returns;   0.0 -> wrong quad
 *          < 1.0 -> depending on the quality;
 */
double QuadShapeQuality(double **Coords, int n1, int n2, int n3, int n4)
{
  double x1, y1, x2, y2, x3, y3, x4, y4;
  double a1, a2, a3, a4, s;

  x1 = Coords[n1][0];
  y1 = Coords[n1][1];
  x2 = Coords[n2][0];
  y2 = Coords[n2][1];
  x3 = Coords[n3][0];
  y3 = Coords[n3][1];
  x4 = Coords[n4][0];
  y4 = Coords[n4][1];

  a1 = AngleCalc(x1, y1, x2, y2, x3, y3);
  a2 = AngleCalc(x2, y2, x3, y3, x4 ,y4);
  a3 = AngleCalc(x3, y3, x4, y4, x1, y1);
  a4 = AngleCalc(x4, y4, x1, y1, x2, y2);

  /* arrowhead shape or small interior angle */
  if(   (a1 > ANGLE_180) || (a1 < ANGLE_10)
     || (a2 > ANGLE_180) || (a2 < ANGLE_10)
     || (a3 > ANGLE_180) || (a3 < ANGLE_10)
     || (a4 > ANGLE_180) || (a4 < ANGLE_10) )
  {
    return(0.0);
  }

  /* sides are intersecting ? */
  if(SegmentIntersect(Coords, n1, n2, n3, n4) == P_YES)
  {
    return(0.0);
  }
  
  if(SegmentIntersect(Coords, n1, n4, n2, n3) == P_YES)
  {
    return(0.0);
  }
  
  s = 0.0;
  s += fabs(ScalarProd(x1, y1, x2, y2, x2, y2, x3, y3));
  s += fabs(ScalarProd(x2, y2, x3, y3, x3, y3, x4, y4));
  s += fabs(ScalarProd(x3, y3, x4, y4, x4, y4, x1, y1));
  s += fabs(ScalarProd(x4, y4, x1, y1, x1, y1, x2, y2));

  return(1.0 - (s / 4.0));
}


/* check the ratio of the adjacent sides */
int QuadSideLengthQuality(double **Coords, int n1, int n2, int n3, int n4)
{
  double x1, y1, x2, y2, x3, y3, x4, y4;
  double x, y, l1, l2, l3, l4;

  x1 = Coords[n1][0];
  y1 = Coords[n1][1];
  x2 = Coords[n2][0];
  y2 = Coords[n2][1];
  x3 = Coords[n3][0];
  y3 = Coords[n3][1];
  x4 = Coords[n4][0];
  y4 = Coords[n4][1];

  x = x1 - x2;
  y = y1 - y2;
  l1 = sqrt((x * x) + (y * y));
  x = x2 - x3;
  y = y2 - y3;
  l2 = sqrt((x * x) + (y * y));
  x = x3 - x4;
  y = y3 - y4;
  l3 = sqrt((x * x) + (y * y));
  x = x4 - x1;
  y = y4 - y1;
  l4 = sqrt((x * x) + (y * y));

  if(   ((l1 / l2) > MAX_SIDE_RATIO) 
     || ((l2 / l1) > MAX_SIDE_RATIO)
     || ((l2 / l3) > MAX_SIDE_RATIO)
     || ((l3 / l2) > MAX_SIDE_RATIO)
     || ((l3 / l4) > MAX_SIDE_RATIO)
     || ((l4 / l3) > MAX_SIDE_RATIO)
     || ((l4 / l1) > MAX_SIDE_RATIO)
     || ((l1 / l4) > MAX_SIDE_RATIO) )
  {
    return(P_BAD);
  }
  else
  {
    return(P_GOOD);
  }

}

static int CheckEvenNoNode(NODE *start, NODE *end)
{
  NODE *curr;
  int   n = 1;
  
  curr = start;
  do
  {
    curr = curr->Next;
    n++;
  }
  while(curr != start && curr != end);

  /* we walked around, 'end' node is on another loop */
  if(curr == start)
  {
    return(P_YES);
  }
  else
  {
    /* is it even number */
    if(n % 2)
    {
      return(P_NO);
    }
    else
    {
      return(P_YES);
    }
  }
}


static int FindBestSegmentMatch(PAVEDMESH *mesh, LOOP *loop, int nsegment,
                                int ncand[P_MAX_SEGMENT],
                                NODE *cand[P_MAX_SEGMENT][P_MAX_CAND],
                                LOOP *cloop[P_MAX_SEGMENT][P_MAX_CAND],
                                int mustresolv,
                                double *bestshape, int *bestindex, int *bestpair)
{
  int     i, j, k, n1, n2, n3, n4, quad;
  NODE   *curr, *next;
  int     node[4];
  double  shape;
  int     quality, even;
  int     check1, check2;

  *bestshape = 0.0;

  /* try all sides with all candidates */
  curr = loop->Nodes;
  for(i = 0; i < nsegment; i++)
  {
    next = curr->Prev;
    n1 = curr->Node;
    n2 = next->Node;
/*    ncand[i] = 0; */

    /* check nodes are floating */
    if(!(   CHK_BND(mesh->NodeFlag[n1])
         || CHK_BND(mesh->NodeFlag[n2]) ))
    {
      for(j = 0; j < ncand[i]; j++)
      {
      
        n3 = cand[i][j]->Node;
        n4 = cand[i][j]->Next->Node;
        
        quad = mesh->NQuads-1;
        for(k = 0; k < 4; k++)
        {
          if(mesh->QuadNodes[quad][k] == n1)
          {
            node[k] = n3;
          }
          else if(mesh->QuadNodes[quad][k] == n2)
          {
            node[k] = n4;
          }
          else
          {
            node[k] = mesh->QuadNodes[quad][k];
          }
        }
        /* invalid quad definition */
        check1 = 0;
        check2 = 0;
        for(k = 0; k < 4; k++)
        {
          if(node[k] == n3)
            check1++;
          if(node[k] == n4)
            check2++;
        }
        if(check1 == 2 || check2 == 2)
          continue;
        
        shape = QuadShapeQuality(mesh->NodeCoords, node[0], node[1], 
                                                   node[2], node[3]);
        quality = QuadSideLengthQuality(mesh->NodeCoords, 
                                        node[0], node[1], node[2], node[3]);
        
        /* check whether the formulated loops would contain even number 
           of nodes */
        if(loop == cloop[i][j])
        {
          even = CheckEvenNoNode(curr->Prev, cand[i][j]);
        }
        /* with other loops it is always true */
        else
        {
          even = P_YES;
        }
        /* 
         * Either the situation must be resolved, therefore any "good" 
         * shape is accepted or it is really acceptable, by all conditions.
         */
        if(   (even == P_YES)
           && (   ((mustresolv == TRUE) && (*bestshape < shape))
               || ((quality == P_GOOD)  && (*bestshape < shape)) ))
        {
          *bestshape = shape;
          *bestindex = i;
          *bestpair  = j;
        }

#ifdef PAVING_DEBUG
        e_GeneralWarning("Quad test %d %d %d %d : shape: %f sidelength ratio: %d", 
                         node[0]+1, node[1]+1, node[2]+1, node[3]+1, shape, quality);
#endif

        /* we could also try node match, but it is NOT implemented at all */
      }
    }

    curr = next;
  }

  return(P_SUCCESS);
}


static int FactoryMonitor(PAVEDMESH *mesh, LOOP *loop, int nsegment,
                          int backmem, int formem)
{
  int    i, n, seamed;
  LOOP  *newloop;

  int    ncand[P_MAX_SEGMENT];             /* no of candidates per segment */
  NODE  *cand[P_MAX_SEGMENT][P_MAX_CAND];  /* candidates per segment */
  LOOP  *cloop[P_MAX_SEGMENT][P_MAX_CAND]; /* loop of the candidate */
  int    mustresolv;                       /* problems should be resolved ? */
  int    nproblem = 0;
  double bestshape;
  int    bestindex, bestpair;
  int    old, new, quad;
  
  if(CHK_BIT(loop->Type, MUST_RESOLV))
  {
    mustresolv = TRUE;
  }
  else
  {
    mustresolv = FALSE;
  }
  
  /* theoretically it is not possible */
  if(nsegment > P_MAX_SEGMENT)
  {
#ifdef PAVING_DEBUG
    e_GeneralError("!!!!! Too many new segments !!!!!");
#endif
    return(P_FACE_UNDO);
  }

  n = mesh->NQuads-1;
  if(   (mustresolv == FALSE)
     && (QuadShapeQuality(mesh->NodeCoords, 
                          mesh->QuadNodes[n][0], mesh->QuadNodes[n][1], 
                          mesh->QuadNodes[n][2], mesh->QuadNodes[n][3] ) == 0) )
  {
    return(P_FACE_UNDO);
  }


  /* try to do seaming at both ends of new segments */
  /* ---------------------------------------------- */
  seamed = FALSE;
  if(SeamingAtLoopHead(mesh, loop, TRUE) == P_SEAM_DONE)
  {
    seamed = TRUE;
  }

  /* go to the other end of new segments */
  for(i = 0; i < nsegment; i++)
  {
    StepBackward(loop, TRUE);
  }

  if(SeamingAtLoopHead(mesh, loop, TRUE) == P_SEAM_DONE) 
  {
    seamed = TRUE;
  }

  if(seamed)
    return(P_FACE_ACCEPT | P_FACE_RESOLV);

  /* restore loop head  */
  for(i = 0; i < nsegment; i++)
  {
    StepForward(loop, TRUE);
    ncand[i] = 0;
  }


  /* 
   * all new segment should be checked for 
   *   - intersection, 
   *   - proximity, 
   * etc ...
   * with all existing loops
   */
  nproblem = CheckAllSegments(mesh, loop, nsegment, 
                              ncand, cand, cloop, &mustresolv);

  /* everything is perfect, though it cannot be negative ! */
  if(nproblem <= 0)
  {
    return(P_FACE_ACCEPT);
  }
  else if(nproblem > 0)
  {
    /* try to find the best matching segment or node (?) */
    FindBestSegmentMatch(mesh, loop, nsegment, ncand, cand, cloop, mustresolv,
                         &bestshape, &bestindex, &bestpair);
    
    /* decide about merging */
    if(mustresolv == TRUE)
    {
      if(bestshape < 0.4)
      {
        return(P_FACE_UNDO);
      }
    }
    else
    {
      if(bestshape < 0.5)
      {
        return(P_FACE_UNDO);
      }
    }

    if(   (loop == cloop[bestindex][bestpair])
       && (backmem || formem) )
    {
      return(P_FACE_UNDO);
    }

#ifdef PAVING_DEBUG
    e_GeneralWarning("!!!!!!!!!!!!!! MERGING !!!!!!!!!!!!!!!!!!!!!!!!");
#endif

    /* -------------------------------------- */
    /* by now we decided to merge, lets do it */
    /* -------------------------------------- */

    /* put the selected node behind loop head */
    StepForward(loop, TRUE);
    for(i = 0; i < bestindex; i++)
    {
      StepBackward(loop, TRUE);
    }
    
    /* renumber the quads */
    old = loop->Nodes->Prev->Node;
    new = cand[bestindex][bestpair]->Node;
    while(mesh->NNodeElem[old] > 0)
    {
      quad = mesh->NodeElem[old][0];
      ChangeElementNode(quad, old, new, mesh, TRUE);
    }

    if(!CHK_BND(mesh->NodeFlag[old]))
    {
      /* remove the node */
      RemoveBehindFrontNode(mesh, loop, TRUE);
      DeleteNode(mesh, old, TRUE);
    }
    else
    {
      StepBackward(loop, TRUE);
    }

    /* renumber the quads */
    old = loop->Nodes->Prev->Node;
    new = cand[bestindex][bestpair]->Next->Node;
    while(mesh->NNodeElem[old] > 0)
    {
      quad = mesh->NodeElem[old][0];
      ChangeElementNode(quad, old, new, mesh, TRUE);
    }

    if(!CHK_BND(mesh->NodeFlag[old]))
    {
      /* remove the node */
      RemoveBehindFrontNode(mesh, loop, TRUE);
      DeleteNode(mesh, old, TRUE);
    }
    else
    {
      StepBackward(loop, TRUE);
    }

    StepBackward(loop, TRUE);

    /* merge or cut */
    MergeCutLoops(mesh, loop->Nodes, cand[bestindex][bestpair], TRUE);
    
    /* same loop, cut */
    if(loop == cloop[bestindex][bestpair])
    {
      /* initialise new loop, count nodes in it */
      newloop = InitLoop(loop->Type, cand[bestindex][bestpair]);
      AddLoop(mesh, newloop, TRUE);
      /* recount the nodes in this loop */
      CountLoopNodes(loop);

      ADD_BIT(newloop->Type, MUST_SMOOTH);
    }
    /* different loops, merge loops, therefore delete one */
    else
    {
      RemoveLoop(mesh, cloop[bestindex][bestpair], TRUE);
      /* recount the nodes in this loop */
      CountLoopNodes(loop);
    }

    ADD_BIT(loop->Type, MUST_SMOOTH);

#ifdef PAVING_GRAPHICS
    if(ShowGraphics)
    {
      mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
                 0, NULL, 
                 mesh->NQuads, mesh->QuadNodes, 0,
                 mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 
    }
#endif

    return(P_FACE_ACCEPT | P_FACE_RESOLV);
  }

  /* should not come here */
  return(P_FACE_UNDO);
}



