#include "mgni.h"
#include "rgr.h"
#include "dtr.h"
#include "eg_lib.h"

/*****************************************************************************/

static void refine_edge(EDGE *edg, double ***fcoords, int *fnp, double umpr,
                        int geomset, MESH *mesh);
static void register_alignment(int start_n, int nmp, int ***alignments,
                               int *na, int *nat);
static void update_seg_crd(int **alignments, int na, double ***seg_crd,
                           int *nsp);
static int misaligned(double *coord, double bottom, double left, double top,
                      double right, double xdif, double ydif, double gprm,
                      double umpr);
static int realign_bnd_nodes(int ***alignments, int *nat, double **coord,
                             int nsp, double bottom, double left, double top,
                             double right, double xdif, double ydif,
                             double gprm, double umpr);

void make_fake_boundary_nodes(double ***fcoords, int *fnp, BND *fbnd,
                              double umpr, int geomset, MESH *mesh)
{
  int e;                        /* inner_edge index */

  refine_edge(&(fbnd->oe), fcoords, fnp, umpr, geomset, mesh);
  for (e = 0; e < fbnd->nie; e++)
    refine_edge(&(fbnd->ie[e]), fcoords, fnp, umpr, geomset, mesh);

  return;
}

static void refine_edge(EDGE *edg, double ***fcoords, int *fnp, double umpr,
                        int geomset, MESH *mesh)
{
  int b;                        /* index on the edge */
  int fn, sn;                   /* bnd node numbers (first, second) */
  double **seg_crd;             /* coordinates of segment nodes */
  double *seg_mpr;              /* and their mpr's */
  double *seg_t;                /* and their t-params in case of curves */
  NURBSCURVE *curve;            /* nurbs curve for curve refine */
  double tfn;                   /* curve param for fn */
  double tsn;                   /* curve param for sn */
  int nsp;                      /* number of segment points */
  int is_curved;                /* was e_crvref or e_segref used */
  int s, i;                     /* counter */

  for (b = 0; b < edg->nen; b++)
  {
    fn = edg->en[b];
    sn = edg->en[(b + 1) % edg->nen];
    if (geomset)
    {
      if (!e_GetCurve(mesh, fn + 1, sn + 1, &curve, &tfn, &tsn))
      {
        nsp = e_SegmentRefine((*fcoords)[fn], umpr, (*fcoords)[sn], umpr,
                              &seg_crd, &seg_mpr, TRUE, E_SEGREF_NONE);
        is_curved = FALSE;
      }
      else
      {
        nsp = e_CurveRefine(curve, tfn, umpr, tsn, umpr,
                            &seg_crd, &seg_mpr, &seg_t, TRUE, E_SEGREF_NONE);
        adjust_curve_refine_output(&nsp, seg_crd);
        is_curved = TRUE;

      }
    }
    else
    {
      nsp = e_SegmentRefine((*fcoords)[fn], umpr, (*fcoords)[sn], umpr,
                            &seg_crd, &seg_mpr, TRUE, E_SEGREF_NONE);
      is_curved = FALSE;
    }
    if (nsp == 0)
      continue;
    (*fcoords) = e_ext_cont_d_matr((*fnp), 3, (*fcoords), nsp);
    for (s = 0; s < nsp; s++)
      edg->en = e_incr_i_vect(edg->nen + s, edg->en, EDGE_INCR);
    for (s = edg->nen - 1; s > b; s--)
      edg->en[nsp + s] = edg->en[s];
    for (s = 0; s < nsp; s++)
    {
      edg->en[b + 1 + s] = (*fnp);
      for (i = 0; i < 3; i++)
        (*fcoords)[(*fnp)][i] = seg_crd[s][i];
      (*fnp)++;
    }
    edg->nen += nsp;
    b += nsp;
    e_free(seg_mpr);
    e_free_cont_d_matr((is_curved ? nsp + 2 : nsp), seg_crd);
    if (is_curved)
      e_free(seg_t);
  }
}

EDGE *incr_edge_vect(unsigned int length, EDGE *vector,
                     unsigned int LengthIncrem)
{
  /* memory allocation function used in setupbnd */
  EDGE *newvect;
  int e;

  newvect = (EDGE *) e_calloc(length + LengthIncrem, sizeof(EDGE));
  memcpy(newvect, vector, length * sizeof(EDGE));
  for (e = length; e < length + LengthIncrem; e++)
  {
    newvect[e].nen = 0;
    newvect[e].en = NULL;
  }
  return newvect;
}

void copy_bnd(BND *source, BND *target)
{
  int i, e;

  /* only for making dummy copies!
     copied bnd can't be extended */

  target->oe.nen = source->oe.nen;
  for (i = 0; i < target->oe.nen; i++)
    target->oe.en = e_incr_i_vect(i, target->oe.en, EDGE_INCR);
  for (i = 0; i < target->oe.nen; i++)
    target->oe.en[i] = source->oe.en[i];
  target->nie = source->nie;
  for (e = 0; e < target->nie; e++)
  {
    target->ie = incr_edge_vect(e, target->ie, 1);
    target->ie[e].nen = source->ie[e].nen;
    for (i = 0; i < target->oe.nen; i++)
      target->ie[e].en = e_incr_i_vect(i, target->ie[e].en, EDGE_INCR);
    for (i = 0; i < target->ie[e].nen; i++)
      target->ie[e].en[i] = source->ie[e].en[i];
  }
}

#ifdef OUT_OF_DATE
int make_aligned_boundary_nodes(double ***coords, int *np, int *tp, int *stnp,
                                double **mpr, BND *bnd, int ***tpnl,
                                double gprm, double umpr)
{
  int nnp;                      /* number of refined aligned boundary nodes */
  double left, right, top, bottom; /* extremities of the mesh */
  double xdif, ydif;            /* differences between positions */
  int row;                      /* row number */

  int b;                        /* boundary segment index */
  int fn, sn;                   /* bnd node numbers (first, second) */
  double **seg_crd;             /* coordinates of segment nodes */
  double *seg_mpr;              /* and their mpr's */
  int nsp;                      /* number of segment points */

  int nmp;                      /* number of misaligned points */
  int **alignments;             /* matrix containing info about the alignments */
  int na;                       /* number of alignements to be made */
  int nat;                      /* number of spaces allocated in this matrix */
  int start_n;                  /* start of a sequence of misaligned nodes */

  EDGE dummy;                   /* dummy boundary */
  int sb;                       /* boundary element counter */

  int n, i, e;                  /* counters */

  (*stnp) = (*np);
  nnp = 0;
  nat = 1;
  alignments = e_alloc_cont_i_matr(nat, 2);

  /* first retrieve the information about the grid */
  left = (*coords)[0][0];
  right = (*coords)[0][0];
  top = (*coords)[0][1];
  bottom = (*coords)[0][1];
  for (n = 0; n < *np; n++)
  {
    if ((*coords)[n][0] < left)
      left = (*coords)[n][0];
    if ((*coords)[n][0] > right)
      right = (*coords)[n][0];
    if ((*coords)[n][1] < bottom)
      bottom = (*coords)[n][1];
    if ((*coords)[n][1] > top)
      top = (*coords)[n][1];
  }

  xdif = umpr;
  ydif = umpr * sqrt(3.0) / 2.0;

  dummy.nen = bnd->oe.nen;
  dummy.en = e_alloc_i_vect(bnd->oe.nen);
  for (i = 0; i < bnd->oe.nen; i++)
    dummy.en[i] = bnd->oe.en[i];

  sb = 0;

  for (b = 0; b < dummy.nen; b++)
  {
    fn = dummy.en[b];
    sn = dummy.en[(b + 1) % dummy.nen];

    nsp = e_SegmentRefine((*coords)[fn], umpr, (*coords)[sn], umpr,
                          &seg_crd, &seg_mpr, TRUE, E_SEGREF_NONE);
    na = 0;
    if (nsp > 4)
    {
      nmp = 0;
      start_n = -1;
      for (n = 0; n < nsp; n++)
      {
        if (misaligned
            (seg_crd[n], bottom, left, top, right, xdif, ydif, gprm, umpr))
        {
          if (start_n == -1)
          {
            start_n = n;
            nmp++;
          }
          else
          {
            nmp++;
          }
        }
        else
        {
          if (start_n != -1)
          {
            if (nmp > 3)
              register_alignment(start_n, nmp, &alignments, &na, &nat, b);
            start_n = -1;
            nmp = 0;
          }
        }
      }
      if (start_n != -1)
        if (nmp > 3)
          register_alignment(start_n, nmp, &alignments, &na, &nat, b);
      update_seg_crd(alignments, na, &seg_crd, &nsp);
    }
    updatebnd(&(bnd->oe), sb, nsp, (*np));
    sb += nsp + 1;
    while ((*tp) - (*np) < nsp)
    {
      (*coords) = e_incr_cont_d_matr((*tp), 3, (*coords), (*tp));
      (*tpnl) = (int **) e_incr_vp_vect((*tp), (void **) (*tpnl), (*tp));
      (*mpr) = e_incr_d_vect((*tp), (*mpr), (*tp));
      (*tp) += (*tp);
    }
    for (n = 0; n < nsp; n++)
    {
      (*coords)[(*np)][0] = seg_crd[n][0];
      (*coords)[(*np)][1] = seg_crd[n][1];
      (*coords)[(*np)][2] = seg_crd[n][2];
      (*tpnl)[(*np)] = e_alloc_i_vect(2);
      (*tpnl)[(*np)][0] = 0;
      (*tpnl)[(*np)][1] = 0;
      (*mpr)[(*np)] = umpr;
      (*np)++;
      nnp++;
    }
    e_free(seg_mpr);
    e_free_cont_d_matr(nnp, seg_crd);
  }

  for (e = 0; e < bnd->nie; e++)
  {
    e_free(dummy.en);
    dummy.en = e_alloc_i_vect(bnd->ie[e].nen);
    for (i = 0; i < bnd->ie[e].nen; i++)
      dummy.en[i] = bnd->ie[e].en[i];
    dummy.nen = bnd->ie[e].nen;
    sb = 0;
    for (b = 0; b < dummy.nen; b++)
    {
      fn = dummy.en[b];
      sn = dummy.en[(b + 1) % dummy.nen];

      nsp = e_SegmentRefine((*coords)[fn], umpr, (*coords)[sn], umpr,
                            &seg_crd, &seg_mpr, TRUE, E_SEGREF_NONE);
      na = 0;
      if (nsp > 4)
      {
        nmp = 0;
        start_n = -1;
        for (n = 0; n < nsp; n++)
        {
          if (misaligned
              (seg_crd[n], bottom, left, top, right, xdif, ydif, gprm, umpr))
          {
            if (start_n == -1)
            {
              start_n = n;
              nmp++;
            }
            else
            {
              nmp++;
            }
          }
          else
          {
            if (start_n != -1)
            {
              if (nmp > 3)
                register_alignment(start_n, nmp, &alignments, &na, &nat, b);
              start_n = -1;
              nmp = 0;
            }
          }
        }
        if (start_n != -1)
          if (nmp > 3)
            register_alignment(start_n, nmp, &alignments, &na, &nat, b);
        update_seg_crd(alignments, na, &seg_crd, &nsp);
      }
      updatebnd(&(bnd->ie[e]), sb, nsp, (*np));
      sb += nsp + 1;
      while ((*tp) - (*np) < nsp)
      {
        (*coords) = e_incr_cont_d_matr((*tp), 3, (*coords), (*tp));
        (*tpnl) = (int **) e_incr_vp_vect((*tp), (void **) (*tpnl), (*tp));
        (*mpr) = e_incr_d_vect((*tp), (*mpr), (*tp));
        (*tp) += (*tp);
      }
      for (n = 0; n < nsp; n++)
      {
        (*coords)[(*np)][0] = seg_crd[n][0];
        (*coords)[(*np)][1] = seg_crd[n][1];
        (*coords)[(*np)][2] = seg_crd[n][2];
        (*tpnl)[(*np)] = e_alloc_i_vect(2);
        (*tpnl)[(*np)][0] = 0;
        (*tpnl)[(*np)][1] = 0;
        (*mpr)[(*np)] = umpr;
        (*np)++;
        nnp++;
      }
      e_free(seg_mpr);
      e_free_cont_d_matr(nnp, seg_crd);
    }
  }
  if (nat > 0)
    e_free_cont_i_matr(nat, alignments);
  return nnp;
}
#endif

int make_aligned_boundary_nodes_incl_nurbs(double ***coords, int *np, int *tp,
                                           int *stnp, double **mpr, BND *bnd,
                                           int ***tpnl, double gprm,
                                           double umpr, int geomset,
                                           MESH *mesh)
{
  int nnp;                      /* number of refined aligned boundary nodes */
  double left, right, top, bottom; /* extremities of the mesh */
  double xdif, ydif;            /* differences between positions */

  int b;                        /* boundary segment index */
  int fn, sn;                   /* bnd node numbers (first, second) */
  double **seg_crd;             /* coordinates of segment nodes */
  double *seg_mpr;              /* and their mpr's */
  int nsp;                      /* number of segment points */

  int **alignments;             /* matrix containing info about the alignments */
  int na;                       /* number of alignements to be made */
  int nat;                      /* number of spaces allocated in this matrix */

  EDGE dummy;                   /* dummy boundary */
  int sb;                       /* boundary element counter */

  NURBSCURVE *curve;            /* nurbs curve for curve refine */
  double tfn;                   /* curve param for fn */
  double tsn;                   /* curve param for sn */
  double *seg_t;                /* curve params for intermediate nodes */
  int is_curved;                /* is the current segment curved */

  int n, i, e;                  /* counters */

  (*stnp) = (*np);
  nnp = 0;
  nat = 1;
  alignments = e_alloc_cont_i_matr(nat, 2);

  /* first retrieve the information about the grid
     necessary for the alignment detection */
  left = (*coords)[0][0];
  right = (*coords)[0][0];
  top = (*coords)[0][1];
  bottom = (*coords)[0][1];
  for (n = 0; n < *np; n++)
  {
    if ((*coords)[n][0] < left)
      left = (*coords)[n][0];
    if ((*coords)[n][0] > right)
      right = (*coords)[n][0];
    if ((*coords)[n][1] < bottom)
      bottom = (*coords)[n][1];
    if ((*coords)[n][1] > top)
      top = (*coords)[n][1];
  }
  xdif = umpr;
  ydif = umpr * sqrt(3.0) / 2.0;

  /* save the original boundary before you start changing it */
  dummy.nen = bnd->oe.nen;
  for (i = 0; i < bnd->oe.nen; i++)
  {
    dummy.en = e_incr_i_vect(i, dummy.en, EDGE_INCR);
    dummy.en[i] = bnd->oe.en[i];
  }

  sb = 0;

  /* the main boundary refinement function */
  for (b = 0; b < dummy.nen; b++)
  {
    fn = dummy.en[b];
    sn = dummy.en[(b + 1) % dummy.nen];

    if (geomset)
    {
      if (!e_GetCurve(mesh, fn + 1, sn + 1, &curve, &tfn, &tsn))
      {
        nsp = e_SegmentRefine((*coords)[fn], umpr, (*coords)[sn], umpr,
                              &seg_crd, &seg_mpr, TRUE, E_SEGREF_NONE);
        is_curved = FALSE;
      }
      else
      {
        nsp = e_CurveRefine(curve, tfn, umpr, tsn, umpr,
                            &seg_crd, &seg_mpr, &seg_t, TRUE, E_SEGREF_NONE);
        adjust_curve_refine_output(&nsp, seg_crd);
        is_curved = TRUE;

      }
    }
    else
    {
      nsp = e_SegmentRefine((*coords)[fn], umpr, (*coords)[sn], umpr,
                            &seg_crd, &seg_mpr, TRUE, E_SEGREF_NONE);
      is_curved = FALSE;
    }

    if (!is_curved)
    {
      if (nsp > 4)
      {
        na =
          realign_bnd_nodes(&alignments, &nat, seg_crd, nsp, bottom, left,
                            top, right, xdif, ydif, gprm, umpr);
        while (na > 0)
        {
              /*** testing ***/
          na = 1;
          alignments[0][0] = 0;
          alignments[0][1] = nsp;
      /*** testing ***/
          update_seg_crd(alignments, na, &seg_crd, &nsp);
          na =
            realign_bnd_nodes(&alignments, &nat, seg_crd, nsp, bottom, left,
                              top, right, xdif, ydif, gprm, umpr);
        }
      }
    }

    updatebnd(&(bnd->oe), sb, nsp, (*np));
    sb += nsp + 1;
    while ((*tp) - (*np) < nsp)
    {
      (*coords) = e_incr_cont_d_matr((*tp), 3, (*coords), (*tp));
      (*tpnl) = (int **) e_incr_vp_vect((*tp), (void **) (*tpnl), (*tp));
      (*mpr) = e_incr_d_vect((*tp), (*mpr), (*tp));
      (*tp) += (*tp);
    }
    for (n = 0; n < nsp; n++)
    {
      (*coords)[(*np)][0] = seg_crd[n][0];
      (*coords)[(*np)][1] = seg_crd[n][1];
      (*coords)[(*np)][2] = seg_crd[n][2];
      (*tpnl)[(*np)] = e_alloc_i_vect(2);
      (*tpnl)[(*np)][0] = 0;
      (*tpnl)[(*np)][1] = 0;
      (*mpr)[(*np)] = umpr;
      (*np)++;
      nnp++;
    }
    e_free(seg_mpr);
    e_free_cont_d_matr(nnp, seg_crd);
  }

  for (e = 0; e < bnd->nie; e++)
  {
    e_free(dummy.en);
    for (i = 0; i < bnd->ie[e].nen; i++)
    {
      dummy.en = e_incr_i_vect(i, dummy.en, EDGE_INCR);
      dummy.en[i] = bnd->ie[e].en[i];
    }
    dummy.nen = bnd->ie[e].nen;
    sb = 0;
    for (b = 0; b < dummy.nen; b++)
    {
      fn = dummy.en[b];
      sn = dummy.en[(b + 1) % dummy.nen];

      if (geomset)
      {
        if (!e_GetCurve(mesh, fn + 1, sn + 1, &curve, &tfn, &tsn))
        {
          nsp = e_SegmentRefine((*coords)[fn], umpr, (*coords)[sn], umpr,
                                &seg_crd, &seg_mpr, TRUE, E_SEGREF_NONE);
          is_curved = FALSE;
        }
        else
        {
          nsp = e_CurveRefine(curve, tfn, umpr, tsn, umpr,
                              &seg_crd, &seg_mpr, &seg_t, TRUE,
                              E_SEGREF_NONE);
          adjust_curve_refine_output(&nsp, seg_crd);
          is_curved = TRUE;
        }
      }
      else
      {
        nsp = e_SegmentRefine((*coords)[fn], umpr, (*coords)[sn], umpr,
                              &seg_crd, &seg_mpr, TRUE, E_SEGREF_NONE);
        is_curved = FALSE;
      }

      na = 0;

      if (!is_curved)
      {
        if (nsp > 4)
        {
          na =
            realign_bnd_nodes(&alignments, &nat, seg_crd, nsp, bottom, left,
                              top, right, xdif, ydif, gprm, umpr);
          while (na > 0)
          {
            update_seg_crd(alignments, na, &seg_crd, &nsp);
            na =
              realign_bnd_nodes(&alignments, &nat, seg_crd, nsp, bottom, left,
                                top, right, xdif, ydif, gprm, umpr);
          }
        }
      }

      updatebnd(&(bnd->ie[e]), sb, nsp, (*np));
      sb += nsp + 1;
      while ((*tp) - (*np) < nsp)
      {
        (*coords) = e_incr_cont_d_matr((*tp), 3, (*coords), (*tp));
        (*tpnl) = (int **) e_incr_vp_vect((*tp), (void **) (*tpnl), (*tp));
        (*mpr) = e_incr_d_vect((*tp), (*mpr), (*tp));
        (*tp) += (*tp);
      }
      for (n = 0; n < nsp; n++)
      {
        (*coords)[(*np)][0] = seg_crd[n][0];
        (*coords)[(*np)][1] = seg_crd[n][1];
        (*coords)[(*np)][2] = seg_crd[n][2];
        (*tpnl)[(*np)] = e_alloc_i_vect(2);
        (*tpnl)[(*np)][0] = 0;
        (*tpnl)[(*np)][1] = 0;
        (*mpr)[(*np)] = umpr;
        (*np)++;
        nnp++;
      }
      e_free(seg_mpr);
      if (is_curved)
        e_free_cont_d_matr(nsp + 2, seg_crd);
      else
        e_free_cont_d_matr(nsp, seg_crd);
    }
  }
  if (nat > 0)
    e_free_cont_i_matr(nat, alignments);
  return nnp;
}

static void register_alignment(int start_n, int nmp, int ***alignments,
                               int *na, int *nat)
{
  /* register an row of misalignments 'nmp' nodes long starting at 'start_n'
     into the alignments matrix which has 'na' rows and 'nat' allocated rows */

  if ((*nat) == (*na))
  {
    (*alignments) = e_ext_cont_i_matr((*nat), 2, (*alignments), 1);
    (*nat)++;
  }
  (*alignments)[(*na)][0] = start_n;
  (*alignments)[(*na)][1] = nmp;
  (*na)++;
}

static void update_seg_crd(int **alignments, int na, double ***seg_crd,
                           int *nsp)
{
  /* update a segment with nsp segment points to include all the extra
     nodes made by realigning according to the alignments matrix */

  int a, n, s;
  double dist[3];
  /* double dif[3], prev_node[2],next_node[2]; */

  (*seg_crd) = e_ext_cont_d_matr((*nsp), 3, (*seg_crd), na);
  for (a = 0; a < na; a++)
  {
    for (n = (*nsp) - 1; n > alignments[a][0] + alignments[a][1] + a - 2; n--) /* -1 */
    {
      for (s = 0; s < 3; s++)
        (*seg_crd)[n + 1][s] = (*seg_crd)[n][s];
    }
    for (s = 0; s < 3; s++)
      dist[s] =
        ((*seg_crd)[alignments[a][0] + a + alignments[a][1]][s] -
         (*seg_crd)[alignments[a][0] + a][s]) / ((double) (alignments[a][1]));
    for (n = 0; n < alignments[a][1] - 1; n++)
    {
      for (s = 0; s < 3; s++)
        (*seg_crd)[alignments[a][0] + a + n + 1][s] =
          (*seg_crd)[alignments[a][0] + a][s] + (double) (n + 1) * dist[s];
      (*seg_crd)[alignments[a][0] + a + n + 1][2] = -1.0;
    }
    (*seg_crd)[alignments[a][0] + a][2] = -20.0;
    (*seg_crd)[alignments[a][0] + a + alignments[a][1]][2] = -20.0;

 /*** version with half nodes
	for(n=(*nsp)-1;n>alignments[a][0]+alignments[a][1]+a-2;n--)
	{
	    for (s=0;s<3;s++)
		(*seg_crd)[n+1][s]=(*seg_crd)[n][s];
	}
	for (s=0;s<2;s++)
	    prev_node[s]=(*seg_crd)[alignments[a][0]+a][s];
	for(n=0;n<alignments[a][1]-1;n++)
	{
	    for (s=0;s<2;s++)
	    {
		next_node[s]=(*seg_crd)[alignments[a][0]+a+1+n][s];
		(*seg_crd)[alignments[a][0]+a+1+n][s]=
		    (prev_node[s]+
		     (*seg_crd)[alignments[a][0]+a+1+n][s] )/2.0;
	    	prev_node[s]=next_node[s];
	    }
	    (*seg_crd)[alignments[a][0]+a+1+n][2]=-1.0;
	}  ***/


    (*nsp)++;
  }
  na = 0;
}

static int misaligned(double *coord, double bottom, double left, double top,
                      double right, double xdif, double ydif, double gprm,
                      double umpr)
{
  double xpos, ypos;
  int row;
  double d;
  double pos[3];

  pos[2] = 0.0;

  /* dirty inefficient way */
  row = 0;
  xpos = left + ((row % 2) + 1) * (xdif / 2.0);
  ypos = bottom + ydif;
  while (ypos < top)
  {
    while (xpos < right)
    {
      pos[0] = xpos;
      pos[1] = ypos;
      d = eg_PointDist2D(pos, coord);
      if ((d > umpr * gprm * 0.99999) && (d < umpr * gprm * 1.2))
        return TRUE;
      xpos += xdif;
    }
    ypos += ydif;
    row++;
    xpos = left + ((row % 2) + 1) * (xdif / 2.0);
  }
  return FALSE;
}

void adjust_curve_refine_output(int *nsp, double **seg_crd)
{
  /* this little function removes the first and last node from the
     output of a e_curve_refine in order to be compatible with
     the output of e_segment_refine

     **** don't forget to freeup the full length of seg_crd **** */

  int s, j;

  for (s = 1; s < *nsp; s++)
    for (j = 0; j < 3; j++)
      seg_crd[s - 1][j] = seg_crd[s][j];
  *nsp -= 2;
}

static int realign_bnd_nodes(int ***alignments, int *nat, double **coords,
                             int nsp, double bottom, double left, double top,
                             double right, double xdif, double ydif,
                             double gprm, double umpr)
{
  /* do the misalignment checks and build the alignments matrix
     return the number of misalignments */

  int na;                       /* the number of misalignments */
  int nmp;                      /* number of misaligned nodes in one series */
  int start_n;                  /* start of a misalignment series */
  int n;                        /* node counter */

  na = 0;
  nmp = 0;
  start_n = -1;
  for (n = 0; n < nsp; n++)
  {
    if (misaligned
        (coords[n], bottom, left, top, right, xdif, ydif, gprm, umpr))
    {
      if (start_n == -1)
        start_n = n;
      nmp++;
      /* new extension to bnd alignment: alignment series maximum 7 nodes long
         if (nmp>6)
         {
         register_alignment(start_n, nmp, &alignments, &na, &nat, b);
         start_n=-1;
         nmp=0;
         } */
    }
    else
    {
      if (start_n != -1)
      {
        if (nmp > 3)
          register_alignment(start_n, nmp, alignments, &na, nat);
        start_n = -1;
        nmp = 0;
      }
    }
  }
  if (start_n != -1)
    if (nmp > 3)
      register_alignment(start_n, nmp, alignments, &na, nat);

  return na;
}

/* end of file */
