/* FILE rgr_main.c
 *
 * a new semi-structured Delaunay based mesh generator
 * made by Jelle and Peter
 *
 */

#include "mgn.h"
#include "mgn_graphics.h"
#include "rgr_graphics.h"
#include "rgr.h"
#include "etr.h"
#include "dtr.h"
#include "eg_lib.h"
#include "mgn_aux.h"

#define RGR_VERSION "2.1b/ESSTIN"


void rgr_main(MGNI *mgni)
{
  double umpr;                  /* the uniform mesh parameter */
  double gprm;                  /* grid acceptance parameter */
  double rprm;                  /* refinement parameter */

  int use_star;                 /* use star technology */
  int load_mpr;                 /* load mesh parameters */
  int bnd_align;                /* use boundary alignment */
  int old_dln;                  /* use old Delaunay */
  int proj;                     /* use projection instead of centroid */
  int geomset;                  /* use NURBS curves */

  int **tr;                     /* triangle matrix */
  int ntr;                      /* number of triangles */
  int ttr;                      /* allocated spaces in triangle matrix */
  double **coords;              /* coordinates matrix */
  double *mpr;                  /* fake nodal mesh parameter vector */
  int np;                       /* number of points */
  int tp;                       /* allocated spaces in coords */
  int stnp;                     /* starting point for insertion of new nodes */
  int nnp;                      /* n of new pnts made by boundary refinement */
  int *hull;                    /* the convex hull */
  int nh;                       /* number of points in the hull */
  int thf;                      /* allocated spaces in hull */
  BND *bnd;                     /* boundary information */
  int **tpnl;                   /* triangle per node listing */
  int sm;                       /* number of smoothing iterations */
  int i, e, n;            			/* counters */

  double **fcoords;             /* all the fake variables */
  int fnp;
  int fntr;
  int **ftr;
  BND *fbnd;

  MTRF mtrf;
  mtrf.used = FALSE;

  /* get the right options */
  geomset = mgni->geomset;
  sm = mgni->smiter;
  use_star = USE_STAR;
  load_mpr = (mgni->ctrl == CTRL_MPR)?TRUE:FALSE;
  bnd_align = USE_BND_ALIGN;
  old_dln = USE_OLD_DLN;
  proj = USE_PROJ_ALIGN;
  umpr = mgni->h;
  gprm = mgni->alfa;
  rprm = mgni->beta;

  /* initialize eg_lib */
  eg_Init();

  /* more initialization */
  coords = NULL;
  np = 0;
  tr = NULL;
  ntr = 0;
  hull = NULL;
  nh = 0;
  thf = 0;
  tpnl = NULL;
  bnd = NULL;
  
  /* parse the coarse mesh */
  parse_initial_mesh(mgni->mesh, &coords, &np);
  tp = np;
  
  /* create the boundary */
  setup_2D_bnd(mgni->mesh, &bnd);

  /* setup the mesh sizing */
  mpr = e_alloc_d_vect(np);
  if (load_mpr)
  {
    for (i = 0; i < np; i++)
      mpr[i] = mgni->mesh->MeshParam.NodalMeshParams[i];
  }
  else
  {
    for (i = 0; i < np; i++)
      mpr[i] = umpr;
  }

  /* reconnect initial mesh */
  inittpnl(&tpnl, np);
  f_connect_mesh(coords, np, &tr, &ntr, &hull, &nh, tpnl, &mtrf);
  ttr = ntr;
  tp = np;

  /* the regular grid */
  if (!load_mpr)
  {
    /* make fake triangulation */
    ShowGraphics = FALSE;
    fcoords = e_alloc_cont_d_matr(np, 3);
    fnp = np;
    memcpy(&(fcoords[0][0]), &(coords[0][0]), 3 * np * sizeof(double));
    fbnd = (BND *) e_calloc(1, sizeof(BND));
    copy_bnd(bnd, fbnd);
    make_fake_boundary_nodes(&fcoords, &fnp, fbnd, 0.5 * umpr, geomset, mgni->mesh);
    fntr = e_2D_triangulate(&fcoords, &fnp, fbnd, &ftr, &mtrf);
    ShowGraphics = mgni->gra;

    /* make grid nodes */
    nnp = make_grid_points(&coords, &np, &tp, &stnp, &mpr, &tpnl, gprm, umpr,
                       		 fcoords, fnp, ftr, fntr);
                           
    /* triangulate grid points */
    if (PrintProgress)
      printf("\n%d new grid nodes to be inserted\n", nnp);
    if (ShowGraphics)
      mgnDrawAll(np-nnp, coords, ntr, tr, 0, NULL, 0, 
        				 mgnDrawExtraNodes, &np, "the grid nodes");
    add_new_nodes(coords, np, nnp, stnp, &tr, &ntr, &ttr, &hull, &nh, tpnl,
                  &mtrf);

    /* free-up fake triangution */
    e_free_cont_d_matr(fnp, fcoords);
    e_free_cont_i_matr(fntr, ftr);
    e_free(fbnd->oe.en);
    for (e = 0; e < fbnd->nie; e++)
      e_free(fbnd->ie[e].en);
    if (e > 0)
      e_free(fbnd->ie);
    e_free(fbnd);
  }

  /* the carefully aligned boundary nodes (dikke zever!) */
  if ((bnd_align) && (!load_mpr))
    nnp = new_algn_bnd_nodes(&coords, &np, &tp, &stnp,
                             &mpr, bnd, &tpnl, gprm, umpr, 
                             geomset, mgni->mesh, nnp);
  else
    nnp = make_boundary_nodes(&coords, &np, &tp, &stnp, CTRL_MPR,
                              geomset, mgni->mesh, NULL, &mpr, 0.0, 
                              bnd, &tpnl, FALSE);
  if (PrintProgress)
    printf("\n%d new boundary nodes to be inserted\n", nnp);
  if (ShowGraphics)
      mgnDrawAll(np-nnp, coords, ntr, tr, 0, NULL, 0, 
        				 mgnDrawExtraNodes, &np, "the boundary nodes");
  add_new_nodes(coords, np, nnp, stnp, &tr, &ntr, &ttr, &hull, &nh, tpnl,
                &mtrf);

  /* boundary recovery */
  while (TRUE)
  {
    nnp = recover_boundaries(&coords, &np, &tp, &stnp, tr, ntr, &tpnl,
                           	 bnd, CTRL_MPR, &mpr, &mtrf);
    if (nnp == 0)
      break;
    if (PrintProgress)
      printf("\nboundary recovery step: %d new nodes\n", nnp);
    if (ShowGraphics)
      mgnDrawAll(np-nnp, coords, ntr, tr, 0, NULL, 0, 
        				 mgnDrawExtraNodes, &np, "boundary recovery nodes");
    add_new_nodes(coords, np, nnp, stnp, &tr, &ntr, &ttr, &hull,
                  &nh, tpnl, &mtrf);
  }
  
  /* determine inside/outside */
  setareacode(bnd, tpnl, tr, ntr);

  /* for adaptive meshes base refinement to 1.0 + gprm */
  if (load_mpr)
    while (TRUE)
    {
      nnp = simple_refine(&coords, &np, &tp, &stnp, &mpr, tr,
                          ntr, &tpnl, 1.0 + gprm, proj);
      if (nnp == 0)
        break;
      if (PrintProgress)
        printf("\nbase refinement: %d new nodes", nnp);
      if (ShowGraphics)
        mgnDrawAll(np-nnp, coords, ntr, tr, 0, NULL, 0, 
      				  	 mgnDrawExtraNodes, &np, "base refinement nodes");
      add_new_nodes(coords, np, nnp, stnp, &tr, &ntr, &ttr, &hull,
                    &nh, tpnl, &mtrf);
    }

  /* refinement procedure */
  while (TRUE)
  {
    if (old_dln)                  /* old delaunay point insertion */
      nnp = simple_refine(&coords, &np, &tp, &stnp, &mpr, tr,
                          ntr, &tpnl, rprm, proj);
    else                          /* new method */
      nnp = simple_dual_refine(&coords, &np, &tp, &stnp, &mpr, tr,
                               ntr, &tpnl, rprm, use_star);
    if (nnp == 0)
      break;
    if (PrintProgress)
        printf("\nrefinement step: %d new nodes", nnp);
    if (ShowGraphics)
      mgnDrawAll(np-nnp, coords, ntr, tr, 0, NULL, 0, 
      				 	 mgnDrawExtraNodes, &np, "refinement nodes");
    add_new_nodes(coords, np, nnp, stnp, &tr, &ntr, &ttr, &hull,
                    &nh, tpnl, &mtrf);
  }

  /* get the 3rd dimension right */
  for (n = 0; n < np; n++)
    coords[n][2] = 0.0;

  /* smoothing by nodal averaging before output */
  if (mgni->smtype == SM_NODAL_AVG)
    smooth_by_nodal_averaging(coords, np, tr, ntr, tpnl,
                              mgni->smiter);

  /* write the output */
  mgni->outputmesh = (MESH *) e_calloc(1, sizeof(MESH));
  e_NullifyMesh(mgni->outputmesh);
  create_outputmesh(coords, np, tr, ntr, mgni->outputmesh);

  if (ShowGraphics)
    mgnDrawAll((mgni->outputmesh)->NMeshPoints,
               (mgni->outputmesh)->MeshPointCoords,
               (mgni->outputmesh)->NElemsTriang1,
               (mgni->outputmesh)->NodesTriang1, 0, NULL, -1, NULL, NULL,
               "output mesh");
               
  /* free up all consumed memory */
  e_free(mpr);
  e_free(hull);
  e_free_cont_i_matr(ttr, tr);
  e_free_cont_d_matr(tp, coords);
  freetpnl(tpnl, np);
}

/* end of file */
